; Scripts for the Skype internet telephony program
; Author: Doug Lee <dgl@dlee.org> (Skype username dlee42)
; URL: http://www.dlee.org/skype/
; Authorship info from 2008 or before:
; Original Author: Chris Nestrud <ccn@panix.com> (Skype username ChrisN)
; Contributing Developers:
; Noel Romey <ner@uark.edu> (Skype username dj_ner)
; Chad Foster
; Tyler Kavanaugh <twkav@kc.rr.com> (Skype Username Compguy2007)
/*
Copyright (c) 2006-2018 Doug Lee.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the copyright holders and contributors may not be used to
  endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

use "skype_jcpDict.jsb"  ; JAWS Control Property manager
use "skype_compat.jsb"  ; compatibility with older JAWS versions

;#pragma StringComparison partial
;#pragma usePoFile 0
include "hjconst.jsh"
include "hjGlobal.jsh"
include "common.jsm"
include "skype.jsh"
include "Skype.jsm"
include "MSAAConst.jsh"

globals
	int Skype___intAutoStartRun,
	string Skype___strTextBuffer,
	int Skype___intGotNewMessage,
	int Skype___lastChatIdx,
	handle Skype___lastChatHwnd,
	object Skype___lastChatList

void function _checkForNewChats()
; Watches for new chat messages.
;Called by Schedule internally and from AutoStartEvent.  Also is called in sayIncomingChat. [TK, 2007-08-20]
var
	handle hwnd,
	object o,
	int cnt
scheduleFunction("_checkForNewChats", 5)
if !opt__getUserOption(opt__SV_RecognizeChat)
	return
endIf
hwnd = getFocus()
if util__isSpecialFocus() || !schat__isChatWindow(hwnd, False)
	return
endIf
hwnd = schat__getWindow(schat__WOutput, hwnd)
if !hwnd
	return
endIf
if hwnd != Skype___lastChatHwnd
	if conv5__applies()
		Skype___lastChatList = conv5__chatContent(hwnd)
	else
		stats__inc("notConv5_01")
		if !sayAllInProgress()
			sutil__sayUnsupportedSkype()
		endIf
		return
	endIf
endIf
o = Skype___lastChatList
if !o
	return
elif o.accRole(0) != Role_System_List && o.accRole(0) != Role_System_Pane
	return
endIf
cnt = schat__getMessageCount(o)
if hwnd != Skype___lastChatHwnd
	; Don't read right after a conversation window switch.
	Skype___lastChatHwnd = hwnd
	Skype___lastChatIdx = cnt
	return
endIf
if Skype___lastChatIdx >= cnt
	; Nothing to read, but the list may have shrunk.
	; This happens when the user reduces displayed history in this window.
	if Skype___lastChatIdx > cnt
		if !sayAllInProgress()
			sayMessage(OT_Message, formatString(msgChatsNoLongerShown, intToString(Skype___lastChatIdx -cnt)))
		endIf
	endIf
	Skype___lastChatIdx = cnt
	return
endIf
if cnt -Skype___lastChatIdx < 5
	; Say the next chat message.
	; The scheduleFunction at the top of this function will catch anything after that.
	Skype___lastChatIdx = Skype___lastChatIdx +1
else
	; Don't say a whole flood.
	; This can happen when the user clicks for conversation history in this window.
	; The one line that is read is probably not new, so don't read it here.
	if !sayAllInProgress()
		sayMessage(OT_Message, formatString(msgNewChatsNotRead, intToString(cnt -Skype___lastChatIdx)))
	endIf
	Skype___lastChatIdx = cnt
	return
endIf
if !sayAllInProgress()
	sayChatEntry(Skype___lastChatHwnd, Skype___lastChatIdx, "n")
endIf
endFunction

void Function AutoStartEvent ()
log__open("", False)
incsearch__setFlags("h")
if (Skype___intAutoStartRun != 1)
	SayFormattedMessage (OT_APP_START, msgSkypeStart_L, msgSkypeStart_S)
	Skype___intAutoStartRun = 1
EndIf
; if we re-enter a conversation window, the new message code is run. force this code not to run for the first haf-second.
Skype___intGotNewMessage=1
ScheduleFunction ("_resetGotNewMessage", 5)
typing__startPolling()
scheduleFunction("appOpt__resetSkypeCommandKey", 10)
Skype___lastChatIdx = 0
Skype___lastChatHwnd = util__vcast(0)
scheduleFunction("_checkForNewChats", 10)
mkcs__setHelpText(msgMKCSHelpText)
endFunction

void function autoFinishEvent()
typing__stopPolling()
endFunction

Script HotkeyHelp ()
;Added 10/24/2004 by ner, outputs hotkey help info and home page access link.
if UserBufferIsActive () UserBufferDeactivate() EndIf
sutil__sayMessageToUserBuffer(msgHotKeyHelp, True)
EndScript

void function keyboardHelpHook(string ScriptName, string FrameName)
; When scripts have numerical arguments, it mistakes it for frame names.
; This converts it to an integer, checks that it is a valid integer, then nulls the frame name argument.
var int iFrame =StringToInt(FrameName)
If (!iFrame  && GetCharacterValue(FrameName)==48)
|| iFrame
	FrameName=cscNull
EndIf
keyboardHelpHook(ScriptName, FrameName)
endFunction

Script SkypeScriptsHomePage ()
sutil__gotoSkypeScriptPage("", False)
EndScript

script ScreenSensitiveHelp()
if !isSameScript()
	performScript ScreenSensitiveHelp()
	return
endIf
sutil__gotoSkypeScriptPage("skypeman6.php", False)
endScript

globals int Skype___insideSayWindowTitle
script sayWindowTitle()
; Enhances JAWSKey+T.
var
	handle hFocus,
	handle hTop,
	string status
Skype___insideSayWindowTitle = True
performScript sayWindowTitle()
Skype___insideSayWindowTitle = False
hFocus = getFocus()
hTop = getTopLevelWindow(hFocus)
if !appwin__isMainSkypeWindow(hTop)
	if schat__isChatWindow(hTop, False)
		; SayWindowTitle will have said the conversation name by now,
		; so just say the status of the other person, if available.
		status = _getChatStatus(hTop, True)
		if status
			sayMessage(OT_Document_Name, status)
		endIf
	endIf
	return
endIf
hTop = schat__getTopOfChat(hFocus)
if schat__isChatWindow(hTop, False)
	status = _getChatStatus(hTop, False)
	if status
		sayMessage(OT_Document_Name, status)
	endIf
	return
endIf
while hFocus
	if getWindowClass(hFocus) == "TConversationForm"
		; Say the name of this conversation.
		sayMessage(OT_Document_Name, schat__getChatName(hFocus, True))
		return
	endIf
	hFocus = getParent(hFocus)
endWhile
endScript

int function focusWindowSupportsUIA()
if Skype___insideSayWindowTitle && getFocus() != getTopLevelWindow(getFocus())
	return False
endIf
return focusWindowSupportsUIA()
endFunction

string function getWindowName(handle hwnd)
if !Skype___insideSayWindowTitle
|| hwnd != getTopLevelWindow(getFocus())
	return getWindowName(hwnd)
endIf
return MSAA__windowName(hwnd)
endFunction

string function _getChatStatus(handle hTop, int excludeName)
; Get a status description for the indicated chat.
var string status, int pos, string tmp
status = schat__getChatName(hTop, True)
if excludeName
	if stringRight(status, 1) == ")"
		pos = stringContainsFromRight(status, "(")
		status = stringChopLeft(status, pos)
		status = stringChopRight(status, 1)
	else
		status = ""
	endIf
endIf
tmp = schat__convDescription(hTop)
if tmp
	if status
		status = status +" "
	endIf
	status = status +tmp
endIf
return status
endFunction

void function jcp__setCustomProps(collection props, string sOrigin)
; Obtains any properties of the indicated control that stray from what JAWS would normally discover and report.
; Sets props to be a structure for other jcp* functions.
; This is a callback function called by skype_jcpdict.jsb.
if util__isSpecialFocus()
|| isVirtualPCCursor()
	return
endIf

var collection loc = jcp__getLocation(sOrigin, "hm")
lbl__fix(props, sOrigin, loc.hwnd, loc.MSAA.obj, loc.MSAA.childID)
endFunction

globals object Skype___CLObj, int Skype___CLID
void function _sayConversationControlPosition(object o, int childID)
; Say the position and count in the conversationsControl window accurately.
; This is done on a delay to avoid slowing navigation through lists.

if o
	Skype___CLObj = o
	Skype___CLID = childID
	scheduleFunction("_sayConversationControlPosition", 10)
	return
endIf
o = Skype___CLObj
childID = Skype___CLID

; Count backward for speed.
var int i = o.accChildCount
; Assume all items count for n and just subtract those that don't, again for speed.
var int n = i
; but count up the position, which favors being near the top of the list a bit.
var int pos = 0
var int curItemFound = 0
while i && !isKeyWaiting()
	; Things that count are all listItems.
	if o.accRole(i) == Role_System_ListItem
	; This keeps out things like "Today" and "Older than one week" in the Conversations list.
	&& stringCompare(o.accName(i), o.accDescription(i), True) != 0
		if i == childID
			curItemFound = True
		endIf
		if curItemFound
			pos = pos +1
		endIf
	else
		n = n -1
	endIf
	i = i -1
endWhile
if !pos
	; If we don't have something accurate to say, say nothing.
	; This applies, for example, to "All other contacts" in the contact list.
	return
endIf
sayMessage(OT_Position, formatString(cmsgMyPosInGroup1, intToString(pos), intToString(n)))
endFunction

string function _getBannerText(handle hwnd)
; Get and return any text from the given (assumed banner or Help/About) window.
var int childID
var object o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	return ""
endIf
var int n = o.accChildCount
childID = 1
var string txt, string part
while childID <= n
	part = o.accName(childID)
	if part
		if txt
			txt = txt +"\n"
		endIf
		txt = txt +part
	endIf
	childID = childID +1
endWhile
return txt
endFunction

void function _clearBanners()
; Wipe any visible banner off the screen.
; These things contain no text visible to JAWS, don't take focus, and block out text in other windows and apps.
; Also works against the Help/About box when not in focus, because it acts the same way.
var
 handle hwnd, handle hwndFocus,
	int safety
if userBufferIsActive()
	return
endIf
hwndFocus = getFocus()
safety = 3
while safety
	hwnd = findTopLevelWindow("TDCBanner", "")
	if !hwnd || !isWindowVisible(hwnd) || hwnd == hwndFocus
		hwnd = findTopLevelWindow("TAboutSplash", "")
		if !hwnd || !isWindowVisible(hwnd)
			return
		endIf
	endIf
	var string txt = _getBannerText(hwnd)
	if txt
		;playSound(findJAWSSoundFile("ascend.wav"))
		pause()
		postMessage(hwnd, 0x10, 0, 0)  ; WM_Close
		pause()
		sutil__sayMessageToUserBuffer(txt, False)
		; Let the next focus change catch any other such windows.
		return
	else
		playSound(findJAWSSoundFile("descend.wav"))
	endIf
	postMessage(hwnd, 0x10, 0, 0)  ; WM_Close
	pause()
	safety = safety -1
endWhile
if !safety
	log__write("Loop ran out.")
endIf
endFunction

globals
; The getTickCount() value when the last FocusChangedEventEx was suppressed.
	int Skype___lastFCESuppressTick,
; A flag used on Tab and Shift+Tab to avoid focus change rerouting.
	int Skype___tabbed

void function focusChangedEventEx(
	handle hwndFocus, int nObject, int nChild,
	handle hwndPrevFocus, int nPrevObject, int nPrevChild,
	int nChangeDepth)
var string wclass = stringLower(getWindowClass(hwndFocus))

var int tabbed = Skype___tabbed
Skype___tabbed = False

; Block chat input focus events on each typed character in Skype 5.1.
; Same for the box for entering phone numbers (tested in Skype 5.3).
; Same for the mood text edit box.
if hwndFocus == hwndPrevFocus
&& stringContains(wclass, "richedit")
	Skype___lastFCESuppressTick = getTickCount()
	return
endIf

; Force focus from invisible TZapCommunicator to chat input edit box.
; This can happen when the Communicator is "Application Access."
; Also applies against ad windows.
var int forceFocus = False
if !tabbed
	forceFocus = !isWindowVisible(hwndFocus) || (
		getWindowClass(hwndFocus) == "TZapCommunicator"
		&& getWindowBottom(hwndFocus) -getWindowTop(hwndFocus) == 0
	) || _isAdWindow(hwndFocus)
	|| getWindowClass(hwndFocus) == "TChatEntryControl"
endIf
if forceFocus
	if appwin__getSkypeWindow(appwin__SWT_ChatInput)
		simulateKey("control+4")
	elif findWindow(getTopLevelWindow(getFocus()), "TSearchControl", "")
		setFocus(findWindow(getTopLevelWindow(getFocus()), "TSearchControl", ""))
	endIf
	return
endIf

; Move focus sensibly on exit from the emoticon selector.
if wclass == "TSmileyButtonPanel"
&& (!getWindowStyleBits(hwndPrevFocus) || getWindowClass(hwndPrevFocus) == "TSmileySelector")
	setFocus(findWindow(getParent(hwndFocus), "TChatRichEdit", ""))
	return
endIf

; Speak alerts like the one for old-style conferences on chat focus.
if wclass == "TChatEntryControl" && nObject == ObjID_Client
	_sayTextItems(hwndFocus)
	; And continue as normal.
endIf

; Fix focus when the Add To Conference button (Skype 5.3) is pressed.
; This brings up a sort of menu whose root window takes focus
; but whose first item does not.
var handle hwndChild = getFirstChild(hwndFocus)
if wclass == "TCallPopup" && getWindowClass(hwndChild) == "TWidgetControl"
	setFocus(hwndChild)
	MSAA__objectFromWindow(hwndChild).accSelect(SelFlag_TakeFocus, 1)
	return
endIf

; Fix focus in the Skype 5.1 Technical Call Info dialog.
; The TCheckBoxUserListControl is from the Send Contacts list.
var handle hwnd
if wclass == "TCallTechInfoForm" && getWindowClass(hwndChild) == "TWidgetControl"
	hwnd =  hwndChild
elif wclass == "TCheckBoxUserListControl"
	hwnd = hwndFocus
endIf
if hwnd
	; Replace this focus event with a better one.
	var object o, int childID, int iRole
	o = getFocusObject(childID)
	iRole = o.accRole(childID)
	if iRole == Role_System_Pane || iRole == Role_System_Client
		o = MSAA__objectFromWindow(hwnd)
		if hwnd != hwndFocus
			o.accSelect(SelFlag_TakeFocus, 0)
		endIf
		; The actual list is represented by a single descendant List object.
		; Focus that.
		; Note that 1 here is for the List object, not the first ListItem.
		o.accSelect(SelFlag_TakeFocus, 1)
		; And swallow the current focus event.
		return
	endIf
endIf

; Fix a problem that applies in JAWS 7.0 and later.
if hwndFocus != hwndPrevFocus && nChangeDepth == -1
	nChangeDepth = getFocusChangeDepth()
endIf
focusChangedEventEx(hwndFocus, nObject, nChild, hwndPrevFocus, nPrevObject, nPrevChild, nChangeDepth)
endFunction

script tab()
Skype___tabbed = True
performScript tab()
endScript

script shiftTab()
Skype___tabbed = True
performScript shiftTab()
endScript


int function _isAdWindow(handle hwnd)
; Returns True if focus is or is within an advertisement window.
while hwnd
	if getWindowClass(hwnd) == "TChatBanner"
		return True
	endIf
	hwnd = getParent(hwnd)
endWhile
return False
endFunction

void function _sayTextItems(handle hwnd)
; Say the text items in the given window's MSAA top level.
var object o = MSAA__objectFromWindow(hwnd)
if !o
	return
endIf
var string buf = ""
var string thisText
var int iRole
var int textSeen = False
var int addThis
var int n = o.accChildCount
var int i = 1
while i <= n
	addThis = False
	iRole = o.accRole(i)
	if iRole == Role_System_StaticText
		textSeen = True
		addThis = True
	elif textSeen && iRole == Role_System_Link
		addThis = True
	elif textSeen
		; Something non-text and non-link after text was seen.
		i = n  ; exit loop early.
	endIf
	if addThis
		thisText = stringTrimTrailingBlanks(stringTrimLeadingBlanks(o.accName(i)))
		if thisText
			buf = buf +" " +thisText
		endIf
	endIf
	i = i +1
endWhile
buf = stringTrimLeadingBlanks(buf)
if buf
	sayMessage(OT_Message, buf)
endIf
endFunction

void function tutorMessageEvent(handle hwndFocus, int nMenuMode)
; Blocks extra TutorMessageEvent speech when Skype 5.1 issues extra
; focusChange events while the user types into a chat window.
if Skype___lastFCESuppressTick && getTickCount() -Skype___lastFCESuppressTick < 500
	Skype___lastFCESuppressTick = 0
	return
endIf
tutorMessageEvent(hwndFocus, nMenuMode)
endFunction

Void Function FocusChangedEvent (handle FocusWindow, handle PrevWindow)
if util__isSpecialFocus()
	focusChangedEvent(focusWindow, prevWindow)
	return
endIf
scheduleFunction("_clearBanners", 5)
var string wclass = getWindowClass(focusWindow)
if wclass == "Shell DocObject View"
|| wclass == "Shell Embedding"
|| appwin__isMainSkypeWindow(focusWindow)
	; Handle focus here by forcing it to the Conversations list.
	; The first two force focus toward an Internet Explorer_Server window.
	scheduleFunction("fixFocus", 2)
	return
endIf
if getTopLevelWindow(focusWindow) != getTopLevelWindow(prevWindow)
	; If we've switched to a conversation window, silence recognizing for a haf-second so it won't inadvertently speak
	if appwin__isChatWindow(focusWindow)
		Skype___intGotNewMessage=1
		ScheduleFunction ("_resetGotNewMessage", 5)
		scheduleFunction("_checkCapsLock", 5)
	EndIf
	var handle thwnd = getTopLevelWindow(focusWindow)
	if thwnd != getTopLevelWindow(prevWindow)
	&& schat__isChatWindow(focusWindow, True)
		Skype___intGotNewMessage=1
		ScheduleFunction ("_resetGotNewMessage", 5)
		if (focusWindow == thwnd || focusWindow == schat__getWindow(schat__WMain, thwnd))
			;sutil__play("DoDoop3.wav")
			util__setFocus(schat__getWindow(schat__WInput, getTopLevelWindow(focusWindow)), "d")
			return
		endIf
	endIf
endIf
focusChangedEvent(focusWindow, prevWindow)
EndFunction

void function fixFocus()
var
	handle focusWindow
focusWindow = getFocus()
var string wclass = getWindowClass(focusWindow)
if wclass == "Shell DocObject View"
|| wclass == "Shell Embedding"
	if getFirstChild(focusWindow)
		setFocus(getFirstChild(focusWindow))
	endIf
	scheduleFunction("fixFocus", 5)
elif appwin__isMainSkypeWindow(focusWindow)
	stopSpeech()
	conv5__focusShowingList()
endIf
endFunction

void function _checkCapsLock()
; Reports if CapsLock is on.
var
	int bIsDown, int bIsOn
if !getKeyState("CapsLock", bIsDown, bIsOn)
	return
endIf
if !bIsOn
	return
endIf
sayMessage(OT_Error, msgCapsLockIsOn)
endFunction

int function saySpecialHighlighting(string moveFunc, string speechFunc, string origScript)
; Speak highlighted text as it draws in windows requiring special highlight handling.
; if moveFunc is non-null, it is called first to move the cursor/highlight.
; If speechFunc is non-null, it is called in some cases to speak text.
; origScript should always be non-null and is the name of the script calling this function.
if util__isSpecialFocus() || !isPCCursor()
	formatStringWithEmbeddedFunctions("<$" +origScript +">")  ; Calls the original script
	return False
elif !speechFunc && (
	handle42conversationNavigation(origScript)
)
	return True
endIf
var handle hwnd = getFocus()
var string lwclass = stringLower(getWindowClass(hwnd))
if (stringContains(lwclass, "history")
|| stringContains(lwclass, "alertlist"))
&& moveFunc
	; Reduce double speech in the History page when the user arrows past the bottom and scrolls the list.
	; SayNextLine speaks when the list scrolls, but the ActiveItemChangedEvent is supposed to speak this.
	formatStringWithEmbeddedFunctions("<" +moveFunc +">")
else
	formatStringWithEmbeddedFunctions("<$" +origScript +">")  ; Calls the original script
endIf
pause()
return True
endFunction

int function handle42conversationNavigation(string origScript)
; For contact/conversation/history list area:
; This makes Up/Down cross between live and non-live, and double Home/End also do so.
var handle hwnd = getFocus()
if getWindowClass(hwnd) != "TConversationsControl"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
childID = o.accFocus
var int goTop = False
if !childID
&& o.accParent.accRole(0) != Role_System_List
	; Probably on the pane object.
	; I haven't found a way to focus the correct item here.  [DGL, 2010-04-25]
	if !stringContains(stringLower(origScript), "home")
		return False
	endIf
	; Make Home try to go to the top of the list.
	; This should make repeated Homes be a fast way to get to an incoming call.
	goTop = True
endIf
var string name = o.accName(childID)
formatStringWithEmbeddedFunctions("<$" +origScript +">")  ; Calls the original script
pause()
var int childID1
o = getFocusObject(childID1)
childID1 = o.accFocus
if childID1 != childID || stringCompare(o.accName(childID1), name, True) != 0
	; We moved, so don't do anything.
	_sayConversationControlPosition(o, childID1)
	return True
endIf
var int delta = -1
if stringContains(stringLower(origScript), "next")
|| stringContains(stringLower(origScript), "end")
	delta = 1
endIf
if goTop
	childID1 = 1
	delta = 1
else
	childID1 = childID1 +delta
endIf
var int safety = 50
var int n = o.accChildCount
while childID1 > 0 && childID1 <= n && safety
	; This detects a "real" contact/conversation entry.
	if o.accRole(childID1) == Role_System_ListItem
	&& !stringIsBlank(o.accValue(childID1))
		o.accSelect(SelFlag_TakeFocus, childID1)
		sutil__play("qw_click1.wav")
		_sayConversationControlPosition(o, childID1)
		return True
	elif o.accRole(childID1) == Role_System_PushButton
	&& o.accName(childID1) == scHideHistory
	&& delta == 1
		; A way to detect History is showing.
		; We only need this when going down.
		; The above MSAA name/value trick fails to find history items because they have no MSAA value.
		; This moves focus from Live to History.
		typeKey("Alt+3")
		sutil__play("qw_click1.wav")
		_sayConversationControlPosition(o, childID1)
		return True
	endIf
	childID1 = childID1 +delta
	safety = safety -1
endWhile
if !safety
	log__write("Loop ran out.")
endIf
_sayConversationControlPosition(o, childID1)
return True
endFunction

script sayNextLine()
if !cycleThroughMSAAControls("next")
	saySpecialHighlighting("nextLine", "", "sayNextLine")
endIf
endScript

script sayPriorLine()
if !cycleThroughMSAAControls("prev")
	saySpecialHighlighting("priorLine", "", "sayPriorLine")
endIf
endScript

script JAWSHome()
if !cycleThroughMSAAControls("first")
	saySpecialHighlighting("JAWSHome", "", "JAWSHome")
endIf
endScript

script JAWSEnd()
if !cycleThroughMSAAControls("last")
	saySpecialHighlighting("JAWSEnd", "", "JAWSEnd")
endIf
endScript

script sayWord()
saySpecialHighlighting("", "sayWord", "sayWord")
endScript

script sayCharacter()
saySpecialHighlighting("", "sayCharacter", "sayCharacter")
endScript

void function sayLineUnit(int unitMovement, /*?*/ int bMoved)
; Avoid double speech when a list, like contacts or conversations, scrolls on arrowing.
var object o, int childID
if util__isSpecialFocus() || !isPCCursor()
	sayLineUnit(unitMovement, bMoved)
	return
endIf
o = getFocusObject(childID)
if !o
	sayLineUnit(unitMovement, bMoved)
	return
endIf
if o.accRole(childID) == Role_System_ListItem
	; Let SayObjectActiveItem handle speech for this navigation.
	return
endIf
sayLineUnit(unitMovement, bMoved)
endFunction

void function sayWordUnit(int unitMovement, /*?*/ int bMoved)
if util__isSpecialFocus() || !isPCCursor()
	; Works around JAWS 14's failure to speak here
	; when focus is in a contact or conversation list.
	sayWord()
	return
endIf
sayWordUnit(unitMovement, bMoved)
endFunction

int function cycleThroughMSAAControls(string cmd)
; In MSAA windows, cycle among the subcontrols available.
; cmd: next, prev, first, or last, to indicate which way to move.
var
	string wclass,
	handle hwnd,
	object o, int childID, int childID1, int n
if !isPCCursor() || isVirtualPCCursor() || util__isSpecialFocus()
	return False
endIf
hwnd = getFocus()
wclass = getWindowClass(hwnd)
if stringCompare(wclass, "TAccessibleRadioButtonEx", False) == 0
	return handleRadioButtonFocus(hwnd, cmd)
endIf
if !(
wclass == "TMyselfControl"
|| wclass == "TPanelButtons"
|| wclass == "TZapCommunicator"
|| wclass == "TCallQualityControlHost"
|| wclass == "TConversationToolbar"
|| wclass == "TContactProfile"
|| wclass == "TConversationProfile"
|| wclass == "TConversationHeader"
|| wclass == "TLiveConversation"
|| wclass == "TConversationVideo"
|| wclass == "TChatContentControl"
|| wclass == "TConversationsControl"
)
	return False
endIf
o = getFocusObject(childID)
if childID != 0 && (
wclass == "TChatContentControl"
|| wclass == "TConversationsControl"
)
	return False
elif wclass == "TConversationsControl" && !childID
&& o.accParent.accRole(0) == Role_System_List
	; A non list item in the conversation or contact list.
	return False
endIf
if stringCompare(o.accName(childID), schat__scVolume, True) == 0
	; See saySpecialHighlighting().
	return False
endIf
typeCurrentScriptKey()
pause()
if getFocus() != hwnd
	return True
endIf
if conv5__applies()
	; Avoids focusing Contact Request button in Skype 5.0 Gold.
	return True
endIf
o = getFocusObject(childID1)
if childID1 != childID
	return True
endIf
n = o.accChildCount
if cmd == "first"
	childID = 1
elif cmd == "last"
	childID = n
elif cmd == "next"
	childID = childID +1
	if childID > n
		childID = 1
	endIf
elif cmd == "prev"
	childID = childID -1
	if childID < 1
		childID = n
	endIf
else
	beep()
	return True
endIf
if childID == childID1
	sayObjectActiveItem()
else
	o.accSelect(SelFlag_TakeFocus, childID)
endIf
return True
endFunction

int function handleRadioButtonFocus(handle hwnd, string cmd)
typeCurrentScriptKey()
pause()
if getFocus() != hwnd
	return True
endIf
hwnd = getParent(hwnd)
var handle hwnd1 = hwnd
if cmd == "next"
	if getNextWindow(hwnd1)
		hwnd1 = getNextWindow(hwnd1)
	else
		hwnd1 = getFirstWindow(hwnd1)
	endIf
elif cmd == "prev"
	if getPriorWindow(hwnd1)
		hwnd1 = getPriorWindow(hwnd1)
	else
		hwnd1 = getLastWindow(hwnd1)
	endIf
elif cmd == "first"
	hwnd1 = getFirstWindow(hwnd1)
else  ; last
	hwnd1 = getLastWindow(hwnd1)
endIf
hwnd1 = getFirstChild(hwnd1)
if hwnd1 == hwnd
	return True
endIf
setFocus(hwnd1)
return True
endFunction

Void Function ActiveItemChangedEvent (handle curHwnd, int curObjectId, int curChildId, handle prevHwnd, int prevObjectId, int prevChildId)

var
	string wclass
wclass = stringLower(getWindowClass(curHwnd))
ActiveItemChangedEvent(curHwnd, curObjectId, curChildId, prevHwnd, prevObjectId, prevChildId)
if wclass == "TInboxList" || stringContains(wclass, "conversation")
	scheduleFunction("sayContactLine", 10)
endIf
endFunction

void function sayChatEntry(handle hwnd, int which, string flags)
; Says one chat message.
; Which: positive = 1-based index, -1 = last line, -2 = line above that, -3 = line two above the last, etc.
; Positive indices only supported starting in Skype 3.5.
; Asking for line 0 returns the last line.
; flags:
;	v for virtual buffer.
;	n for new (incoming) message / no special info like "not delivered."
var
	object o, int childID,
	int iRole,
	int nmsgs
flags = stringLower(flags)
if !hwnd
	hwnd = schat__getWindow(schat__WOutput, getTopLevelWindow(getFocus()))
endIf
if !which
	which = -1  ; latest line sent
endIf
if stringContains(flags, "v")
	schat__virtualizeChatMessages(hwnd, which, 100, 0)
	return
endIf

if conv5__applies()
	o = conv5__chatContent(hwnd)
else
	sutil__sayUnsupportedSkype()
	return
endIf
iRole = o.accRole(0)
if !o || (iRole != Role_System_List && iRole != Role_System_Pane)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
nmsgs = schat__getMessageCount(o)
if which < 0
	which = which +nmsgs +1
endIf
if which < 1 || which > nmsgs
	if nmsgs < 1
		sayMessage(OT_Error, msgNoChatLines)
	else
		sayMessage(OT_Error, msgNotThatManyChatLines)
	endIf
else
	var string msg
	msg = schat__getMessageText(o, which, flags)
	if stringContains(flags, "v")
		sutil__sayMessageToUserBuffer(msg, False)
	else
		sutil__sayURMessage(msg, "", False)
	endIf
endIf  ; valid message number requested
endFunction

script sayAChatLine()
; Attached to alt+1 through alt+0, which read the last message on up respectively.
var
	handle hwnd,
	string keyname,
	int idx,
	string flags
if !schat__isChatWindow(getFocus(), False)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
hwnd = schat__getWindow(schat__WOutput, getTopLevelWindow(getFocus()))
keyname = getCurrentScriptKeyName()
idx = stringContains("1234567890", stringRight(keyname, 1))
flags = ""
if isSameScript()
	flags = flags +"v"
endIf
sayChatEntry(hwnd, 0-idx, flags)
endScript

int function _countSelectedChatMessages(object o)
; Return how many messages are selected in the given MSAA list.
var
	int i, int n, int cnt
cnt = 0
n = o.accChildCount
i = 1
while i <= n
	if (o.accState(i) & State_System_Selected)
	&& o.accRole(i) == Role_System_ListItem
		cnt = cnt +1
	endIf
	i = i +1
endWhile
return cnt
endFunction

script sayRelativeChatLine(int which)
; Say the previous, current, next, first, or last chat message.
; If IsSameScript is True and the current message is being requested, the current message and context is virtualized.
; If IsSameScript is True and the first message is being requested, the first unread message is read.
var
	handle hwnd,
	object o, int msgno,
	int msgcount,
	string keyName,
	string flags
if !schat__isChatWindow(getFocus(), False)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
hwnd = schat__getWindow(schat__WOutput, getTopLevelWindow(getFocus()))
if !hwnd
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
flags = ""
if conv5__applies()
	o = conv5__chatContent(hwnd)
else
	sutil__sayUnsupportedSkype()
	return
endIf
if !o
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
msgcount = schat__getMessageCount(o)
if which == -2 || which == 2
	if which == -2
		msgno = 1
		if isSameScript()
			msgno = _firstUnreadMessage(o)
		endIf
	elif which == 2
		msgno = -1
	endIf
	if msgno > msgcount
		beep()
		return
	endIf
	sayChatEntry(hwnd, msgno, flags)
	schat__selectMessage(o, msgno)
	return
endIf
msgno = schat__selectedMessage(o)
if !msgno
	beep()
	msgno = 1
	schat__selectMessage(o, msgno)
	pause()
	msgno = schat__selectedMessage(o)
endIf
if !msgno
	beep()
	return
endIf
msgno = msgno +which
if msgno < 1 || msgno > msgcount
	beep()
	msgno = msgno -which
else
	schat__selectMessage(o, msgno)
endIf
if which == 0 && !isSameScript()
	var int msgno1, string buf
	msgno1 = schat__selectedMessage(o)
	if msgno1 > 0
		buf = formatString(msgSelectedChatMessage, intToString(msgno1), intToString(msgcount))
	else
		msgno1 = _countSelectedChatMessages(o)
		buf = formatString(msgSelectedChatMessages, intToString(msgno1), intToString(msgcount))
	endIf
	sutil__sayURMessage(buf, "", False)
endIf
if isSameScript() && !which
	flags = flags +"v"
endIf
sayChatEntry(hwnd, msgno, flags)
endScript

int function _firstUnreadMessage(object o)
; Return the index of the first unread message in the indicated message list.
var int i = 1
var int n = o.accChildCount
var string name
var string sNew = " " +schat__scNew
var int len = stringLength(sNew)
for i = 1 to n
	name = o.accName(i)
	if stringRight(name, len) == sNew
		return i
	endIf
endFor
beep()
return n
endFunction

globals
	handle Skype___lastChatWindow
void function sayIncomingChat()
if sutil__getSkypeVersionCode(False) < "v3.5"
	; Work with the last window flagged by the NewTextEvent as having new message text.
	if !sayAllInProgress()
		sayChatEntry(Skype___lastChatWindow, -1, "n")
	endIf
else
	_checkForNewChats()
endIf
endFunction

Void Function NewTextEvent (handle hwnd, string buffer, int nAttributes,
int nTextColor, int nBackgroundColor, int nEcho, string sFrameName)
var
	int doDefault,
	string strClass,
	int stc

typing__schedulePoll(3)
doDefault = False
strClass = StringLower(GetWindowClass (hwnd))
if getScreenEcho() != 1
	; Handle everything in the default manner if screenEcho is None or All.
	doDefault = True
elif stringStartsWith(strClass, "tty")
|| stringStartsWith(strClass, "telnet")
|| stringStartsWith(strClass, "console")
	; Works around a bug in JAWS 8/9/10 or so that makes redrawn console text speak when a Skype dialog closes and uncovers it.
	return
elif !appwin__isSkypeWindow(hwnd) || appwin__isSkypeNotifyWindow(hwnd)
	; Handle non-Skype windows in the default manner.
	; Handle Skype notifications in the default manner so SkypeWatch can catch them.
	doDefault = True
endIf
if doDefault
	newTextEvent(hwnd, buffer, nAttributes, nTextColor, nBackgroundColor, nEcho, sFrameName)
	return
endIf

; Don't repeat text we've just seen
if !stringCompare(buffer, Skype___strTextBuffer, True)
Return
EndIf

; Speak various messages automatically.
if stringStartsWith(strClass, "TLoginControl")  ; during login
|| stringStartsWith(strClass, "TWhatsNewGuide")  ; Getting Started wizard pages
|| ((
	stringStartsWith(strClass, "TZapCommunicator")  ; Questions for active calls
	|| stringStartsWith(strClass, "TCallQuality")  ; pop-up caused by pressing the Check Settings button
	|| stringStartsWith(strClass, "TConversationVideo")  ; Allow/Do Not Allow prompts for video etc.
	) && !Skype___intGotNewMessage  ; new text not caused by a top-level focus change
)
;|| stringStartsWith(strClass, "TOptionTabCaptionControl")  ; Descriptions of Options pages
	sayMessage(OT_Message, buffer)
	focusPopupIfNecessary(hwnd)
	return
elif getWindowClass(getTopLevelWindow(hwnd)) == "TChangePasswordForm"
&& strClass == "TRichLabel"
	sayMessage(OT_Message, buffer)
	return
endIf

; Automate speech of banner controls.
if stringContains(strClass, "banner")
	sayMessage(OT_Message, buffer)
	return
endIf

Skype___strTextBuffer = buffer

; only speak conversation text if we're actually in the relevant conversation window, the user wants it, and the window update was not just a "Last message was posted x minutes ago" update.
if (appwin__isChatWindow(getFocus())
		&& (hwnd != getFocus())
		&& (buffer != "*** ")
		&& hwnd == schat__getWindow(schat__WOutput, getFocus())
		&& (opt__getUserOption(opt__SV_RecognizeChat) == 1))
	if (Skype___intGotNewMessage == 0)
		Skype___lastChatWindow = hwnd
		scheduleFunction("sayIncomingChat", 5)
		Skype___intGotNewMessage=1
		ScheduleFunction ("_resetGotNewMessage", 20)
	EndIf
	Return
EndIf
newTextEvent(hwnd, buffer, nAttributes, nTextColor, nBackgroundColor, nEcho, sFrameName)
EndFunction

globals
handle Skype___hwndPopup
void function focusPopupIfNecessary(handle hwnd)
; Focus popup call warning and feedback windows when they appear if apparently appropriate.
; Called from NewTextEvent with a window handle,
; and called internally on a schedule to make the actual focus change.
var
	string strClass
if !hwnd
	; Called from below after a delay.
	if Skype___intGotNewMessage
		; Prevent focus forcing when the focusChangedEvent fires after the newTextEvent on a switch into a conversation.
		return
	endIf
	if getFocus() != Skype___hwndPopup
		setFocus(Skype___hwndPopup)
	endIf
	return
endIf

; Called directly by NewTextEvent.
strClass = getWindowClass(hwnd)
if strClass != "TZapCommunicator"
; Call Quality omitted because it takes focus too often.
; TODO: Should be included but only when newly focusing a conversation doesn't trigger it.
;&& !stringStartsWith(strClass, "TCallQuality")
		; We handle only those window classes just listed here.
		return
	elif Skype___intGotNewMessage
		; Focus just switched to this conversation,
		; which means the NewTextEvent was probably from that rather than from a fresh popup.
		; Don't focus this in that case.
		return
	endIf
	Skype___hwndPopup = hwnd
	scheduleFunction("focusPopupIfNecessary", 5)
endFunction

Void Function sayHighlightedText (handle hwnd, string buffer)
var
	int tc, int stc
; If we're in a special situation, don't play with it.
if util__isSpecialFocus()
	sayHighlightedText(hwnd, buffer)
	return
endIf
; Don't speak highlighting outside of the current top-level window.
; Example:  When someone signs in/out and you're in a conversation but the Contacts tab is also visible on screen,
; JAWS otherwise repeats the currently highlighted Contact List line.
; Exception: In at least the Look In combo box of Send File, JAWS 7.0 and older won't speak during arrowing unless the ComboLBox exception is included here.
if getTopLevelWindow(hwnd) != getTopLevelWindow(getFocus())
&& getWindowClass(hwnd) != "ComboLBox"
	return
endIf
tc = getWindowTypeCode(hwnd)
; Don't speak highlighting in Skype 2.0+ contact lists because lines are spoken via MSAA there.
; Exception for JAWS 7.0 and lower:  ActiveItemChangeEvent doesn't fire there during arrowing.
; WT_MultiLine_Edit added 2010-08-10 because this otherwise broke speaking of text selection, like Ctrl+A.
if tc != WT_Edit && tc != WT_ComboBox && tc != WT_MultiLine_Edit
&& sutil__isMSAASupport(hwnd)
	return
endIf
; Don't speak non-focused list highlight changes, such as the Hotkey list in Options when the user moves through the options page list to reach that page.
if getWindowTypeCode(hwnd) == WT_ListBox && hwnd != getFocus()
	return
endIf
sayHighlightedText(hwnd, buffer)
EndFunction

Script PickToolBarItem()
var
	string names, string data,
	string curcontact,
	int n, int i
n = tb__makeList(names, data)
if !n
	sayMessage(OT_Error, msgNoToolbarItems)
	return
endIf

i = dlgSelectItemInList(names, msgToolbarItemSelectionTitle, False)
speechOff()
stopSpeech()
pause()
speechOn()
if !i
	sayMessage(OT_Status, msgCanceled)
else
	tb__clickItem(stringSegment(data, "|", i))
endIf
pause()
scheduleFunction("sayObjectTypeAndText", 3)
EndScript

script sayTypingStatus()
var
	string status
if !typing__getText(status)
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
if status
	if isSameScript() > 1
		spellString(status)
	else
		sutil__sayURMessage(status, "", False)
	endIf
	return
endIf
sutil__sayURMessage(msgNoTyping, "", False)
endScript

script toggleChatMessageTimes()
var
	handle hwnd,
	string msg
hwnd = getFocus()
if !schat__isChatWindow(hwnd, False)
|| conv5__focusedListName() == conv5__ListContacts
	sayCurrentScriptKeyLabel()
	typeCurrentScriptKey()
	return
endIf
msg = opt__userOptionCycler(opt__SV_ChatMessageTimes, opt__msgChatMessageTimeChoices, False)
msg = msgMessageTimes +" " +msg
sayUsingVoice(VCTX_Message, msg, OT_Status)
endScript

globals
	int Skype___keyCaseSch,
	string Skype___keyCaseID

void function keyCaseSet(string case)

Skype___keyCaseID = case
if Skype___keyCaseSch
	unscheduleFunction(Skype___keyCaseSch)
	Skype___keyCaseSch = 0
endIf
if !case
	return
endIf
var handle hwnd
hwnd = stringToHandle(case)
JAWSCursor()
moveToWindow(hwnd)
if getCurrentWindow() != hwnd && getParent(getCurrentWindow()) != hwnd
	beep()
endIf
if !util__setRestrictionToWindow(hwnd)
	beep()
endIf
if !isRepeatKey()
PCCursor()
endIf
sayMessage(OT_Message, getWindowText(hwnd, Read_Everything))
endFunction

string function keycaseGet()
return Skype___keyCaseID
endFunction

void function keyCaseCheck(int direct)
if direct
	Skype___keyCaseSch = scheduleFunction("keyCaseCheck", 5)
	return
endIf
if keycaseGet()
	if !isJAWSCursor()
		saveCursor() JAWSCursor()
		setRestriction(RestrictAppWindow)
		restoreCursor()
		keyCaseSet("")
	elif getCurrentWindow() != stringToHandle(keycaseGet())
	&& getParent(getCurrentWindow()) != stringToHandle(keycaseGet())
		beep()
		moveToWindow(stringToHandle(keycaseGet()))
	endIf
else
endIf
endFunction

Script ScriptFileName ()
ScriptAndAppNames (msgScriptName)
if !isSameScript()
	sutil__sayURMessage(customInfo(), "", False)
endIf
EndScript

void function addHotkeyLinks()
userBufferAddText(customInfo())
addHotkeyLinks()
endFunction

string function customInfo()
var
	int swStatus,
	string buf
buf = formatString(msgSkypeAppVersionString_L, sutil__Script_Version, sutil__Script_Revision)
buf = buf +"\13\10"
swStatus = sutil__isSkypewatchLoaded()
if swStatus == 2
	buf = buf +msgSWLoaded
elif swStatus == 1
	buf = buf +msgSWLoaded_old
else
	buf = buf +msgSWNotLoaded
endIf
return buf
endFunction

script sayAppVersion()
var
	string buf
; The funny length restriction makes versions like "3.0" get ignored so we can get a more specific answer.
; Skype, starting around v3.0.32.171 (and possibly 3.0.0.171), began returning versions like "3.0" for ProductVersion (used by default JAWS script)
; and the full version for FileVersion.  [DGL, 2006-12-03]
if stringLength(getVersionInfoString (getAppFilePath(), "ProductVersion")) < 5
	buf = getVersionInfoString(getAppFilePath(), "ProductName")
	if !buf
		buf = getAppFileName()
	endIf
	buf = formatString(msgSkypeAppVersionFmt, buf, stringChopLeft(sutil__getSkypeVersionCode(False), 1))
	if isSameScript()
		sutil__sayMessageToUserBuffer(buf, False)
	else
		sutil__sayURMessage(buf, "", False)
	endIf
	return
endIf
performScript sayAppVersion()
endScript

int function sayContactLine()
var handle hwnd = getFocus()
if !isPCCursor() || util__isSpecialFocus() || !sutil__isMSAASupport(hwnd)
	return False
endIf
var string wclass = stringLower(getWindowClass(hwnd))

if wclass == "TInboxList" || stringContains(wclass, "conversation")
	var object o
	if conv5__applies()
		o = conv5__chatContent(hwnd)
	else
		sutil__sayUnsupportedSkype()
		return False
	endIf
	if wclass == "TConversationList" || wclass == "TInboxList" || wclass == "TConversationsControl"
		var collection cContact = sutil__getCurrentContactInfo()
		var string flist
		if !stringCompare(cContact.Name, cContact.ID, True)
			flist = "Time,ExtStat"
		else
			flist = "ID,Time,ExtStat"
		endIf
		var string fld, string val, string line
		forEach fld in cContact
			val = cContact[fld]
			if val
				if !(fld == "Name" || fld == "ID" || fld == "Status")
					line = line +","
				endIf
				line = line +" " +val
			endIf
		endForEach
		sayMessage(OT_Line, line)
	else  ; not historyList or Contacts page
		line = sutil__removeHTML(o.accHelp(o.accFocus+0))
		if !line
			line = getObjectHelp()
		endIf
		sayMessage(OT_Line, line)
	endIf
	return True
endIf
log__write("Something got past the MSAA code.")
return False
endFunction

Script SayLine ()
; Makes SayLine read Options/Sounds checkbox list lines more effectively.
; Also makes a triple SayLine virtualize chats from a chat list window.
var int iss = isSameScript()
if util__isSpecialFocus() || iss == 1 || !isPCCursor()
	performScript sayLine()
return
endIf
if iss > 1
	var handle hwndFocus = getFocus()
	var handle hwnd = schat__getWindow(schat__WOutput, getTopLevelWindow(hwndFocus))
	if hwndFocus == hwnd
		; Allows virtualization of any message from the message list window.
		var object o, int msgno
		if conv5__applies()
			o = conv5__chatContent(hwnd)
		else
			sutil__sayUnsupportedSkype()
			return
		endIf
		msgno = schat__selectedMessage(o)
		sayChatEntry(hwnd, msgno, "v")
		return
	endIf
endIf
performScript sayLine()
EndScript

script switchToTab()
var
	variant null,
	handle hwnd, handle hwndWanted,
	handle hwndCurrent,
	string which,
	object o, int childID,
	int succeeded
util__waitForKeysToClear(15)
which = stringRight(getCurrentScriptKeyName(), 1)
if !conv5__applies()
	sutil__sayUnsupportedSkype()
	return
endIf
var string swt = stringSegment(formatString("%1|%2|%3|%4|%5|%6",
	appwin__SWT_ContactList, appwin__SWT_ConversationList, appwin__SWT_LiveConversations,
	appwin__SWT_ChatInput, appwin__SWT_ChatOutput, appwin__SWT_ChatContact
), "|", stringToInt(which))
if !swt
	beep()
	return
endIf

if swt == appwin__SWT_ContactList
	typeKey("Alt+1")
	pause()
	; Fall through to refine focus.
elif swt == appwin__SWT_ConversationList
	typeKey("Alt+2")
	pause()
	; Fall through to refine focus.
elif swt == appwin__SWT_ChatInput
	hwnd = appwin__getSkypeWindow(swt)
	if hwnd
		if getFocus() != hwnd
			util__setFocus(hwnd, "d")
		else
			sayObjectTypeAndText()
		endIf
	else
		sayMessage(OT_Error, msgWindowNotFound)
	endIf
	return
elif swt == appwin__SWT_ChatContact
	; Profile area, just read except on isSameScript.
	hwnd = appwin__getSkypeWindow(swt)
	if !hwnd
		sayMessage(OT_Error, msgWindowNotFound)
		return
	endIf
	if !isSameScript()
		var object oProfile = MSAA__objectFromWindow(hwnd)
		if oProfile
			sayMessage(OT_User_Requested_Information, schat___getExtraMessageMaterial(oProfile, 0, True))
		else
			sayMessage(OT_User_Requested_Information, getWindowText(hwnd, Read_Everything))
		endIf
		return
	endIf
	if getFocus() == hwnd
		sayObjectTypeAndText()
	else
		util__setFocus(hwnd, "d")
	endIf
	return
endIf

; Refine focus for contacts, conversations, live conversations, and chat message list.
stopSpeech()
o = appwin__getMSAAObject(swt)
if !o
	sayMessage(OT_Error, msgWindowNotFound)
	return
elif o.accState(childID) & State_System_Invisible
	sayMessage(OT_Error, msgWindowNotVisible)
	return
elif swt == appwin__SWT_LiveConversations
&& o.accChildCount == 0
	sayMessage(OT_Error, msgNoLiveConversations)
	return
endIf
if o.accParent.accState(0) & State_System_Focused
	sayObjectTypeAndText()
else
	o.accParent.accSelect(SelFlag_TakeFocus, 0)
endIf
o.accSelect(SelFlag_TakeFocus, 0)
endScript

string function positionInGroup()
; Override of standard JAWS function to make the current position in some combo boxes speak.
; Author:  Doug Lee
var
	handle hwnd,
	string result,
	object o, int childID,
	int i, int n
hwnd = getCurrentWindow()
result = positionInGroup()
; The "navigable" part is for TNavigableTntComboBox windows.
; These occurred in an active call tab as part of the input/output sound control widgets.
if stringLength(result) && !(result == "0 " && getWindowClass(hwnd) == "tNavigable")
	return result
endIf
if getWindowTypeCode(hwnd) == WT_ComboBox || stringContains(stringLower(getWindowClass(hwnd)), "combo")
	; Window.client.child3 (which is the list window) .client (which is the actual list object)
	o = getObjectFromEvent(hwnd, 0-4, 0, childID).accChild(3).accChild(-4)
	if !o
		return ""
	endIf
	i = o.accSelection +0
	n = o.accChildCount
	if i
		return formatString(cmsgMyPosInGroup1, intToString(i), intToString(n))
	else
		return formatString(cmsgMyPosInGroup2, intToString(n))
	endIf
endIf
return ""
endFunction

script MoveToChatEntryField()
simulateKey("control+4")
endScript

Void Function _resetGotNewMessage ()
Skype___intGotNewMessage=0
EndFunction

int function dialogActive()
; Override to make the result more accurate for Skype.
; Author:  Doug Lee
var
	handle hwnd,
	string wclass
if dialogActive()
	return True
elif util__isSpecialFocus()
	return False
endIf
hwnd = getTopLevelWindow(getFocus())
wclass = getWindowClass(hwnd)
if stringRight(wclass, 13) == ".UnicodeClass"
	wclass = stringChopRight(wclass, 13)
endIf
if wclass == "Tsk" && stringRight(wclass, 4) == "Form" && wclass != "TskMainForm"
	return True
endIf
return False
endFunction

script Enter()
var
	handle hwnd, handle hwnd1
if util__isSpecialFocus() || isVirtualPCCursor()
	performScript enter()
	return
endIf
hwnd = getFocus()
typeCurrentScriptKey()
if !dialogActive()
	return
endIf
pause()
if hwnd != getFocus()
	return
endIf
hwnd = getTopLevelWindow(hwnd)
hwnd1 = findWindow(hwnd, "", _scOK)
if !hwnd1
	hwnd1 = findWindow(hwnd, "", _scSave)
endIf
if !hwnd1
	hwnd1 = findWindow(hwnd, "", _scSearch)
endIf
if !hwnd1
	hwnd1 = findWindow(hwnd, "", _scNext)
endIf
if !hwnd1
	hwnd1 = findWindow(hwnd, "", _scFinish)
endIf
if hwnd1
	saveCursor() JAWSCursor() saveCursor()
	moveTo( (getWindowLeft(hwnd1)+getWindowRight(hwnd1))/2, (getWindowTop(hwnd1)+getWindowBottom(hwnd1))/2)
	leftMouseButton()
endIf
endScript

globals
	int Skype___KPELastKey,
	int Skype___KPERepeatCount

int function isRepeatKey()
; Like isSameScript without a time limit between presses of the same key/combination.
return Skype___KPERepeatCount
endFunction

void function keyPressedEvent(int nKey, string strKeyName, int nIsBrailleKey, int nIsScriptKey)
; Makes the Applications (Context Menu) key work where it should but doesn't already.
; Helps with setting the Skype command key.
; Helps JAWS keep up when Space or Ctrl+Space is pressed in a contact list.
; Makes JAWS speak volume sliders correctly on arrows.
; Fixes a Skype 6.9+ issue where JAWS can't see the content of a chat edit on first message.
; Adds incremental search to the Skype emoticon panel.
if userBufferIsActive() || menusActive()
	keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
	return
endIf
var handle hwnd = getFocus()
var string wclass = getWindowClass(hwnd)
if wclass == "TSmileySelector"
|| (wclass == "TPopupContainer" && getWindowClass(getParent(hwnd)) == "TPickerPopup")
	keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
	incsearch__gotoNext(strKeyName)
	return
endIf
if handleSkypeCommandKey(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
	return
endIf
if nKey == Skype___KPELastKey
	Skype___KPERepeatCount = Skype___KPERepeatCount +1
else
	Skype___KPELastKey = nKey
	Skype___KPERepeatCount = 0
endIf
keyPressedEvent(nKey, strKeyName, nIsBrailleKey, nIsScriptKey)
keyCaseCheck(True)
if wclass == "TChatRichEdit"
	scheduleFunction("cre__manage", 4)
	return
endIf

var int isSpace = !stringCompare(strKeyName, "space", False)
if isSpace && (
	wclass == "TCheckBoxUserListControl"
	|| live5selector__selected() != 0
)
	saveCursor()  JAWSCursor()  saveCursor()
	routeJAWSToPC()
	leftMouseButton()
	restoreCursor()  restoreCursor()
	scheduleFunction("sayLine", 2)
elif isSpace || !stringCompare(strKeyName, "control+space", False)
	; Works in JAWS 8+ but causes no error messages in earlier versions.
	scheduleFunction("MSAARefresh", 5)
elif (isSpace || stringContains(stringLower(strKeyName), "arrow"))
&& wclass == "TVolumeControl"
	scheduleFunction("stopSpeech", 0)
	scheduleFunction("MSAARefresh", 2)
	scheduleFunction("sayWord", 3)
endIf
if nKey == 93
	scheduleFunction("checkForMenu", 3)
endIf
endFunction

int function handleSkypeCommandKey(int nKey, string strKeyName, int nIsBrailleKey, int nIsScriptKey)
; Handles the KeyPressedEvent part of the Skype command key system.
; Returns True if something was done and False if not.
if Skype__appOpt__giSettingSkypeCommandKey
	if nIsScriptKey || nIsBrailleKey
		; Scripts abort the change because otherwise things like Enter can make quite a mess!
		beep()
	elif stringContains(",esc,enter,tab,space,", ","+stringLower(strKeyName)+",")
		; Disallowed Skype Command keys.
		beep()
		if strKeyName == "Esc"
			; Abort the change.
		else
			; Stay in this mode.
			return
		endIf
	else
		; Write to the .ini from here, where we actually have the key name.
		opt__setSkypeVar(opt__SV_SkypeCommandKey, strKeyName)
	endIf
	; The Space will cause another jvoSkypeCommandKey() call to be issued by DlgSelectFunctionToRun().
	; That will update the screen to show the new key, and jvoSkypeCommandKey() will wrap up this process.
	typeKey("Space")
	return True
endIf
return False
endFunction

void function checkForMenu()
; Helper for KeyPressedEvent.
if util__isSpecialFocus()
	return
endIf
if !stringContains(stringLower(getWindowClass(getFocus())), "profile")
&& !stringContains(stringLower(getWindowClass(getFocus())), "conversationheader")
	return
endIf
saveCursor() JAWSCursor() saveCursor()
routeJAWSToPC()
pause()
if getTopLevelWindow(getWindowAtPoint(getCursorCol(), getCursorRow())) == getTopLevelWindow(getFocus())
	rightMouseButton()
else
	beep()
endIf
endFunction

script sayBottomLineOfWindow()
; Says the total Skype online user count if possible,
; or otherwise reads the bottom line of the current window as usual.
; A double press does the normal function regardless.
var string s = conv5__peopleOnline()
if s
	sutil__sayURMessage(s, "", False)
	return
endIf
var
	handle hwnd,
	int oldRestriction,
	string buf
hwnd = appwin__getSkypeWindow(appwin__SWT_ContactList)
if hwnd && isWindowVisible(hwnd) && !isWindowObscured(hwnd)
	hwnd = getParent(hwnd)
	saveCursor() invisibleCursor() saveCursor()
	moveToWindow(hwnd)
	oldRestriction = getRestriction()
	util__setRestrictionToWindow(hwnd)
	JAWSPageDown()
	JAWSEnd() JAWSHome()
	buf = getLine()
	sutil__sayURMessage(buf, "", False)
	setRestriction(oldRestriction)
	return
endIf
performScript sayBottomLineOfWindow()
endScript

string function _shortenFolderPath(object byRef fso, string fname)
; Convert long folder paths (names containing components not adhering to the old DOS 8.3 format) to short (8.3-compatible) paths.
; Used by CompileSkypeScripts to avoid numerous problems with passing space-containing file paths through batch files.
var
	object oFolder
if !fso
	fso = util__getFSObject()
	if !fso
		; Can't shorten, so return what we got.
		return fname
	endIf
endIf
oFolder = fso.getFolder(fname)
if !oFolder
	return fname
endIf
return oFolder.shortPath
endFunction

script compileSkypeScripts()
; Compile all scripts in this script set and, in JAWS 5.10 and later, reload them and make them take effect.
var
	object fso,
	string compiler, string JAWSDrive, string JAWSDir, string scriptDir,
	string cmd
JAWSDir = _shortenFolderPath(fso, getJAWSDirectory())
scriptDir = _shortenFolderPath(fso, getJAWSSettingsDirectory())
compiler = scriptDir +"\\SkypeComp.bat"
JAWSDrive = substring(compiler, 1, 2)
JAWSDir = stringChopLeft(JAWSDir, 2)
scriptDir = stringChopLeft(scriptDir, 2)
cmd = compiler +" " +JAWSDrive +" " +JAWSDir +" " +scriptDir
run(cmd)
scheduleFunction("reloadAllConfigs", 50)
endScript

Script ReadBoxInTabOrder ()
; Makes JAWS read text more effectively in Skype dialog boxes.
If !InHJDialog ()
	SayWindow(GetTopLevelWindow (GetFocus ()), Read_Everything)
	return
EndIf
PerformScript ReadBoxInTabOrder()
EndScript

script leftMouseButton()
if isPCCursor()
	saveCursor() JAWSCursor() saveCursor()
	routeJAWSToPC()
endIf
performScript leftMouseButton()
endScript

script rightMouseButton()
if isPCCursor()
	saveCursor() JAWSCursor() saveCursor()
	routeJAWSToPC()
endIf
performScript rightMouseButton()
endScript

script sayAChatName(int which)
; Say the name of a conversation, or list the names if which == 0.
schat__switchToChat(2, which)
endScript

script sayAllChatNames()
; Says the names of all open conversations.
schat__switchToChat(2, 0)
endScript

script switchToChat(int which)
; Switch to a specific conversation by index, or bring up a list for selection by passing 0.
schat__switchToChat(1, which)
endScript

void function _refreshMSAAInfo()
var
	handle hwnd
if util__isSpecialFocus()
	return
endIf
hwnd = getFocus()
if !isMSAAWindow(hwnd)
	return
endIf
if getWindowClass(getParent(hwnd)) == "TConversationsControl"
&& stringContains(stringLower(getWindowClass(hwnd)), "list")
	if getWindowClass(hwnd) == "THistoryList"
		MSAARefresh()
	endIf
endIf
endFunction

void function sayObjectActiveItem(int bSayPosition)
_refreshMSAAInfo()
sayObjectActiveItem(bSayPosition)
endFunction

void function ObjStateChangedEvent(handle hObj, int iObjType, int nChangedState, int nState, int nOldState)
; Block extra repeats of History List items during arrowing.
if !util__isSpecialFocus() && isPCCursor()
&& getWindowClass(getFocus()) == "THistoryList"
	return
endIf
ObjStateChangedEvent(hObj, iObjType, nChangedState, nState, nOldState)
endFunction

void function ValueChangedEvent(handle hwnd, int objId, int childId, int nObjType, string sObjName, string sObjValue,optional int bIsFocusObject)
if getWindowClass(hwnd) == "TNavigableComboBox"
	Say(sObjValue, Ot_line)
	return
endIf
ValueChangedEvent(hwnd, objId, childId, nObjType, sObjName, sObjValue, bIsFocusObject)
endFunction

script toggleDialPad()
; Toggle the DialPad on/off and focus it when on.
var
	handle hwnd,
	object o, int childID
hwnd = findWindow(appwin__getSkypeWindow(appwin__SWT_ChatMain), "TLiveConversation", "")
if hwnd && getWindowClass(getFirstChild(hwnd)) == "TLiveConversation"
	hwnd = getFirstChild(hwnd)
endIf
if !hwnd || !isWindowVisible(hwnd) || !isWindowVisible(getParent(hwnd))
	sayMessage(OT_Error, msgCantFindCT)
	return
endIf
o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	sayMessage(OT_Error, msgCantGetCT)
	return
endIf
childID = MSAA__childByName(o, scCallToolBar)
if !childID
	sayMessage(OT_Error, msgNoDialPadButton)
	return
endIf
o = o.accChild(childID)
childID = MSAA__childByName(o, scDialPad)
if !childID
	sayMessage(OT_Error, msgNoDialPadButton)
	return
endIf
if !o.accDefaultAction(childID)
endIf
o.accDoDefaultAction(childID)
pause()
o = o.accParent
var int isOpen = False
if MSAA__childByName(o, scLiveDialPad)
	isOpen = True
endIf
if isOpen
	sayMessage(OT_Status, msgDPActive)
else
	sayMessage(OT_Status, msgDPClosed)
endIf
endScript

script JAWSDelete()
; Makes Delete work in lists where it otherwise deletes what's selected instead of what's in focus.
; At this writing, that's the Contact list and the Conversation list.
var
	string wclass,
	object o, int childID
wclass = getWindowClass(getFocus())
if util__isSpecialFocus() || (wclass != "TConversationList" && wclass != "TInboxList")
	performScript JAWSDelete()
	return
endIf
saveCursor()  PCCursor()
o = sutil__getFocusObject(childID)
if !(o.accState(childID) & State_System_Selected)
	typeKey("Space")
	pause()
endIf
performScript JAWSDelete()
endScript

script InsertEmoticon()
var handle hwnd = getFocus()
var string wclass = getWindowClass(hwnd)
if wclass != "TChatRichEdit"
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
hwnd = getPriorWindow(hwnd)
if getWindowClass(hwnd) != "TSmileyButtonPanel"
	sayMessage(OT_Error, msgNotInChatWindow)
	return
endIf
var object o = MSAA__objectFromWindow(hwnd)
var int n = o.accChildCount
if n == 3  ; starts in Skype 7.16.
	n = 2
elif n == 5 || n == 6  ; Same version
	n = 5
elif n == 1  ; older Skypes back through 6.22 or so.
	; Leave it at 1.
else
	n = 0
endIf
if !n || o.accRole(n) != Role_System_PushButton
	sayMessage(OT_Error, msgNotInChatWindow)
	return
elif o.accState(n) & State_System_Unavailable
	sayMessage(OT_Error, cMsgNotAvailable)
	return
endIf
o.accSelect(SelFlag_TakeFocus, n)
pause()
if o.accFocus() != n
	beep()
	return
endIf
typeKey("space")
endScript

int function hasUnreadContent(object o, int childID)
; Return True if the given MSAA item is a "new" conversation.
; Assumes the item is a conversation list entry.
; "New" means "contains something unread."
; New conversations' entries are taller than read ones.
if sutil__getMSAAItemDim(o, childID, "h") > 30
	return True
endIf
return False
endFunction

void function sayCharacterUnit(int unit)
if getObjectSubtypeCode() == WT_LeftRightSlider
	MSAARefresh()
	pause()
elif lbl__isTCIHeaderItem()
	sayLine()
endIf
sayCharacterUnit(unit)
endFunction

int function caretVisible()
; For some reason, the Skype 4 contact list reports caretVisible() True briefly after an up/down navigation that scrolls the list.
; Symptom:  JAWS saying "blank" or reading the just-left contact line before reading the newly selected one.
; This little intercept is to stop that.
if !util__isSpecialFocus() && isPCCursor()
&& stringContains(stringLower(getWindowClass(getFocus())), "list")
	return False
endIf
return caretVisible()
endFunction

script CloseSkype()
; Closes Skype without causing it to prompt for confirmation, for example, to confirm permission to close an active call.
; This is a convenience over having to close Skype from the system tray.
var
	handle hwnd
hwnd = appwin__getSkypeWindow(appwin__SWT_Main)
if !hwnd
	sayMessage(OT_Error, msgNoMainWindow)
	return
endIf
if !isSameScript()
	sayMessage(OT_User_Requested_Information, msgCloseSkypeConfirmation)
	return
endIf
; Heads up:  This will close Skype instantly, with all calls, file transfers, etc.
postMessage(hwnd, 0x10, 0, 0)  ; 0x10 is WM_Close
endScript

script upALevel()
; Close a Skype conversation without removing its entry from the Conversation list.
; Alt+F4 does this in Compact view, but no key does it natively in Default view:
; Ctrl+F4 removes from the Conversations list, Alt+F4 closes Skype's main window,
; and Esc does nothing.
var
	handle hwnd
hwnd = getFocus()
if util__isSpecialFocus() || !schat__isChatWindow(hwnd, False)
	performScript upALevel()
	return
endIf
while hwnd
	if getWindowClass(hwnd) == "TConversationForm" && getParent(hwnd)
		; Non top-level conversation; Default view.
		postMessage(hwnd, 0x10, 0, 0)  ; WM_Close
		return
	endIf
	hwnd = getParent(hwnd)
endWhile
; Not in a Default-view conversation, so just do what the key would have done.
performScript upALevel()
endScript

handle function appGetWindow(string which)
; Test harness compatibility; standard name for this function.
return appwin__getSkypeWindow(which)
endFunction

script sayWhoIsActive()
; Say who is in a multiperson voice conference.
sayConferenceParticipants(False)
endScript

script sayWhoIsSpeaking()
; Say who in a multi-person conference is currently speaking.
; Technically this really says who is currently making the most noise.
sayConferenceParticipants(True)
endScript

void function sayConferenceParticipants(int justSpeakers)
; Say who in a multi-person conference is currently participating or speaking.
var collection speakers = conf__getParticipants(justSpeakers)
var int err = speakers.err
if err
	if err == 1
		sayMessage(OT_Error, msgNotInChatWindow)
	elif err == 2
		sayMessage(OT_Error, msgNotInLiveChat)
	elif err == 3
		sayMessage(OT_Error, msgNotInConference)
	else
		sayMessage(OT_Error, formatString(msgUnknownError, intToString(err)))
	endIf
	return
endIf

var int i, collection speaker
var string buf = ""
for i = 1 to speakers.count
	speaker = speakers[intToString(i)]
	if buf
		buf = buf +" + "
	endIf
	buf = formatString("%1%2", buf, speaker.name)
endFor
if !buf
	buf =  msgNobodySpeaking
endIf
;buf = buf + " " +intToString(speakers.tc)
sayMessage(OT_User_Requested_Information, buf)
endFunction

; Interceptions of JAWS 15 scripts, to prevent "Unknown script call" messages.
script readHistoryItem()
sayCurrentScriptKeyLabel()
typeCurrentScriptKey()
endScript

script AssignConversation(int iConversationID)
sayMessage(OT_Error, msgNotInTheseScripts)
endScript


;======================== Module schat (from SkypeChat) ========================
; Subsystem for handling Skype conversation windows and their contents.
; Interface:
;	schat__isChatWindow(hwnd, restricted)  --> lWellIsIt:  Indicates if the given handle is part of a valid conversation window structure.
;	schat__getChatName(hwnd[, includeStatus]) --> sChatName:  Gets a speakable/printable name for a conversation from its handle.
;	schat__getChatWindows() --> cChats:  Gets a collection of all open conversation window handles.
;	schat__switchToChat(int absolute, int delta) --> lSuccess:  Name, list, or move to a conversation.
;	schat__getWindow(hChatTop, sWhich) --> hwnd:  Gets a window by SKCW_* identifier from a given conversation.
;	schat__virtualizeChatMessages(hwnd, start, contextSize, contextPos) --> lSuccess: Virtualizes conversation messages with context.
;
; Author:  Doug Lee

const
; Link types for schat___addMessageToUserBuffer and supporting code.
	schat___LTMsg = 1,  ; The whole message is a single link ("Continue last conversation with ...").
	schat___LTHistory = 2,  ; "Show messages from: ..." history links.
	schat___LTURL = 3,  ; URLs in user-typed messages.
	schat___LTPlay = 4   ; A voicemail notification terminated by a Play link.

int function schat__isCompactView()
; Returns True if Skype is in Compact view and False if it is in Default view.
var
	handle hwnd
hwnd = appwin__getSkypeWindow(appwin__SWT_Main)
if hwnd
	hwnd = findWindow(hwnd, "TSkypeSplitter", "")
	if hwnd
		return !isWindowVisible(hwnd)
	endIf
endIf
; TODO: We can't determine the view, so Default is assumed here.
beep()
log__write("Unable to determine the active view.")
return False
endFunction

handle function schat__getTopOfChat(handle hwnd0)
; Get the top-level window of the conversation containing the given window.
; In Compact view, this will actually be a top-level window,
; but in Default view, it will be the TConversationForm containing the conversation,
; which will still be a descendant of the main Skype window.
var
	handle hwnd,
	handle hTop,
	variant null
hTop = getTopLevelWindow(hwnd0)
hwnd = hwnd0
while hwnd
	if getWindowClass(hwnd) == "TConversationForm"
		return hwnd
	endIf
	if hwnd == hTop
		hwnd = null
	else
		hwnd = getParent(hwnd)
	endIf
endWhile
; For Skype 4 Default view, in which exactly 0 or 1 conversations can be active.
; This means any window in the Skype default view is considered part of a conversation,
; as long as a conversation is actually showing.
hwnd = appwin__getActiveChatWindow(hTop)
if hwnd
	return hwnd
endIf
return 0
endFunction

int function schat__isChatWindow(handle hwnd, int restricted)
; Returns True if the given window is part of a Skype conversation window structure.
; Restricted: True if only windows inside a conversation should be allowed to return True.  Useful in Default view.
var
	handle hTop
hTop = schat__getTopOfChat(hwnd)
if !hTop
	return False
elif !restricted
	return True
endIf
while hwnd
	if hwnd == hTop
		return True
	endIf
	hwnd = getParent(hwnd)
endWhile
return False
endFunction

globals
; Cache to minimize Conversation List scans below.
	object Skype__schat___dvobj,
	int Skype__schat___dvid
int function schat___findMatchingCLEntry(string whichList, string wantedName)
var
	object o, int childID, int n
o = appwin__getMSAAObject(whichList)
if !o
	return False
endIf
n = o.accChildCount
childID = 1
while childID <= n
	if !stringCompare(wantedName, o.accName(childID), True)
		Skype__schat___dvobj = o
		Skype__schat___dvid = childID
		return True
	endIf
	childID = childID +1
endWhile
return False
endFunction

object function schat___getConversationObjectAndID(handle hwnd, int byRef childID)
; Get the Conversation List entry (MSAA object and childID) corresponding to the given conversation window.
var
	variant null,
	string wantedName
wantedName = getWindowName(hwnd)
if !stringCompare(wantedName, Skype__schat___dvobj.accName(Skype__schat___dvid), True)
	childID = Skype__schat___dvid
	return Skype__schat___dvobj
endIf
Skype__schat___dvobj = null
Skype__schat___dvid = 0
childID = 0
if schat___findMatchingCLEntry(appwin__SWT_ConversationList, wantedName)
|| schat___findMatchingCLEntry(appwin__SWT_LiveConversations, wantedName)
	childID = Skype__schat___dvid
	return Skype__schat___dvobj
endIf
childID = 0
return null
endFunction

int function schat__isLiveConversation(handle hwnd)
; Returns True if the given call represents a live conversation.
var
	object o, int childID
hwnd = schat__getTopOfChat(hwnd)
if !hwnd
	return False
endIf
o = schat___getConversationObjectAndID(hwnd, childID)
if !o
	return False
endIf
return (o.accName(0) == schat__scLive)
endFunction

string function schat__getChatName(handle hwnd, int includeStatus)
; Returns the name of the conversation indicated by the given window handle.
; Multichats tend to get a message line as a name.
var
	string name,
	int pos,
	object o, int childID, string buf
hwnd = schat__getTopOfChat(hwnd)
if !hwnd
	return ""
endIf
name = getWindowName(hwnd)
pos = stringContainsFromRight(name, ")")
if pos
	name = stringLeft(name, pos)
endIf
if !includeStatus
	return name
endIf
o = schat___getConversationObjectAndID(hwnd, childID)
if !o
	return name
endIf
buf = stringTrimLeadingBlanks(stringSegment(o.accDescription(childID), ",", 1))
if buf
	name = formatString("%1 (%2)", name, buf)
endIf
return name
endFunction

collection function schat___getChatWindows(handle hwnd, int requireOneVisible)
; Helper for schat__getChatWindows.
var
	handle hwndVisible,
	int compactView,
	collection cChats cChats = new collection
compactView = !(getParent(hwnd))
while hwnd
	if (!compactView || isWindowVisible(hwnd)) && schat___isAvailableConversation(hwnd)
		if isWindowVisible(hwnd)
			hwndVisible = hwnd
		endIf
		cChats[intToString(hwnd)] = hwnd
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
if !hwndVisible && requireOneVisible
	; If none are visible (can happen in Default view), there really aren't any available conversations.
	return ""
endIf
return cChats
endFunction

collection function schat__getChatWindows()
; Returns a collection of all the conversation windows currently open.
; The list is guaranteed to be in a stable order from one call to the next; however, this order is not specifically defined.
var
	handle hwnd,
	handle hTop,
	collection cChats
; If the empty canvas is visible, Default view is active and there is no current conversation.
hTop = appwin__getSkypeWindow(appwin__SWT_Main)
if isWindowVisible(findWindow(hTop, "TEmptyCanvas", ""))
	hwnd = findWindow(hTop, "TConversationForm", "")
	if !hwnd
		hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
	endIf
	if hwnd
		cChats = schat___getChatWindows(hwnd, False)
		return cChats
	endIf
endIf
; Try Compact view next because a visible conversation window at top level definitively identifies that this view is active.
cChats = new collection
hwnd = findTopLevelWindow("TConversationForm", "")
if !hwnd
	hwnd = findTopLevelWindow("TConversationForm.UnicodeClass", "")
endIf
if hwnd
	cChats = schat___getChatWindows(hwnd, True)
	if collectionItemCount(cChats) > 0
		return cChats
	endIf
endIf
; Try Default view if Compact view is not active.
cChats = new collection
hwnd = findWindow(hTop, "TConversationForm", "")
if !hwnd
	hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
endIf
if hwnd
	cChats = schat___getChatWindows(hwnd, False)
endIf
return cChats
endFunction

variant function schat__getWindow(string which, handle hTop)
; Returns the requested window from under hTop, which should be the top-level window of a Skype conversation, though it can actually be any window in the conversation.
; If which is "?" the list of understood window names is returned.  This is for testing.
var
	handle hwnd
hwnd = schat__getTopOfChat(hTop)
if !hwnd
	hwnd = appwin__getActiveChatWindow(hTop)
endIf
if !hwnd
	return 0
endIf
hTop = hwnd
if which == "?"
	return schat__WMain +"|" +schat__WInput +"|" +schat__WOutput +"|" +schat__WTopic +"|" +schat__WUserList
	+"|" +schat__WConvToolBar +"|" +schat__wContactArea
elif which == schat__WMain
	return hTop
elif which == schat__WInput
	var handle hwnd1
	hwnd1 = findWindow(hTop, "TChatEntryControl", "")
	if hwnd1
		return findWindow(hwnd1, "TChatRichEdit", "")
		|| findWindow(hwnd1, "TChatRichEdit.UnicodeClass", "")
	endIf
elif which == schat__WOutput
	hwnd = findWindow(hTop, "TChatContentControl", "")
	if hwnd
		return hwnd
	endIf
	hwnd = winfinder__findWindowByClassSubstring(hTop, "mecontrol")
	if !hwnd || !winfinder__findWindowByClassSubstring(getFirstWindow(hwnd), "richview")
		return 0
	endIf
	hwnd = winfinder__findWindowByClassSubstring(getFirstWindow(hwnd), "richview")
	if hwnd
		return hwnd
	endIf
	hwnd = getFirstChild(hTop)
	while hwnd && (getWindowClass(hwnd) != "TPanel" || !findWindow(hwnd, "", "pnlSmileys"))
		hwnd = getNextWindow(hwnd)
	endWhile
	if hwnd
		hwnd = findWindow(hwnd, "TRichView", "")
		return hwnd
	endIf
elif which == schat__WTopic
	return findWindow(findWindow(hTop, "", "pnlTopic"), "TRichView", "")
elif which == schat__WUserList
	return findWindow(hTop, "TChatUserList", "")
elif which == schat__WConvToolBar
	return findWindow(hTop, "TConversationToolbar", "")
elif which == schat__WContactArea
		hwnd = findWindow(hTop, "TConversationHeader", "")
		if !hwnd
			hwnd = findWindow(hTop, "TContactProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hTop, "TConversationProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hTop, "TParticipantList", "")
		endIf
		return hwnd
else
	log__write(formatString(schat__msgUnknownWindow, which), True)
endIf
endFunction

int function schat___isAvailableConversation(handle hwnd)
; Returns True if hwnd is a valid and either active or activatable conversation window.
; "Available" means it has a corresponding Conversation List entry.
; Skype 4 can leave invisible conversation windows around that can't be activated.
var
	int compactView,
	object o, int childID
if getWindowClass(hwnd) != "TConversationForm"
	; Rule out windows with the wrong class.
	return False
endIf
compactView = !getParent(hwnd)
if compactView && !isWindowVisible(hwnd)
	; Rule out invisible windows unless in Default view.
	return False
endIf
if !schat__isChatWindow(hwnd, False)
	return False
endIf
if isWindowVisible(hwnd)
	; By now, it's valid if it's visible.
	; This allows default-view conversations without corresponding Conversation List entries to be valid if visible.
	return True
endIf
if compactView
	return False
endIf
o = schat___getConversationObjectAndID(hwnd, childID)
if o
	return True
endIf
return False
endFunction

int function schat___focusDefaultViewConversation(handle hwnd)
; Focus the given Default view conversation window by clicking its corresponding Conversation list entry.
; Intended for use only in Default view, where conversations all descend from the main Skype window.
var
	handle hwndInbox,
	object o, int childID
; First make sure the Conversation List is visible.
if conv5__applies()
	o = appwin__getMSAAObject(appwin__SWT_ConversationList)
	if o.accState(0) & State_System_Invisible || !o
		typeKey("Alt+2")
		pause()
		stopSpeech()
		o = appwin__getMSAAObject(appwin__SWT_ConversationList)
		if o.accState(0) & State_System_Invisible || !o
			beep()
			return False
		endIf
	endIf
else
	sutil__sayUnsupportedSkype()
	return False
endIf
o = schat___getConversationObjectAndID(hwnd, childID)
if !o
	return False
endIf
; Fire this next to make sure the entry is visible and its text present on screen.
; This will not actually switch focus to the Conversation List window.
o.accSelect(SelFlag_TakeFocus, childID)
pause()
if !MSAA__click(o, childID)
	beep()
	return False
endIf
pause()
hwnd = schat__getWindow(schat__WInput, hwnd)
if getFocus() != hwnd
	stopSpeech()
	setFocus(hwnd)
endIf
return True
endFunction

int function schat__switchToChat(int absolute, int delta)
; Switch forward or backward by delta conversations from the current one, say or show a list of conversations, or say the name of a conversation:
;	absolute == 0 or 1, delta == 0:  Show a list of conversations for selection.
;	absolute == 0, delta != 0:  Move delta (plus or minus) conversations from the current one in round-robin fashion.
;	absolute == 1, delta != 0:  Move to a specific conversation (negative counts from end).
;	absolute == 2, delta == 0:  Say a list of conversations.
;	absolute == 2, delta != 0:  Say the name of a specific conversation (negative counts from end).
; If moving relatively and focus is not currently in a conversation, a delta of 1 will go to the first conversation, -1 to the last, 0 to a list with the first as default, and others logically from this specification.
; Returns True on success and False on failure or user cancellation.
var
	collection cChats, int nchats, int i,
	handle hwnd,
	string chatNames, string name,  ; used when building a list of conversations to show the user
	string desc,
	int curPos

; Collect conversation window handles, a count of them, and the current position among them, and make sure delta is in bounds.
cChats = schat__getChatWindows()
nchats = collectionItemCount(cChats)
if !nchats
	sayMessage(OT_Error, schat__msgNoChatsOpen)
	return False
endIf
var handleArray aChats aChats = new handleArray[nchats]
var string shwnd
i = 0
forEach shwnd in cChats
	i = i +1
	aChats[i] = stringToHandle(shwnd)
endForEach

if absolute == 1 && !delta
	absolute = 0
endIf
if absolute
	curpos = delta
	if !curpos
		sutil__sayURMessage(
			string__quantify(formatString(schat__msgChatsOpen_L, intToString(nchats))),
			string__quantify(formatString(schat__msgChatsOpen_S, intToString(nchats))),
		False)
		forEach shwnd in cChats
			hwnd = stringToHandle(shwnd)
			name = schat__getChatName(hwnd, False)
			string__replace(name, "|", " ", "ect")
			desc = schat__convDescription(hwnd)
			if desc
				name = name +" (" +desc +")"
			endIf
			sutil__sayURMessage(name, "", False)
		endForEach
		return True
	elif curpos < 0
		curpos = nchats +curpos +1
	endIf
	if curpos < 1 || curpos > nchats
		sayMessage(OT_Error, schat__msgNotThatManyChats_L, schat__msgNotThatManyChats_S)
		return False
	endIf
	if absolute == 2
		name = schat__getChatName(aChats[curpos], False)
		string__replace(name, "|", " ", "ect")
		sutil__sayURMessage(name, "", False)
		  return True
	endIf
else  ; not absolute
	var int wantList
	wantList = !delta
	var handle curHwnd = schat__getTopOfChat(getFocus())
	curPos = nchats
	while curPos && aChats[curpos] != curHwnd
		curPos = curPos -1
	endWhile
	if !curPos && delta < 0
		curPos = 1
	endIf
	delta = delta %nchats  ; guarantees reasonable bounds in case somebody does something crazy like delta==125

	; Set curPos to be the conversation we want to go to next.
	if wantList
		; Show list of conversations, starting on current one, and set curPos from the user's choice.
		i = 1
		while i <= nchats
			hwnd = aChats[i]
			name = schat__getChatName(hwnd, False)
			string__replace(name, "|", " ", "ect")
			desc = schat__convDescription(hwnd)
			if desc
				name = name +" (" +desc +")"
			endIf
			chatNames = string__join(chatNames, "|", name)
			i = i +1
		endWhile
		curPos = DLGSelectItemInList(chatNames, schat__msgSelectAChat, False, curPos)
		if !curPos
			sayMessage(OT_Message, schat__msgChatSelectCanceled)
			return False
		endIf
	else
		; Round-robin ring of conversations...
		curPos = curPos +delta
		if curPos < 1
			curPos = curPos +nchats
		elif curPos > nchats
			curPos = curPos -nchats
		endIf
	endIf
endIf  ; absolute

; Switch to the indicated conversation.
hwnd = aChats[curpos]
if !hwnd
	log__write(schat__msgChatSwitchFailed, True)
	log__write("Chat window not found", True)
	return False
endIf
var handle hwndInput
hwndInput = schat__getWindow(schat__WInput, hwnd)
if getFocus() == hwndInput
	if nchats == 1
		sayMessage(OT_Message, schat__msgOnlyChat)
	endIf
	sayObjectTypeAndText()
elif sutil__isAncestorOf(getFocus(), hwnd)
	; SetFocus should work within the current conversation.
	setFocus(hwndInput)
else
	; This is a conversation switch, and setFocus may need a little help.
	; In default view, a lot of help.
	if hwnd == getTopLevelWindow(hwnd)
		; Compact view:  Conversations are separate Windows tasks.
		; We first have to focus and possibly unobscure the conversation's main window.
		if schat___windowIsFullyVisible(hwnd)
			setFocus(hwnd)
		else
			if !schat___focusViaTaskButton(hwnd)
				beep()
				return False
			endIf
		endIf
		pause()
		if !getFocus()
			; The switched-to window is minimized, so unminimize it.
			typeKey("Enter")
		endIf
		if getFocus() != hwndInput
			setFocus(hwndInput)
		endIf
	else
		; Default view:  Conversations are all under the Skype main window.
		; The new one will be invisible, and it can only be made visible by clicking its corresponding Conversation List entry.
		if !schat___focusDefaultViewConversation(hwnd)
			log__write(schat__msgChatSwitchFailed, True)
			log__write("Default-view focus failed.", True)
			return False
		endIf
		if getFocus() != hwndInput
			setFocus(hwndInput)
		endIf
	endIf
endIf
return True
endFunction

int function schat___windowIsFullyVisible(handle hwnd)
; Return True if hwnd is really visible according to MSAA.
; Chats that pop up because of an incoming message tend to return True for isWindowVisible(hwnd)
; but still be invisible in MSAA.
; SetFocus will not work to focus these under JAWS 11.
; This function actually uses a shortcut that may be Skype specific:
; If the window is completely off screen, we assume it's invisible.
if getWindowRight(hwnd) < 0 || getWindowTop(hwnd) < 0
	return False
endIf
return True
endFunction

int function schat___focusViaTaskButton(handle hwnd)
; Attempt to set focus to hwnd by clicking its corresponding taskbar button.
var string name = getWindowName(getTopLevelWindow(hwnd))
hwnd = util__getTaskBarWindow()
if !hwnd
	return False
endIf
if getWindowTypeCode(getFirstChild(hwnd)) == WT_ToolBar
	hwnd = getFirstChild(hwnd)
endIf
var object o, int childID
o = getObjectFromEvent(hwnd, -4, 0, childID)  ; -4 == ObjID_Client
if !o
	return False
endIf
childID = MSAA__childByName(o, name, "c")
if childID
	o.accDoDefaultAction(childID)
	return True
endIf
return False
endFunction

int function schat__isSameText(string s1, string s2)
; Stripped-blank text comparison.
s1 = stringReplaceChars(s1, "\10\13\160", " ")
s2 = stringReplaceChars(s2, "\10\13\160", " ")
return s1 == s2
endFunction

int function schat__getTypist(object o, int i, string byRef typist)
; Returns by reference the typist indicated by the given MSAA element.
; The typist may be empty if the given element does not represent a typist.
; Returns False on encountering a "real" chat message and True until then.
; This function is meant to be called from a loop that travels through chat messages in reverse from the end.
var
	string name, string desc
typist = ""
; ChildID 1 is not allowed to be a typist because either:
;	(1) It's the "Show messages from" entry, or
;	(2) This is a brand-new incoming IM window, which had to be opened by an IM that should be here, or
;	(3) This is a brand-new outgoing IM, in which case you're typing and they're not yet.
; For case (1), the date/time stamp is wrong just as for typing indicators,
; and we use that funny date as a heuristic below (thanks Aaron Smith of GW Micro for this idea),
; which is why it's worth making this exclusion now.
; Non listItems such as buttons and links are excluded for more obvious reasons.
if i <= 1 || o.accRole(i) != Role_System_ListItem
	return False
endIf
var int nState = o.accState(i)
if !(nState & State_System_Unavailable)
	return False
endIf
desc = stringTrimLeadingBlanks(o.accDescription(i))
if desc
	return False
endIf
name = stringTrimLeadingBlanks(o.accName(i))
if !name
	return True
endIf
if stringRight(name, stringLength(schat___scTyping)) != schat___scTyping
	return False
endIf
if nState & State_System_OffScreen
	return True
endIf
name = stringChopRight(name, stringLength(schat___scTyping))
typist = name
return True
endFunction

int function schat___yearDifference(string desc)
; Returns the absolute value of the difference between the current year and the one in desc.
; If there is an error, returns -1.
; Used to help distinguish between typing indicators and regular messages.
; Idea from Aaron Smith of GW Micro, 2009-09-18.
var
	int curYear, int messageYear,
	int diff,
	int pos
curYear = stringToInt(sysGetDate("yyyy"))
if !curYear
	return -1
endIf
pos = stringContainsFromRight(desc, " ")
if !pos
	return -1
endIf
messageYear = stringToInt(stringChopLeft(desc, pos))
if messageYear < 2000
	return -1
endIf
diff = messageYear -curYear
if diff < 0
	diff = 0 -diff
endIf
return diff
endFunction

int function schat__selectMessage(object o, int msgno)
; Select the indicated message.
; msgno: 1-based (positive) message number, or negative number indicating a message relative to the last one.
; -1 is the last message, -2 the next to last, etc.
; 0 is not a valid value of msgno.
var int n = schat__getMessageCount(o)
if msgno < 0
	msgno = n +1 +msgno
endIf
o.accSelect(SelFlag_TakeSelection, msgno)
	return True
endFunction

int function schat___selectedMessageID(object o)
; Indicate the MSAA ChildID of the focused or, failing that, selected chat message.
var int childID = o.accFocus +0
if stringToInt(intToString(childID))
	return childID
endIf
childID = o.accSelection +0
if stringToInt(intToString(childID))
	return childID
endIf
childID = 1
var int n = o.accChildCount
while childID <= n
	if (o.accState(childID) & (State_System_Selected|State_System_Focused))
	&& o.accRole(childID) == Role_System_ListItem
		return childID
	endIf
	childID = childID +1
endWhile
; Default result if nothing is selected.
return 1
endFunction

int function schat__selectedMessage(object o)
; Indicate the 1-based message number of the selected chat message.
return schat___selectedMessageID(o)
endFunction

int function schat__focusedMessageID(object o)
; Indicate the number of the focused chat message.
; If MSAA doesn't indicate which message is focused, the last "real" message is returned.
; "Real" means not a typist.
var int childID = o.accFocus +0
if childID
	return childID
endIf
if o.accRole(2) != Role_System_ListItem
	return 1
endIf
return schat__getMessageCount(o)
endFunction

int function schat__getMessageCount(object o)
; Return the number of "real" messages in the given chat.
; "Real" means not typist items.
var
	int n,
	string typist
n = o.accChildCount
while n
	if !schat__getTypist(o, n, typist)
		return n
	endIf
	n = n -1
endWhile
return 0
endFunction

int function schat__getMessageIndex(object o)
; Given the MSAA object for a single message, return its 1-based index among messages.
; Returns 0 on error or failure.
var object oList = o.accParent
if !oList
	return 0
endIf
var int n = schat__getMessageCount(oList)
if !n
	return 0
endIf
; Bounding rectangle for the given message.
var int ml, int mr, int mt, int mb
if !MSAA__rect(o, 0, ml, mr, mt, mb)
	return 0
endIf

; This is a binary search based on bounding rectangle coordinates.
; Coordinates increase monotonically but can start negative.
; [DGL, 2014-09-14, Skype 6.18.59.106]
var int lbound = 1, int ubound = n, int i = 1
; Message to test against and its bounding rectangle.
var object oTest, int tl, int tr, int tt, int tb
var int safety = 100
while safety && ubound >= lbound
	oTest = oList.accChild(i)
	if !MSAA__rect(oTest, 0, tl, tr, tt, tb)
		return 0
	endIf
	if tt > mt
		; This message is below, hence after, the given one.
		ubound = i -1
		i = (lbound +i) /2
	elif tt < mt
		; This message is above, hence before, the given one.
		lbound = i +1
		i = (ubound +i) /2
	elif tl == ml && tr == mr && tb == mb
		; We have a winner!
		;sayInteger(100 -safety +1)
		return i
	endIf
	if i < lbound
		; Can happen due to integer arithmetic.
		i = lbound
	endIf
	safety = safety -1
endWhile
; The safety net ran out.
return 0
endFunction

collection function schat___parseMessage(object oMsgs, int childID)
; Get and return a collection breaking down the indicated message into component parts.
; This function is definitive on how to interpret Skype chat MSAA messages.
; Returned fields:
;	raw: The raw message text from MSAA.
;	sender, time, text: Parts of the message if the message is parsable.
;	message: The actual message (includes sender and text).
;		For unparseable messages, this is the same as raw.
;	meta: Meta info about the message, from its end, if parsable.
;		message + meta == raw.*
;	flags: edited, new, not delivered, etc.; drawn from meta.
;		Currently understood special messages: Edited, New, Not delivered.
;		For unparsable messages, this will be blank.
;		NOTE: SMS statuses sent and delivered are not handled specially here.
;	extra: Extra material drawn from elements under the message, if any.
;		This occurs for special "messages" like file transfers.
;	msgno, msgcount: 1-based index of message and message count.
; Assumes a message's accName property takes the approximate form
; from <sender>: <message>, sent at <date/time>[ edited by <editor> at <date/time>][ not delivered|new]
; This is current as of Skype 6.18 for most message types. [DGL, 2014-08-19]
; * info.message may have schat___scFrom removed for convenience in reading messages,
; and messages starting with sender name (/me lines and contact sharing)
; have the duplicate text removed; so message+meta may not always be equal to raw.
var
	collection info,
	string buf,
	string flags,
	int pos

; raw, extra, msgno and msgcount can be set right away.
; Set the rest to blank until filled in later.
buf = oMsgs.accName(childID)
info = new collection
info.raw = buf
info.sender = ""
info.time = ""
info.text = ""
info.message = ""
info.meta = ""
info.flags = ""
info.extra = schat___getExtraMessageMaterial(oMsgs, childID)
info.msgno = childID
info.msgcount = schat__getMessageCount(oMsgs)

; Identify "normal" message types.
if !buf || !stringStartsWith(buf, schat___scFrom)
	; We don't know how to parse these or there's nothing there.
	info.message = buf
	return info
endIf
; This just makes reading messages faster, and users have requested it.
buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(schat___scFrom)))

; Split meta from message.
var string sep = schat___scSent
pos = stringContainsFromRight(buf, sep)
if !pos
	; Special case for file transfers.
	sep = schat___scOn
	pos = stringContainsFromRight(buf, sep)
	if !pos
		info.message = buf
		return info
	endIf
endIf
info.message = stringLeft(buf, pos-1)
; This leaves sep in meta as well.
; This means info.message +info.meta == info.raw.
; Well almost; see above removal of schat___scFrom...
buf = stringChopLeft(buf, pos-1)
info.meta = buf

; Go back and get sender and text from message.
pos = stringContains(info.message, ", ")
if pos
	info.sender = stringLeft(info.message, pos-1)
	info.text = stringChopLeft(info.message, pos+1)
	; Extra simplification for contact detail sharing and /me lines:
	; If the sender name heads the message, the duplicate is cut.
	if stringStartsWith(info.text, formatString("%1 ", info.sender))
		info.sender = ""
	endIf
endIf

; Now to pull flags from the metadata, which remains in buf.
; Complex metadata example:
;	Wednesday, July 01, 2009 1:08:44 PM Edited by Doug Lee at Wednesday, July 01, 2009 1:09:17 PM Not delivered
; The initial time is from "sent on."
; I believe date and time values are formatted per locale here.

flags = ""
pos = stringContainsFromRight(buf, schat___scEditedBy)
if pos
	; Looks like this message has been edited.
	flags = flags +schat___msgEdited
	info.time = stringLeft(buf, pos-1)
endIf
if stringRight(buf, stringLength(schat__scNew)) == schat__scNew && childID != 1
	; childID != 1: Skype 4.1 (at least) likes to put "New" on the line with the links for showing history.
	flags = flags +schat__scNew
	buf = stringChopRight(buf, stringLength(schat__scNew))
endIf
if stringRight(buf, stringLength(schat___scNotDelivered)) == schat___scNotDelivered
	flags = flags +schat___scNotDelivered
	buf = stringChopRight(buf, stringLength(schat___scNotDelivered))
endIf
; Any unrecognized flags will end up in info.time.
info.time = buf
flags = stringTrimLeadingBlanks(flags)
info.flags = flags
return info
endFunction

string function schat___getExtraMessageMaterial(object o, int childID, optional int textOnly)
; Get and return any special content drawn from descendants of the indicated message.
; This applies to voicemails, file transfers, contact requests, etc.
; Currently also used to get user profile (conversation header) text in a conversation panel.
; o/childID may be the message list and a childID or the message object and 0.
; If textOnly is True, buttons and links and such are ignored.
var int isFirst = (childID == 1)
if childID
	o = o.accChild(childID)
	childID = 0
endIf
var int n = o.accChildCount
if !n
	return ""
endIf

; ToDo: Slow but functional depth-first search without recursion.
var string buf = ""
var collection stack = schat__stl__stack(20)
schat__stl__push(stack, o)
schat__stl__push(stack, 0)
while schat__stl__size(stack)
	childID = schat__stl__pop(stack).value +0
	o = schat__stl__pop(stack).value
	childID = childID +1
	if childID <= o.accChildCount
		var object o1 = o.accChild(childID)
		if o1
			schat__stl__push(stack, o)
			schat__stl__push(stack, childID)
			o = o1
			childID = 0
		endIf
		var int iRole = o.accRole(childID)
		if (isFirst || iRole != Role_System_Link)
		&& !(textOnly && iRole != Role_System_StaticText)
			var string tmp = o.accName(childID)
			if !stringIsBlank(tmp)
				buf = buf +" " +tmp
			endIf
			tmp = o.accDescription(childID)
			if !stringIsBlank(tmp)
				buf = buf +" " +tmp
			endIf
			tmp = o.accHelp(childID)
			if !stringIsBlank(tmp)
				buf = buf +" " +tmp
			endIf
		endIf  ; not a link
		schat__stl__push(stack, o)
		schat__stl__push(stack, childID)
	endIf
endWhile
return stringTrimLeadingBlanks(buf)
endFunction

string function schat__getMessageText(object oMsgs, int msgno, string flags)
; Get the speakable/Braillable text of the indicated message.
; flags:
;	n for new (incoming) message / no special info like "not delivered."
;	a: Called as part of a SayAll operation.
;	Any other flags are silently ignored.
var
	int childID,
	string specialInfo, int noSpecialInfo,
	string msg
if !oMsgs
	return ""
elif msgno <= 0
	return ""
elif msgno > schat__getMessageCount(oMsgs)
	return ""
endIf
var collection info = schat___parseMessage(oMsgs, msgno)
msg = info.text
if info.sender
	if stringContains(stringLower(flags), "a")
		msg = formatString("%1 %2", info.sender, msg)
	else
		msg = formatString("%1:\n%2", info.sender, msg)
	endIf
endIf
if stringIsBlank(msg)
	msg = info.message
endIf
if info.extra
	msg = formatString("%1\n%2", msg, info.extra)
endIf
if opt__getUserOption(opt__SV_ChatMessageTimes) == 1
	; Using message+meta allows extra to be put between them.
	var string meta = info.meta
	if stringContains(",:", stringLeft(meta, 1))
		meta = stringTrimLeadingBlanks(stringChopLeft(meta, 1))
	endIf
	msg = formatString("%1\n%2", msg, meta)
endIf
flags = stringLower(flags)
noSpecialInfo = (stringContains(flags, "n"))
if !noSpecialInfo
	specialInfo = info.flags
endIf
if specialInfo && !noSpecialInfo
	msg = specialInfo +", " +msg
endIf
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(msg))
endFunction

globals
; The set of messages being used by an active userBuffer of messages,
; the index of the current message among them,
; and the index of the wanted link within that.
	object Skype__schat___oMsgs, int Skype__schat___msgno, int Skype__schat___linkno

int function schat___clickItem(int msgno, int linkno)
; Clicks a link when the user activates it from a JAWS user buffer or its link list.
; Part of the support code for schat__virtualizeChatMessages().
if userBufferIsActive()
	userBufferDeactivate()
endIf
if msgno
	;sayString(formatString("Click message %1 item %2", intToString(msgno), intToString(linkno)))
	Skype__schat___msgno = msgno
	Skype__schat___linkno = linkno
	scheduleFunction("schat___clickItem", 0)
	return True
endIf

stopSpeech()
msgno = Skype__schat___msgno
linkno = Skype__schat___linkno
Skype__schat___msgno = 0
Skype__schat___linkno = 0

var object oMsg = Skype__schat___oMsgs.accChild(msgno)
if !oMsg
	beep()
	return False
endIf
return oMsg.accDoDefaultAction(linkno)
endFunction

int function schat___addMessageToUserBuffer(handle hwnd, object byRef oMsgs, int msgno)
; Adds the indicated message to a JAWS virtual view of conversation messages.
; Links in the message become links in the JAWS view as well.
; Part of the support code for schat__virtualizeChatMessages().
if !oMsgs
	if conv5__applies()
		oMsgs = conv5__chatContent(hwnd)
	else
		sutil__sayUnsupportedSkype()
		return False
	endIf
endIf
if oMsgs.accRole(0) != Role_System_List && oMsgs.accRole(0) != Role_System_Pane
return False
endIf
var string msg = schat__getMessageText(oMsgs, msgno, "")
var object oMsg = oMsgs.accChild(msgno)
if !oMsg
	oMsg = oMsgs
endIf
if !oMsg.accChildCount
	userBufferAddText(msg)
	return
endIf

; This message includes links and/or buttons and/or special text parts.
; To avoid duplication of text, we now have the slippery task of
; finding, in msg, the right occurrence of each link/button's text,
; so we can wrap userBuffer action support around it.
; The plan is to split msg into parts around each actionable item.
var int i, int iRole, string name, int pos, int wt
var int n = oMsg.accChildCount
var int checkForColon, string punct
for i = 1 to n
	iRole = oMsg.accRole(i)
	if iRole == Role_System_Link || iRole == Role_System_PushButton
		name = oMsg.accName(i)
		pos = 0
		checkForColon = False
		punct = ""
		if iRole == Role_System_Link
			pos = stringContains(msg, name)
			checkForColon = True
		elif iRole == Role_System_PushButton
		; Buttons don't split message text; they occur after it;
		; but we added extra stuff, including button names, to msg already.
		; We therefore scan from the right for these.
			pos = stringContainsFromRight(msg, name)
		endIf
		if pos
			; Add msg text up to but not including the actionable item.
			var string lhs = stringTrimTrailingBlanks(stringLeft(msg, pos-1))
			if lhs
				userBufferAddText(lhs)
			endIf
			; Then just keep whatever comes after the actionable item.
			msg = stringTrimLeadingBlanks(stringChopLeft(msg, pos +stringLength(name) -1))
			if checkForColon && stringStartsWith(msg, ":")
				msg = stringTrimLeadingBlanks(stringChopLeft(msg, 1))
				punct = ":"
			endIf
		elif msg
			; Add and clear all message text.
			userBufferAddText(msg)
			msg = ""
		endIf
		; Add the actionable item itself with its text.
		if iRole == Role_System_PushButton
			wt = WT_Button
		else
			wt = WT_Link
		endIf
		userBufferAddText(name +punct,
		"schat___clickItem(" +intToString(msgno) +", " +intToString(i) +")",
		; Include message number in links so, e.g., "Play" for five different
		; voicemails in one userBuffer will be unique in the Insert+F7 list.
		formatString(schat___msgMessageLinkInList, name, intToString(msgno)),
		"", 0, 0, IgnoreColor, IgnoreColor, True,
		wt, 0
		)
	endIf
endFor
if msg
	userBufferAddText(msg)
endIf
return True
endFunction

int function schat__virtualizeChatMessages(handle hwnd, int start, int contextSize, int contextPos)
; Virtualize contextSize+1 messages containing start.
; Negative start counts from the end (-1 is the last message).
; contextPos determines where start appears in the messages shown:
;	< 0:  Last message.
;	= 0:  Center message (contextSize/2 messages above and below).
;	> 0:  First message.
var
	object oMsgs, int offset,
	int n, int nmsgs,
	int idx, int idx1, int idx2,
	string title
hwnd = schat__getWindow(schat__WOutput, hwnd)
if !hwnd
	sayMessage(OT_Error, schat__msgNoChatOutputWindow)
	return False
endIf
if conv5__applies()
	oMsgs = conv5__chatContent(hwnd)
else
	sutil__sayUnsupportedSkype()
	return False
endIf
if !oMsgs
	log__write(schat__msgNoMSAA, True)
	return False
endIf

; Set start, idx1, and idx2.
; The shown message range is from idx1 through idx2 inclusive.
; Start is the message where the cursor begins and must be between idx1 and idx2 inclusive.
n = contextSize +1
nmsgs = schat__getMessageCount(oMsgs)
if start < 0
	start = nmsgs +1 +start
endIf
start = min(nmsgs, max(1, start))
if contextPos > 0
	idx1 = start
elif contextPos == 0
	idx1 = start - (contextSize /2)
else  ; contextPos < 0
	idx1 = start -contextSize
endIf
idx1 = min(nmsgs, max(1, idx1))
idx2 = idx1 +contextSize
;sayString(formatString("%1 %2 %3", intToString(idx1), intToString(start), intToString(idx2)))
offset = idx2 -nmsgs
if offset > 0
	idx1 = idx1 -offset
	idx2 = idx2 -offset
	idx1 = max(1, idx1)
endIf
Skype__schat___oMsgs = oMsgs
if userBufferIsActive()
	userBufferDeactivate()
endIf
userBufferClear()
var int pos
title = schat__getChatName(hwnd, False)
pos = stringContainsFromRight(title, "(")
if pos
	; Take off what is probably the conversation partner's online status.
	title = stringTrimTrailingBlanks(stringLeft(title, pos-1))
endIf
title = formatString(schat__msgVChatTitle,
		intToString(idx1), intToString(idx2), intToString(nmsgs), title)
userBufferAddText(title)
offset = 0
idx = idx1
userBufferActivate()
while idx <= idx2
	if idx == start
		; +3: 1 to go from end of last message to start of this one,
		; 1 to skip the blank line atop this one,
		; 1 more to skip the origin and go to the message itself.
		offset = getLineCount() +3
	endIf
	userBufferAddText(" ")
	schat___addMessageToUserBuffer(hwnd, oMsgs, idx)
	idx = idx +1
endWhile
userBufferAddText(" ")
var string buf, int msgcount
if idx1 > 1
	msgcount = min(n-1, idx1-1)
	buf = formatString(schat__msgVChatPrev, intToString(msgcount))
	userBufferAddText(buf,
			formatString("schat___deferredVirtualizeChatMessages(%1, %2, %3, -1)",
				intToString(hwnd), intToString(idx1), intToString(contextSize)),
			buf)
endIf
if idx2 < nmsgs
	msgcount = min(n-1, nmsgs-idx2+1)
	buf = formatString(schat__msgVChatNext, intToString(msgcount))
	userBufferAddText(buf,
			formatString("schat___deferredVirtualizeChatMessages(%1, %2, %3, 1)",
				intToString(hwnd), intToString(idx2), intToString(contextSize)),
			buf)
endIf
userBufferAddText(cmsgBuffExit)
if offset
	goToLineNumber(offset)
endIf
if stringIsBlank(getLine())
	priorLine()
endIf
sayLine()
sayUsingVoice(VCTX_Message, title, OT_Help)
return True
endFunction

globals
	handle Skype__schat___vcHwnd,
	int Skype__schat___vcStart,
	int Skype__schat___vcContextSize,
	int Skype__schat___vcContextPos
int function schat___deferredVirtualizeChatMessages(handle hwnd, int start, int contextSize, int contextPos)
; Used by schat__virtualizeChatMessages when creating a new buffer from a Previous or Next link.
; This avoids the problem of secondary buffers just saying "F7" instead of bringing up a link list on JAWSKey+F7.
; [DGL, 2008-05-09]
if hwnd
	Skype__schat___vcHwnd = hwnd
	Skype__schat___vcStart = start
	Skype__schat___vcContextSize = contextSize
	Skype__schat___vcContextPos = contextPos
endIf
speechOn()
schat__virtualizeChatMessages(Skype__schat___vcHwnd, Skype__schat___vcStart, Skype__schat___vcContextSize, Skype__schat___vcContextPos)
endFunction

collection function schat__convInfo(handle hwnd)
; Return a collection whose properties describe the given conversation window.
; Properties returned (not all for every conversation):
;	window: The window handle to the conversation.
;	callActive: True if this conversation is (or contains) an active (connected) call.
;	multi: True if this conversation has multiple participants.
;	callHeld: Blank, "l" for local hold, or "r" for remote hold.
;	callMuted: True if muted (can be True on hold).
;	videoOut: y, n, or blank, for yes, no, or apparently not available.
;	videoIn: Same as for videoOut for incoming video signal.
;	callQuality: Call quality description if present.
;	callDuration: Call duration if valid, in [hh:]mm:ss format.
; All schat___convInfo* functions called by this one must know the above properties.
var collection null null = new collection
var collection oConv oConv = new collection
if conv5__applies()
	schat__chat5__getInfo(oConv, hwnd)
else
	sutil__sayUnsupportedSkype()
	return null
endIf
return oConv
endFunction

string function schat__convDescription(variant which)
; Return a brief description of the conversation indicated by the indicated conversation.
; which: Window handle of conversation, or collection from a schat__convInfo() call on that conversation.
var
	collection oConv,
	string sMulti, string sCall, string sMuted, string sHeld,
	string duration,
	string buf
if intToString(which) != 0
	oConv = schat__convInfo(which)
else
	oConv = util__vcast(which)
endIf
buf = ""
sMulti = ""
if oConv.multi
	sMulti = " " +schat__msgMulti
endIf

sCall = ""
if oConv.callActive
	if oConv.videoIn == "y"
		if oConv.videoOut == "y"
			sCall = " " +schat__msgVideoCall
		else  ; Both disabled and unsupported outgoing video cases.
			sCall = " " +schat__msgVideoInCall
		endIf
	else  ; No incoming video.
		if oConv.videoOut == "y"
			sCall = " " +schat__msgVideoOutCall
		elif oConv.videoOut == "n"
			sCall = " " +schat__msgVoiceCall
		else
			sCall = " " +schat__msgCall
		endIf
	endIf
	duration = oConv.callDuration
	if duration
		sCall = sCall +" " +duration
	endIf
else
	; Nothing else to say...
	return stringTrimLeadingBlanks(sMulti)
endIf

sMuted = ""
if oConv.callMuted
	sMuted = " " +schat__msgMuted
endIf

sHeld = ""
if oConv.callHeld == "l"
	sHeld = " " +schat__msgLocalHold
elif oConv.callHeld == "r"
	sHeld = " " +schat__msgRemoteHold
endIf

buf = formatString(schat__msgConvDescFormat, sMulti, sCall, sMuted, sHeld)
return stringTrimLeadingBlanks(buf)
endFunction


;------------------------ Module chat5 (from SkypeChat5) -----------------------
; Code for filling the conversation info structure from a Skype 5.x conversation window tree.
; Supports Skype 5.0 beta 2 and later.
; Does not support Skype 3.x, 4.x, or 5.0 beta 1.
;
; Author:  Doug Lee

void function schat__chat5__getInfo(collection byRef oConv, handle hwnd)
; Fill the given conversation info collection from the given Skype 5.0 (beta 2 or later) conversation window.
; oConv:  An empty conversation info collection; see schat__convInfo() for details..  Must be empty on call.
; hwnd: A conversation window or a member of one.
var
	handle hwndConvToolBar,
	handle hwndContactArea,
	object o, int childID,
	string buf

; First check for a Skype 5.1/5.2 live call area.
var collection oLiveInfo = live5__getInfo(hwnd)
if !util__collectionIsEmpty(oLiveInfo)
	oConv.window = hwnd
	oConv.callActive = True
	var collection oParticipants = oLiveInfo.participants
	oConv.multi = (collectionItemCount(oParticipants) > 1)
	oConv.callHeld = oLiveInfo.held
	oConv.callMuted = oLiveInfo.muted
	oConv.videoOut = oLiveInfo.videoOut
	oConv.videoIn = oLiveInfo.videoIn
	oConv.callQuality = oLiveInfo.callQuality
	oConv.callDuration = oLiveInfo.duration
	return
endIf
hwndConvToolBar = schat__getWindow(schat__wConvToolBar, hwnd)
if !hwndConvToolBar
	return
endIf
o = getObjectFromEvent(hwndConvToolBar, ObjID_Client, 0, childID)
if !o
	return
endIf
oConv.window = hwnd

; Does this conversation contain multiple paerticipants?
hwndContactArea = schat__getWindow(schat__WContactArea, hwnd)
if hwndContactArea
	if MSAA__windowProp(hwndContactArea, "Name") == schat__scParticipants
		oConv.multi = True
	else
		oConv.multi = False
	endIf
endIf

; Does this conversation contain or consist of a call?
if MSAA__childByName(o, schat__scEndCall)
	oConv.callActive = True
else
	oConv.callActive = False
	; The rest of the properties only apply to calls, so we're done.
	return
endIf

; Hold status/type.
if MSAA__childByName(o, schat__scResume)
	; This is a button.
	oConv.callHeld = "l"
elif MSAA__childByName(o, schat__scRemoteHold)
	; This has a role of Toolbar.
	oConv.callHeld = "r"
endIf

; Mute status.
if MSAA__childByName(o, schat__scUnmute)
	oConv.callMuted = True
else
	oConv.callMuted = False
endIf

; Outgoing video status.
if MSAA__childByName(o, schat__scStopVideo)
	oConv.videoOut = "y"
elif MSAA__childByName(o, schat__scStartVideo)
	oConv.videoOut = "n"
endIf
; TODO: Receiving part not done.

; Call quality if present.
; This is apparently blank unless the call quality is suboptimal.
childID = MSAA__childByName(o, schat__scCallQuality, "p")
if childID
	buf = o.accName(childID)
	buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(schat__scCallQuality)))
	if buf
		oConv.callQuality = buf
	endIf
endIf

; Call duration if valid.
if conv5__applies()
endIf
return
endFunction

;---------------------- End module chat5 (from SkypeChat5) ---------------------


;---------------------------------- Module stl ---------------------------------
; Standard data structures for JAWS.
;
; Author: Doug Lee of Level Access

const
; List types supported.
	schat__stl___LstStack = 1,
	schat__stl___LstQueue = 2

collection function schat__stl__stack(optional int maxSize)
; Return a stack for use via schat__stl__push/schat__stl__pop etc.
return schat__stl___new(schat__stl___LstStack, maxSize)
endFunction

collection function schat__stl__queue(optional int maxSize)
; Return a queue for use via schat__stl__push/schat__stl__pop etc.
return schat__stl___new(schat__stl___LstQueue, maxSize)
endFunction

collection function schat__stl___new(int ltype, optional int maxSize)
; Return a structure of the indicated type.
var collection lst lst = new collection
lst.type = ltype
if ltype == schat__stl___LstQueue
	; ToDo: maxSize ignored.
	lst.size = 0
	return lst
endIf

lst.maxSize = 0
lst.top = 0
if maxSize
	; More efficient.
	lst.maxSize = maxSize
	var variantArray ar ar = new variantArray[lst.maxSize]
	lst.content = ar
else
	; A bit less efficient but more flexible.
	var collection nc nc = new collection
	lst.content = nc
endIf
return lst
endFunction

int function schat__stl__push(collection struct, variant item)
; Push an item onto whatever structure is passed.
; Returns True on success and False on failure.
if struct.type == schat__stl___LstQueue
	var collection node node = new collection
	node.value = item
	var collection node0 = struct.tail
	if node0
		node0.next = node
		struct.tail = node
	else
		struct.head = node
		struct.tail = node
	endIf
	struct.size = struct.size +1
	return True
endIf

struct.top = struct.top +1
if struct.maxSize
	if struct.top > struct.maxSize
		struct.top = struct.top -1
		return False
	endIf
	var variantArray ar = struct.content
	ar[struct.top] = item
	;sayString("Push " +intToString(struct.top))
else
	var collection col = struct.content
	col[intToString(struct.top)] = item
	;sayString("Dynamic push " +intToString(struct.top))
endIf
return True
endFunction

collection function schat__stl__top(collection struct, optional int alsoPop)
; Return the top element of whatever structure is passed.
; Also pops the element off if alsoPop is True.
; The item is returned as the .value property of a simple collection.
; On error, null is returned rather than a collection.
var collection item item = new collection
if struct.type == schat__stl___LstQueue
	var collection node = struct.head
	if !node
		return
	elif alsoPop
		struct.head = node.next
		if !struct.head
			struct.tail = struct.head
		endIf
		struct.size = struct.size -1
	endIf
	return node
endIf

if struct.top < 1
	return
endIf
if struct.type == schat__stl___LstStack
	if struct.maxSize
		var variantArray ar = struct.content
		item.value = ar[struct.top]
	else
		var collection col = struct.content
		item.value = col[intToString(struct.top)]
	endIf
	if alsoPop
		;sayString("Pop " +intToString(struct.top))
		struct.top = struct.top -1
	endIf
	return item
endIf
return
endFunction

collection function schat__stl__pop(collection struct)
; Pop an item from whatever structure is passed.
; The item comes back as the .value property of a collection.
; On error, returns null instead of a collection.
return schat__stl__top(struct, True)
endFunction

int function schat__stl__size(collection struct)
; Return the size (item count) of whatever structure is passed.
if struct.type == schat__stl___LstQueue
	return struct.size
endIf
return struct.top
endFunction

;-------------------------------- End module stl -------------------------------

;====================== End module schat (from SkypeChat) ======================


;================================== Module cre =================================
; Code for fixing the following issue in Skype 6.9+ TChatRichEdit input boxes:
; On just the first message typed into a particular chat window,
; if enhanced edit support is used (default), no text is seen.
; In this case, using the OSM is better, until the message is sent.
;
; Author: Doug Lee

/*
Notes:
In Default view, where chats stay in the main Skype window, chat input
edit controls often disappear and reappear with new window handles. On
the first appearance of any window handle in  Default view, JAWS 15
(through at least 15.0.6025) seems unable to retrieve the
iTextDocument interface internally and/or use it effectively to
implement its enhanced edit mode. When this happens, users are unable
to see any text in the window, locate text to edit, hear anything but
"blank" on arrows and backspace, etc. Using the OSM, i.e., avoiding
enhanced edit mode, works better in this case. Alt+Tabbing out of and
back into Skype, forcing focus changes away from and back to the edit
window, and even reloading JAWS do not fix this problem. However, as
soon as one message is sent from the window, things begin to work
properly in enhanced edit mode (assuming it's still active).

The above problem can be detected programmatically by checking for a
non-blank MSAA accValue for the edit control and a blank
getObjectValue() result for it. These two conditions occur
simultaneously only when JAWS is unable to access the control
correctly.

This module deactivates enhanced edit mode when the problem is
detected. Window handles of chat input controls are remembered both to
minimize time required to grab IAccessible objects for them and to
allow tracking of which windows have fixed themselves. Enhanced mode
is enabled again when either the accValue is again empty or a
correctly-working chat window takes focus when enhanced mode is not in
effect.
*/

globals
; The current edit window, 0 if the focused window is not one.
	handle Skype__cre___curWindow,
; The per-window data for this module. Fields:
;	hwnd: The window handle.
;	obj: The MSAA Client object for the window.
;	fixed: True if enhanced edit mode has been seen to work in this window.
	collection Skype__cre___hwnds

int function cre__manage()
; Manage the Enhanced Edit Support usage for TChatRichEdit windows.
if getJFWVersion() >= 1600000
	; In JAWS 16, getObjectValue() returns null even when things are working here.
	; Also clipboard functions are broken there by this code.
	; So we do nothing in 16+.
	return False
endIf
var handle hwnd = getFocus()
cre___updateCurrentWindow(hwnd)
var int opt = Opt_Edit_Use_OSM

; If the MSAA mode is above 1, the below system doesn't work.
if getJCFOption(OPT_MSAA_Mode) > 1
	return False
endIf

if getWindowClass(hwnd) != "TChatRichEdit"
	return False
endIf
var collection rec = cre__winrecs__record(Skype__cre___hwnds, hwnd)
if !rec.hwnd
	rec.obj = MSAA__objectFromWindow(hwnd)
	if !rec.obj
		return False
	endIf
	rec.fixed = False
	rec.hwnd = hwnd
endIf

; ToDo: Undocumented JAWS function.
var int enhanced = usingEnhancedEditSupport(hwnd)
if rec.fixed
	; Once fixed, always fixed until the window goes away.
	if !enhanced
		setJCFOption(opt, 0)
	endIf
	return False
endIf

; If the actual value is blank, we can't tell if we have a problem yet.
var object o = rec.obj
if !o.accValue(0)
	; But we can make sure to use enhanced mode in case it starts working.
	if !enhanced
		setJCFOption(opt, 0)
	endIf
	return False
endIf

if !enhanced
	; Apparently already compensating for this issue.
	return False
endIf

; If getObjectValue is not blank, things should be working normally.
; False means do not use MSAA here.
if getObjectValue(False)
	rec.fixed = True
	return False
endIf

; Problem: objectValue blank, MSAA value not blank.
; Meaning: JAWS can't read the edit control in enhanced mode.
; Solution: Quit using enhanced edit support.
setJCFOption(opt, 1)
return True
endFunction

void function cre___updateCurrentWindow(handle hwnd)
; Update data associated with the current edit window.
if hwnd == Skype__cre___curWindow
	return
endIf
; ToDo: Update any required info.
Skype__cre___curWindow = util__vcast(0)
endFunction


;-------------------------------- Module winrecs -------------------------------
; A manager for data structures (records) indexed by a window handle.
;
; Author:  Doug Lee of Level Access

collection function cre__winrecs__record(collection byRef recs, handle hwnd, optional int noCreate)
; Return the record for the given window.
; Registers this window in the set if necessary.
; Also removes stale records for windows that no longer exist.
; The caller must handle record structure creation.
; Suggestion: Use an "hwnd" property so there's a uniform way to test for a new record.
; Pass noCreate if you need to avoid creating a new record.
cre__winrecs__clean(recs)
if !hwnd
|| !getWindowStyleBits(hwnd)
	return
endIf
var string shwnd = intToString(hwnd)
var collection rec
if !recs
	recs = new collection
endIf
if !collectionItemExists(recs, shwnd)
	if noCreate
		return
	endIf
	; We have to make a new record for this one.
	rec = new collection
	recs[shwnd] = rec
	; Caller must detect that this record is empty and build it.
else
	rec = recs[shwnd]
endIf
return rec
endFunction

int function cre__winrecs__remove(collection recs, handle hwnd)
; Remove a window's record.
var string shwnd = intToString(hwnd)
if !collectionItemExists(recs, shwnd)
	return False
endIf
collectionRemoveItem(recs, shwnd)
return True
endFunction

int function cre__winrecs__clean(collection recs, /*?*/ int clearAll)
; Remove from recs any records for windows that no longer exist.
; If clearAll is True, removes all records.
; Returns the number of records removed.
; Note that this function is called by cre__winrecs__record(),
; so manual use of it is optional.
var string shwnd
var int nremoved = 0
if clearAll
	var int n = 0
	forEach shwnd in recs
		n = n +1
	endForEach
	collectionRemoveAll(recs)
	return n
endIf
var handle hwnd
forEach shwnd in recs
	hwnd = stringToHandle(shwnd)
	if !hwnd || !getWindowStyleBits(hwnd)
		if cre__winrecs__remove(recs, hwnd)
			nremoved = nremoved +1
		endIf
	endIf
endForEach
return nremoved
endFunction

;------------------------------ End module winrecs -----------------------------

;================================ End module cre ===============================


;================================== Module opt =================================
; User option code.

const
; A constant used to detect missing, as opposed to set-to-empty, .ini keys.
	opt___NoVal = 12345678

variant function opt__getUserOption(string varName)
; Retrieves and returns the value of the indicated user-settable variable from skype.ini.
; This function makes sure to return integer 0 for the integer value 0, rather than returning the string "0" which shows up as True in a boolean test.
var
	string val
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
val = iniReadString(sutil__INISectName, varname, intToString(opt___NoVal), sutil__INIFileName)
; Favor integer 0 (False) rather than string "0" (True).
if stringContains(stringLeft(stringTrimLeadingBlanks(val), 1), "0")
	return ""
endIf
if val == opt___NoVal
	; All default values for options go here.
	if varname == opt__SV_RecognizeChat
		val = "1"
	elif varname == opt__SV_ChatMessageTimes
		val = "1"
	elif varname == opt__SV_SpeakOnlineStatus
		val = "1"
	elif varname == opt__SV_TypingAnnounce
		val = "1"
	elif varname == opt__SV_TypingSounder
		val = "1"
	elif varname == opt__SV_SkypeCommandKey
		val = "$"
	elif varname == opt__SV_SkypeWatchMessages
		val = "1"
	else
		val = ""
	endIf
endIf
return val
endFunction

int function opt__setSkypeVar(string varName, variant val)
; Sets the indicated user-settable variable in skype.ini to the given value.
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
return iniWriteString(sutil__INISectName, varname, val, sutil__INIFileName)
endFunction

variant function opt__userOptionCycler(string varname, string choices, int iRetCurVal)
; Used by DlgSelectFunctionToRun cycler/toggler functions to cycle/toggle variable values.
; Varname should be the option name being handled (SV_* constant).
; Choices should be a list of text values, separated by vertical bars (|).
; iRetCurVal should be the integer taken by the calling toggler/cycler function.
; The return value of this function behaves according to specifications for DlgSelectFunctionToRun functions and should be returned directly by the toggler/cycler function.
; The integer value saved in skype.ini will be the 0-based index of the chosen option.
var
	int i,
	int n
i = opt__getUserOption(varname)
n = string__segmentCount(choices, "|")
if !iRetCurVal
	; Update it.
	i = i +1
	if i >= n
		i = 0
	endIf
	opt__setSkypeVar(varname, i)
endIf
; i+1 rather than just i because i is 0-based and stringSegment is 1-based.
return stringSegment(choices, "|", i+1)
endFunction

;================================ End module opt ===============================


;================================ Module appOpt ================================
; User option code.

String Function jvoRecognizeChat (int RetCurValue)
return opt__userOptionCycler(opt__SV_RecognizeChat, opt__msgDisabledEnabled, RetCurValue)
EndFunction

String Function jvoSpeakOnlineStatus (int RetCurValue)
return opt__userOptionCycler(opt__SV_SpeakOnlineStatus, opt__msgDisabledEnabled, RetCurValue)
EndFunction

String Function jvoTypingAnnounce (int RetCurValue)
return opt__userOptionCycler(opt__SV_TypingAnnounce, opt__msgDisabledEnabled, RetCurValue)
EndFunction

String Function jvoTypingSounder (int RetCurValue)
var
	string result,
	string idx
result = opt__userOptionCycler(opt__SV_TypingSounder, opt__msgTypingSounderChoices, RetCurValue)
if !RetCurValue
	; A change, so play a sample
	idx = opt__getUserOption(opt__SV_TypingSounder)
	if idx
		sutil__play(formatString(sutil__fNameSkypeTyping, idx))
	endIf
endIf
return result
EndFunction

String Function jvoSkypeWatchMessages (int RetCurValue)
return opt__userOptionCycler(opt__SV_SkypeWatchMessages, opt__msgSkypeWatchMessageChoices, RetCurValue)
EndFunction

String Function jvoChatMessageTimes (int RetCurValue)
return opt__userOptionCycler(opt__SV_ChatMessageTimes, opt__msgChatMessageTimeChoices, RetCurValue)
EndFunction

String Function jvoSkypeCommandKey (int iRetCurVal)
; A special verbosity/options function for managing the Skype command key.
var
	string varname,
	string sKey
varname = opt__SV_SkypeCommandKey
sKey = opt__getUserOption(varname)
if !iRetCurVal
	if Skype__appOpt__giSettingSkypeCommandKey == 0
		; The user just pressed Space on this entry:
		; Ask the user for a key .
		sayMessage(OT_No_Disable, appOpt__msgKeyChangePrompt)
		; Prevent the next key from doing anything except calling KeyPressedEvent.
		trapKeys(True)
		; Tell the KeyPressedEvent we're expecting a Skype Command key to be typed now.
		Skype__appOpt__giSettingSkypeCommandKey = 1
		; Prevent JAWS from saying the current key value again.
		speechOff()
		scheduleFunction("speechOn", 2)
		; but then return as normal.
		; JAWS should fall silent after the above prompt.
		; The next user key will hit the KeyPressedEvent and get stored,
		; and then KeyPressedEvent will simulate a Space,
		; which will make DlgSelectFunctionToRun call this function again,
		; with iRetCurVal False but Skype__appOpt__giSettingSkypeCommandKey still 1,
		; and control will continue below.
	elif Skype__appOpt__giSettingSkypeCommandKey == 1
		; KeyPressedEvent caused this call on receiving (and setting) the new command key.
		trapKeys(False)
		; Set all relevant jkm lines to match the newly stored key.
		appOpt__resetSkypeCommandKey()
		Skype__appOpt__giSettingSkypeCommandKey = 0
	endIf  ; Skype__appOpt__giSettingSkypeCommandKey
endIf  ; !iRetCurVal
return sKey
EndFunction

void function appOpt__resetSkypeCommandKey()
var
	string sect, string fname,
	string val,
	string sKey, string sKeyList,
	int i, int n,
	string seg,
	int pos,
	int changing
sKey = opt__getUserOption(opt__SV_SkypeCommandKey)
mkcs__setCommandKey(sKey)
endFunction

void function appOpt___addVerbosityOption(string byRef lst, string optionData)
; Adds an option to the given verbosity (JAWS 8-) / option (JAWS 9+) list.
var
	string varAndDesc
varAndDesc = formatString("jvo%1:%2",
	stringSegment(optionData, ":", 1),
	stringSegment(optionData, ":", 2))
if lst
	lst = lst +"|"
endIf
lst = lst +varAndDesc
endFunction

Script AdjustJawsOptions ()
performScript adjustJAWSVerbosity()
endScript

Script AdjustJawsVerbosity ()
; Uses Jaws's central verbosity coding.
; Modified for skype usage by Ner
; To write more toggle functions for this, see toggle functions
; They *must* be done in the same manner.
var
	string lst

lst = ""
appOpt___addVerbosityOption(lst, opt__SV_RecognizeChat)
appOpt___addVerbosityOption(lst, opt__SV_ChatMessageTimes)
appOpt___addVerbosityOption(lst, opt__SV_SpeakOnlineStatus)
appOpt___addVerbosityOption(lst, opt__SV_TypingAnnounce)
appOpt___addVerbosityOption(lst, opt__SV_TypingSounder)
appOpt___addVerbosityOption(lst, opt__SV_SkypeCommandKey)
if sutil__isSkypeWatchLoaded()
	appOpt___addVerbosityOption(lst, opt__SV_SkypeWatchMessages)
endIf
OptionsTreeCore(lst)
EndScript

String Function NodeHLP (string NodeName)
	if NodeName == GetActiveConfiguration ()
return appOpt__msgNodeHlp
endIf
return nodeHlp(nodeName)
EndFunction

String Function jvoRecognizeChatHlp()
return stringSegment(opt__SV_RecognizeChat, ":", 3)
EndFunction

String Function jvoChatMessageTimesHlp()
return stringSegment(opt__SV_ChatMessageTimes, ":", 3)
EndFunction

String Function jvoSpeakOnlineStatusHlp()
return stringSegment(opt__SV_SpeakOnlineStatus, ":", 3)
EndFunction

String Function jvoTypingAnnounceHlp()
return stringSegment(opt__SV_TypingAnnounce, ":", 3)
EndFunction

String Function jvoTypingSounderHlp()
return stringSegment(opt__SV_TypingSounder, ":", 3)
EndFunction

String Function jvoSkypeCommandKeyHlp()
return stringSegment(opt__SV_SkypeCommandKey, ":", 3)
EndFunction

String Function jvoSkypeWatchMessagesHlp()
return stringSegment(opt__SV_SkypeWatchMessages, ":", 3)
EndFunction

;============================== End module appOpt ==============================


;=============================== Module winfinder ==============================
; Ways to search for windows by typeCode, class substring match, and by calling test functions.
;
; Author:  Doug Lee

globals
; Variables used to communicate between winfinder__findWindow and winfinder___findWindowHelper.
; Required in order to use enumerateChildWindows().
	handle Skype__winfinder___windowFound,
	string Skype__winfinder___windowClass,
	string Skype__winfinder___windowName

int function winfinder___findWindowHelper(handle hwnd)
; Called via EnumerateChildWindows.
var
	int pos,
	string wclass,
	int found
wclass = Skype__winfinder___windowClass
if stringContains("@!", stringLeft(wclass, 1))
	pos = stringContains(wclass, "%%")
	if pos
		wclass = stringLeft(wclass, pos-1) +intToString(hwnd) +stringChopLeft(wclass, pos+1)
	endIf
endIf
if wclass == "@"
	found = stringToInt(formatStringWithEmbeddedFunctions("<" +stringChopLeft(wclass, 1) +">"))
elif wclass == "!"
	found = !stringToInt(formatStringWithEmbeddedFunctions("<" +stringChopLeft(wclass, 1) +">"))
else
	found = (!stringCompare(getWindowClass(hwnd), wclass, False) && !stringCompare(getWindowName(hwnd), Skype__winfinder___windowName, False))
endIf
if found
	Skype__winfinder___windowFound = hwnd
	return False
endIf
return True
endFunction

handle function winfinder__findWindow(handle hTop, string wclass, string wname)
; Generalized window finder. Can be used just like JAWS' FindWindow function.
; Also accepts in wclass @funcCall or !funcCall to use arbitrary function calls to test windows.
; funcCall should be a function call with %% in place of the hwnd parameter.
; NOTE: The Name option uses getWindowName, not WM_GETTEXT. They can differ.
; The JAWS built-in FindWindow function uses WM_GETTEXT according to impirical data.
var
handle hwnd,
	handle nullHandle
hwnd = findWindow(hTop, wclass, wname)
if hwnd
	return hwnd
endIf
Skype__winfinder___windowClass = wclass
Skype__winfinder___windowName = wname
Skype__winfinder___windowFound = nullHandle
EnumerateChildWindows (hTop, "winfinder___findWindowHelper")
return Skype__winfinder___windowFound
endFunction

int function winfinder___isTypeCode(handle hwnd, int typeCode, int isSubtype)
; Returns True if the given window has the given type or subtype code.
if isSubtype
	return (getWindowSubtypeCode(hwnd) == typeCode)
else
	return (getWindowTypeCode(hwnd) == typeCode)
endIf
endFunction

handle function winfinder__findWindowByTypeCode(handle hTop, int typeCode, int isSubtype)
; Finds the first window under hTop (inclusive) having the given type or subtype code.
; Author:  Doug Lee
return findWindowByType(hTop, typeCode)
endFunction

int function winfinder__getWindowRole(handle hwnd)
; Returns the given window's MSAA client object's role.
; This role, if not Role_System_Client, should represent the window's actual function.
; Author:  Doug Lee
var int childID  ; required by GetObjectFromEvent but otherwise ignored here
return getObjectFromEvent(hwnd, -4, 0, childID).accRole(0)
endFunction

int function winfinder__getWindowValue(handle hwnd)
; Returns the given window's MSAA client object's value.
; Author:  Doug Lee
var int childID  ; required by GetObjectFromEvent but otherwise ignored here
return getObjectFromEvent(hwnd, -4, 0, childID).accValue(0)
endFunction

int function winfinder___isMSAARole(handle hwnd, int iRole)
; Returns True if the given window's MSAA client object has the given role.
; Author:  Doug Lee
var int childID
return (winfinder__getWindowRole(hwnd) == iRole)
endFunction

handle function winfinder__findWindowByMSAARole(handle hTop, int iRole)
; Finds the first window under hTop (inclusive) whose MSAA client object has the given role.
; Author:  Doug Lee
return winfinder__findWindow(hTop, "@winfinder___isMSAARole(%%, " +intToString(iRole) +")", "")
endFunction

int function winfinder___isClassMatch(handle hwnd, string classPart)
; Returns True if the given window's class name contains classPart.
if stringContains(stringLower(getWindowClass(hwnd)), stringLower(classPart))
	return True
endIf
return False
endFunction

handle function winfinder__findWindowByClassSubstring(handle hTop, string classPart)
; Finds the first window under hTop (inclusive) whose class name contains classPart.
; Author:  Doug Lee
return winfinder__findWindow(hTop, "@winfinder___isClassMatch(%%, " +classPart +")", "")
endFunction

;============================= End module winfinder ============================


;================================ Module appwin ================================
; getSkypeWindow and related app window identification code.

globals
; The main Skype window
	handle Skype__appwin___mainWindow,
; A place to cache the MSAA contact list object.
	object Skype__appwin___contactListObject

int function appwin__isMainSkypeWindow(handle hwnd)
var
	string wclass
wclass = getWindowClass(hwnd)
if wclass == "TskMainForm"
|| wclass == "TZapMainForm"
	return True
endIf
return False
endFunction

handle function appwin__getMainSkypeWindow()
; Returns the top-level main Skype window if possible.
; If the main window cannot be found, also clears the MSAA contact list object cache.
var
	object nullObject,
	handle hwnd
hwnd = getTopLevelWindow(getFocus())
if appwin__isMainSkypeWindow(hwnd)
	Skype__appwin___mainWindow = hwnd
	return hwnd
endIf
hwnd = Skype__appwin___mainWindow
if appwin__isMainSkypeWindow(hwnd)
	return hwnd
endIf
hwnd = findTopLevelWindow("tSkMainForm", "")
if !hwnd
	hwnd = findTopLevelWindow("tSkMainForm.UnicodeClass", "")
endIf
if hwnd
	Skype__appwin___mainWindow = hwnd
	return hwnd
endIf
Skype__appwin___contactListObject = nullObject
return 0
endFunction

int function appwin__isChatWindow(handle hwnd)
; Returns 1 (True) if the given window is or is within a Skype Instant Messaging window.
while hwnd
	if getWindowClass(hwnd) == "TConversationForm"
		return 1
	endIf
	hwnd = getParent(hwnd)
endWhile
return 0
endFunction

int function appwin___isSkypeToolTip(handle hwnd)
; Returns True if the given window is the Skype tooltip window.
return hwnd && stringContains(stringLower(getWindowClass(hwnd)), appwin___cwsToolTip)
		;&& getParent(hwnd)
		;&& getWindowClass(getParent(hwnd)) == appwin___wcApp && getWindowName(getParent(hwnd)) == appwin__wnSkype
endFunction

int function appwin__isSkypeNotifyWindow(handle hwnd)
; Returns True if the given window is a Skype system tray notification window or call alert.
var
	string wclass
wclass = getWindowClass(hwnd)
return (wclass == "TCallNotificationWindow"
|| wclass == "TTrayAlert")
endFunction

handle function appwin__findFirstSkypeNotifyWindow()
; Find the first available Skype notification window and return its handle.
; Returns null on failure.
var handle h1 = findTopLevelWindow("TTrayAlert", "")
var handle h2 = findTopLevelWindow("TCallNotificationWindow", "")
if h1 && h2
&& getWindowHierarchyX(h2) < getWindowHierarchyX(h1)
	h1 = h2
elif !h1
	h1 = h2
endIf
return h1
endFunction

int function appwin___isRequestedSkypeWindow(handle hwnd, string which)
; Returns True if the given window is the given Skype window.
if which == appwin__SWT_ToolTip
	return appwin___isSkypeToolTip(hwnd)
elif which == appwin__SWT_Notify
	return appwin__isSkypeNotifyWindow(hwnd)
else
	return False
endIf
endFunction

int function appwin__isSkypeWindow(handle hwnd)
; Returns True if the given window is part of Skype.
var
	string wclass
if stringRight(string__ref(getWindowOwner(hwnd)), 9) == "Skype.exe"
	return True
endIf
return False
endFunction

handle function appwin___findTopLevelSkypeWindow(string which)
; Returns the handle of the requested Skype top-level window.
; This window may not exist, such as right after Skype loads.
var
	handle hwnd
hwnd = util__getFirstTopLevelWindow()
while hwnd && !appwin___isRequestedSkypeWindow(hwnd, which)
	hwnd = getNextWindow(hwnd)
endWhile
if appwin___isRequestedSkypeWindow(hwnd, which)
	return hwnd
endIf
return 0
endFunction

handle function appwin__getActiveChatWindow(handle hTop)
var
	handle hwnd
hwnd = findWindow(hTop, "TConversationForm", "")
if !hwnd
	hwnd = findWindow(hTop, "TConversationForm.UnicodeClass", "")
endIf
if !hwnd || isWindowVisible(hwnd)
	; Got it or there isn't such a thing.
	return hwnd
endIf
hwnd = getNextWindow(hwnd)
while hwnd
	if isWindowVisible(hwnd) && appwin__isChatWindow(hwnd)
		return hwnd
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

handle function appwin__getSkypeWindow(string which)
; Returns the handle of the window of the indicated type.
; Note that this only works for window types that are unique.
; This function should be kept in sync with the set of SWT_* constants in skype.jsh.
var
	handle hMain, handle hwnd, handle hwnd1,
	string hList
if which == "?"
	hList = appwin__SWT_Main +"|" +appwin__SWT_Myself
		+"|" +appwin__SWT_ContactList +"|" +appwin__SWT_ConversationList +"|" +appwin__SWT_LiveConversations
		+"|" +appwin__SWT_History
	hList = hList +"|" +appwin__SWT_ChatMain +"|" +appwin__SWT_ChatInput +"|" +appwin__SWT_ChatOutput +"|" +appwin__SWT_ChatList +"|" +appwin__SWT_ChatContact +"|" +appwin__SWT_Communicator +"|" +appwin__SWT_CallQuality
	; Omitted knowingly:  appwin__SWT_ToolTip, appwin__SWT_Notify.
	return hList
endIf
hMain = appwin__getMainSkypeWindow()
if !hMain
	return 0
endIf
if which == appwin__SWT_Main
	return hMain
endIf
if which == appwin__SWT_Myself
	return winfinder__findWindowByClassSubstring(hMain, appwin___cwsMyself)
elif which == appwin__SWT_ContactList
	hwnd1 = findWindow(hMain, "TConversationList", "")
	if hwnd1
		if which == appwin__SWT_ContactList
			return hwnd1
		endIf
		return getParent(hwnd1)
	endIf
	return 0
elif which == appwin__SWT_ConversationList
	return findWindow(hMain, "TInboxList", "")
elif which == appwin__SWT_LiveConversations
	hwnd = findWindow(hMain, "TLiveConversationList", "")
	return hwnd
elif which == appwin__SWT_History
	return winfinder__findWindowByClassSubstring(hMain, "history")
elif which == appwin__SWT_ToolTip
	return appwin___findTopLevelSkypeWindow(appwin__SWT_ToolTip)
elif which == appwin__SWT_Notify
	var handle hwndn1, handle hwndn2
	hwndn1 = findTopLevelWindow("TCallNotificationWindow", "")
	hwndn2 = findTopLevelWindow("TTrayAlert", "")
	if hwndn1 && hwndn2
		if getWindowHierarchyX(hwndn1) < getWindowHierarchyX(hwndn2)
			hwnd1 = hwndn1
		else
			hwnd1 = hwndn2
		endIf
	elif hwndn1
		hwnd1 = hwndn1
	else
		hwnd1 = hwndn2
	endIf
	return hwnd1
; TZapCommunicator is in each conversation in Compact view,
; but in Default view, it's at the same level as the conversation windows.
elif which == appwin__SWT_Communicator
	hwnd1 = findWindow(hMain, "TZapCommunicator", "")
	if hwnd1
		return hwnd1
	endIf
endIf
if which == appwin__SWT_ChatPrefix || which == appwin__SWT_CallQuality
	; Conversations can be part of the Skype window tree or their own independent tree.
	hMain = getTopLevelWindow(getFocus())
	if !appwin__isChatWindow(hMain)
		hMain = appwin__getActiveChatWindow(hMain)
	endIf
	if !hMain
		return 0
	endIf
	if which == appwin__SWT_ChatMain
		return hMain
	endIf
	if which == appwin__SWT_ChatInput
		return winfinder__findWindowByClassSubstring(hMain, "TChatRichEdit")
	elif which == appwin__SWT_ChatList
		return findWindow(hMain, "TParticipantList", "")
	elif which == appwin__SWT_ChatContact
		hwnd = findWindow(hMain, "TConversationHeader", "")
		if !hwnd
			hwnd = findWindow(hMain, "TContactProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hMain, "TConversationProfile", "")
		endIf
		if !hwnd
			hwnd = findWindow(hMain, "TParticipantList", "")
		endIf
		return hwnd
	elif which == appwin__SWT_ChatOutput
		return findWindow(hMain, "TChatContentControl", "")
	elif which == appwin__SWT_CallQuality
		return findWindow(hMain, "TCallQualityControlHost", "")
	endIf
	return 0
endIf
return 0
endFunction

object function appwin__getMSAAObject(string which)
; Returns the requested MSAA object, which should be an SWT_* constant.
; appwin__SWT_MenuBar is special:  It returns the Skype menu bar object but has no corresponding purpose in appwin__getSkypeWindow().
var
	object o, int childID, int n,
	handle hwnd
if conv5__applies()
	; In Skype 5.0 Gold, these have MSAA but no window.
	if which == appwin__SWT_ContactList
		return conv5__listByName(conv5__ListContacts)
	elif which == appwin__SWT_LiveConversations
		return conv5__listByName(conv5__ListLive)
	elif which == appwin__SWT_ConversationList
		return conv5__listByName(conv5__ListConversations)
	elif which == appwin__SWT_History
		return conv5__listByName(conv5__ListHistory)
	elif which == appwin__SWT_ChatOutput
		return conv5__chatContent(appwin__getSkypeWindow(appwin__SWT_ChatOutput))
	endIf
endIf
if which != appwin__SWT_MenuBar
	hwnd = appwin__getSkypeWindow(which)
	if !hwnd
		return 0
	endIf
	o = getObjectFromEvent(hwnd, -4, 0, childID)  ; Gets the Role_System_Client (or more specific role) object for this window
	return o
endIf
hwnd = appwin__getMainSkypeWindow()
if !hwnd
	return 0
endIf
o = getObjectFromEvent(hwnd, 0, 0, childID)  ; Gets the Role_System_Window object for this window
if !o
	return 0
endIf
return o.accChild(-3)  ; application menu bar
endFunction

object function appwin__getContactListObject()
; Returns the MSAA object for the contact list, if it can be found.
var
	handle hwnd,
	object o, int childID,
	object nullObject

; Clear out the cached object if Skype is not loaded (appwin__getMainSkypeWindow does this).
if !appwin__getMainSkypeWindow()
	return nullObject
endIf

; Return the cached object if it is valid.
if Skype__appwin___contactListObject
	if Skype__appwin___contactListObject.accRole(0) == Role_System_List
	|| Skype__appwin___contactListObject.accRole(0) == Role_System_Pane
		return Skype__appwin___contactListObject
	elif Skype__appwin___contactListObject.accRole(0) == Role_System_Client
		return nullObject
	endIf
endIf

Skype__appwin___contactListObject =
conv5__listByName(conv5__ListContacts)
if Skype__appwin___contactListObject
	return Skype__appwin___contactListObject
endIf

hwnd = appwin__getSkypeWindow(appwin__SWT_ContactList)
if hwnd
	o = getObjectFromEvent(hwnd, 0-4, 0, childID)
	if !o
		Skype__appwin___contactListObject = nullObject
		return nullObject
	elif o.accRole(0) != Role_System_List && o.accRole(0) != Role_System_Pane
		Skype__appwin___contactListObject = nullObject
		return nullObject
	endIf
	Skype__appwin___contactListObject = o
	return Skype__appwin___contactListObject
endIf
Skype__appwin___contactListObject = nullObject
return nullObject
endFunction

;============================== End module appwin ==============================


;================================== Module tb ==================================
; Skype toolbar handling code.
;
; Usage:
;	Call tb__makeList(names, data) (strings) to get toolbar item names and info.
;	Call tb__clickItem(seg) to activate an item (seg from data).
;
; TODO: Depends on a number of out-of-module constants.
;
; Author:  Doug Lee

string function tb___cleanOffLeadingGraphic(string s)
; Clean off a leading graphic if there is one (e.g., "graphic 327") from the beginning of the given string, and return the result.
var
	int n, int len, int i
if s == tb___JAWSGraphic
	s = stringTrimLeadingBlanks(stringChopLeft(s, stringLength(tb___JAWSGraphic)))
	len = stringLength(s)
	i = 1
	while i <= len && stringContains("0123456789", substring(s, i, 1))
		i = i +1
	endWhile
	if i > 1
		s = stringTrimLeadingBlanks(stringChopLeft(s, i-1))
	endIf
endIf
return s
endFunction

int function tb___addGraphicButtonsWithGivenNames(string byRef names, string byRef data, int n, string givenNames)
; Adds graphic buttons to names and data.
; Assumes the JAWS or (preferably) invisible cursor is on the first button and that buttons can be reached successively via NextWord().
; Also assumes all graphics stop the cursor; use OPT_Include_Graphics=2 for best results.
; Returns the number of buttons successfully added.
var
	int i
i = 1
while i <= n
	if !getGraphicID()
		beep()
		i = i-1  ; don't count the current "icon" that really isn't one.
		log__write(formatString("Ran out of graphics, found %1 of %2",
			intToString(i), intToString(n)
		))
		return i
	endIf
	names = string__join(names, "|", stringSegment(givenNames, "|", i))
	; Window handle, x/y of natural JAWS/invisible cursor position for graphic, childID of 0.
	data = string__join(data, "|",
			intToString(getCurrentWindow()) +","
			+intToString(getCursorCol()) +"," +intToString(getCursorRow()) +",0"
			)
	nextWord()
	i = i +1
endWhile
endFunction

int function tb___addWindow(string byRef names, string byRef data, handle hwnd, string givenName)
; Add the given window, as a named toolbar item, to names and data.
; Assumes the givenName includes all type/value/state information desired; nothing is drawn from the window itself except coordinates.
; Returns 1 if successful and 0 if not.
names = string__join(names, "|", givenName)
; Window handle, x/y of center of window, childID of 0.
data = string__join(data, "|",
		intToString(hwnd) +","
		+intToString((getWindowLeft(hwnd)+getWindowRight(hwnd))/2) +","
		+intToString((getWindowTop(hwnd)+getWindowBottom(hwnd))/2) +",0"
		)
return 1
endFunction

int function tb___addIconsWithTextLabels(string byRef names, string byRef data, handle hwnd, string nameSuffix)
; Adds all icons with text labels in the given window to names and data.
; Grayed-out items are still listed but shown as disabled.
; nameSuffix, if set, is appended to each item's text.
; Returns the number of items successfully added.
var
	int oldRestriction,
	int done,
	string buf,
	int nitems, int safetyCount
saveCursor() invisibleCursor() saveCursor()
moveToWindow(hwnd)
oldRestriction = getRestriction()
if !util__setRestrictionToWindow(hwnd)
	return 0
endIf
JAWSPageUp() JAWSHome()
nextChunk() priorChunk() priorChunk()  ; Make sure we're squarely on the first one
done = 0
safetyCount = 0
while !done && getCurrentWindow() == hwnd && safetyCount < 50
	buf = tb___cleanOffLeadingGraphic(stringTrimLeadingBlanks(getChunk()))
	buf = stringTrimLeadingBlanks(stringTrimTrailingBlanks(buf))
	if stringLength(buf)
		if nameSuffix
			buf = buf +" " +nameSuffix
		endIf
		names = string__join(names, "|", buf)
		if getColorName(getColorText()) == tb___JAWSGray
			; Commented out 2007-07-10 because chat toolbar items are grey when enabled.
			;names = names +" " +tb___msgDisabled
		endIf
		; Window handle, x/y of current position on text label, childID of 0.
		; A little offset is added to the x coordinate to avoid missing the button.
		data = string__join(data, "|",
				intToString(hwnd) +","
				+intToString(getCursorCol()+5) +"," +intToString(getCursorRow()) +",0"
				)
		nitems = nitems +1
	endIf
	done = !nextChunk()
	safetyCount = safetyCount +1
endWhile
if safetyCount >= 50
	log__write("Loop ran out.")
endIf
setRestriction(oldRestriction)
return nitems
endFunction

int function tb___addMSAAIconsWithGivenNames(string byRef names, string byRef data, handle hwnd, string givenNames)
; Adds all MSAA toolbar items in the given window to names and data.
; Returns the number of items successfully added.
var
	object o, int childID,
	int i, int nitems, int nreal,
	string newname
o = getObjectFromEvent(hwnd, 0-4, 0, childID)
nitems = o.accChildCount
nreal = 0
i = 1
while i <= nitems
	newname = sutil__getToolBarItemDesc(stringSegment(givenNames, "|", i), o, i, False)
	if newname
		names = string__join(names, "|", newname)
		; Window handle, 0,0, childID.
		data = string__join(data, "|",
				intToString(hwnd) +",0,0,"
				+intToString(i)
				)
		nreal = nreal +1
	endIf
	i = i +1
endWhile
return nreal
endFunction

int function tb___addContactGroupItems(string byRef names, string byRef data, handle hwnd)
var
	int oldRestriction,
	int x, int y
saveCursor() invisibleCursor() saveCursor()
moveToWindow(hwnd)
oldRestriction = getRestriction()
setRestriction(restrictWindow)
JAWSPageUp() JAWSHome()
y = getCursorRow()
names = string__join(names, "|", tb___msgContactButtons)
x = getWindowRight(hwnd)
data = string__join(data, "|", intToString(hwnd) +"," +intToString(x-30) +"," +intToString(y) +",0")
data = string__join(data, "|", intToString(hwnd) +"," +intToString(x-15) +"," +intToString(y) +",0")
setRestriction(oldRestriction)
return True
endFunction

globals
	string Skype__tb___names,
	string Skype__tb___data

int function tb___enumHelper(handle hwnd)
; Window enumeration helper for tb__makeList.
var
	object o, int childID,
	int iRole, string sRole, int iState,
	int n,
	string name
if !isWindowVisible(hwnd)
	return True
elif getWindowClass(hwnd) == "TSkMoreButton"
	var string wname
	wname = getWindowText(hwnd, Read_Everything) +" button"
	if isWindowDisabled(hwnd)
		wname = wname +" " +tb___msgDisabled
	endIf
	tb___addWindow(Skype__tb___names, Skype__tb___data, hwnd, wname)
endIf
o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
if !o
	return True
endIf
iRole = o.accRole(0)
if !iRole
	; "Give Feedback" button and Call Phones/Directory/Shop.
	tb___addIconsWithTextLabels(Skype__tb___names, Skype__tb___data, hwnd, "button")
elif iRole != Role_System_Pane
	return True
endIf
n = o.accChildCount
childID = 1
while childID <= n
	iRole = o.accRole(childID)
	if iRole != Role_System_Window
		iState = o.accState(childID)
		if !(iState & (State_System_Invisible +State_System_Offscreen))
		&& sutil__isUnobscured(o, childID)
			name = sutil__removeHTML(o.accName(childID))
			if !name
				name = MSAA__rectText(o, childID)
			endIf
			if !name
				name = getWindowName(hwnd)
				if !name
					name = getWindowClass(hwnd)
					if stringLeft(name, 1) == "t"
					&& stringContains("ABCDEFGHIJKLMNOPQRSTUVWXYZ", substring(name, 2, 1))
						name = stringChopLeft(name, 1)
						string__replace(name, "control", "", "e")
					endIf
				endIf
				name = formatString("%1:%2", name, intToString(childID))
			endIf
			if iState & State_System_Unavailable
				name = name +" " +tb___msgDisabled
			endIf
			sRole = getRoleText(iRole)
			; TODO: This is English-specific but probably can't meaningfully be localized just by making string constants.
			if sRole == "push"
				sRole = stringChopLeft(sRole, 4)
			endIf
			name = name +" " +sRole
			Skype__tb___names = string__join(Skype__tb___names, "|", name)
			; Window handle, 0,0, childID.
			Skype__tb___data = string__join(Skype__tb___data, "|",
					intToString(hwnd) +","
					+"0,0," +intToString(childID)
					)
		endIf  ; visible and on-screen items
	endIf  ; non-Role_System_Window items
	childID = childID +1
endWhile  ; childIDs in window
return True
endFunction

int function tb__makeList(string byRef names, string byRef data)
; Makes a list of buttons/controls based on the current location within Skype.
; This function is responsible for including all toolbar buttons and any special exceptions.
; Returns the number of items in the list.
var
	handle hTop
Skype__tb___names = ""
Skype__tb___data = ""
hTop = getTopLevelWindow(getFocus())
enumerateChildWindows(hTop, "tb___enumHelper")
names = Skype__tb___names
data = Skype__tb___data
Skype__tb___names = ""
Skype__tb___data = ""
return string__segmentCount(names, "|")
endFunction

int function tb__clickItem(string data)
; Click (or otherwise activate) the button/control indicated by the given data string.
; Returns True on success and False on failure.
var
	handle hwnd, int x, int y,
	object o, int childID,
	int left, int right, int top, int bottom
hwnd = stringToHandle(stringSegment(data, ",", 1))
x = stringToInt(stringSegment(data, ",", 2))
y = stringToInt(stringSegment(data, ",", 3))
childID = stringToInt(stringSegment(data, ",", 4))
if childID
	o = getObjectFromEvent(hwnd, 0-4, childID, childID)
	; o.accDoDefaultAction(childID) does nothing in this app.
	if MSAA__rect(o, childID, left, right, top, bottom)
		saveCursor() JAWSCursor() saveCursor()
		moveTo( (left+right)/2, (top+bottom)/2 )
		leftMouseButtonLock()
		delay(1)
		leftMouseButtonLock()
		return True
	endIf
else
	saveCursor() JAWSCursor() saveCursor()
	moveTo(x, y)
	leftMouseButton()
	return True
endIf
return False
endFunction

;================================ End module tb ================================


;================================= Module conv5 ================================
; Utilities for managing the Skype 5 beta 2 and Gold, and (hopefully) later conversations control.
; Also includes a function to obtain the chat output message list in these Skype versions.

globals
; The main conversations window.
	handle Skype__conv5___hwndConv,
; A flag indicating whether this window has been found to be compatible with this module.
; -1 means no, 1 means yes, 0 means don't know yet.
	int Skype__conv5___compat,
; Cached list collections for live, contacts, conversations, and history.
; Key is a constant passed to conv5__listByName(),
; value is an MSAA object.
	collection Skype__conv5___listCache,
; RealName-to-scriptName map for the cache.
; Key is an actual MSAA list's name,
; value is the constant from this module used to find the MSAA object in the cache.
	collection Skype__conv5___listNames

int function conv5__applies()
; Returns True if this appears to be a Skype version supporting this module's interface.
var variant null
var handle hwnd = conv5___mainWindow()
if !hwnd
	Skype__conv5___listCache = null
	Skype__conv5___listNames = null
	return False
endIf
if hwnd != Skype__conv5___hwndConv
	Skype__conv5___compat = 0
	Skype__conv5___listCache = null
	Skype__conv5___listNames = null
endIf
if Skype__conv5___compat == 1
	return True
elif Skype__conv5___compat == -1
	return False
endIf
var int childID
var object o = getObjectFromEvent(hwnd, -4, 0, childID)
if !o
	return False
endIf
Skype__conv5___hwndConv = hwnd
var int n = o.accChildCount
if n > 20
	; That's permanent; we know this isn't compatible.
	Skype__conv5___compat = -1
	return False
endIf
childID = 1
while childID <= n
	if o.accRole(childID) == Role_System_List
		Skype__conv5___compat = 1
		return True
	endIf
	childID = childID +1
endWhile
; Don't mark permanently incompatible,
; in case this just means no list has yet been formed.
return False
endFunction

handle function conv5___mainWindow()
; Return the main conversations control window.
var handle hwnd = appwin__getSkypeWindow(appwin__SWT_Main)
if !hwnd
	return
endIf
hwnd = findWindow(hwnd, "TConversationsControl", "")
if !hwnd
	return
endIf
return hwnd
endFunction

object function conv5___mainClient()
; Return the main conversations control MSAA client object.
var handle hwnd = conv5___mainWindow()
if !hwnd
	return
endIf
var int childID
return getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
endFunction

string function conv5__peopleOnline()
; Return the string indicating how many people are online, if it can be found.
; The algorithm used is intended to be language independent.
if !conv5__applies()
	return ""
endIf
var object o = conv5___mainClient()
if !o
	return ""
endIf
var int childID = 1
var int n = o.accChildCount
if n > 20
	; Too much here to be right for Skype 5.
	return ""
endIf
while childID <= n
	if o.accRole(childID) == Role_System_StaticText
	&& conv5___isPeopleCountString(o.accName(childID))
		return o.accName(childID)
	endIf
	childID = childID +1
endWhile
return ""
endFunction

int function conv5___isPeopleCountString(string s)
; Returns True if s appears to represent the number of people online in Skype.
; The algorithm used is intended to be language independent.
var string digs = "0123456789"
if !stringContainsChars(s, digs)
	return False
endIf
var int i = 1
var int n = stringLength(s)
var int cnt = 0
var string ch
while i <= n
	ch = substring(s, i, 1)
	if stringContains(digs, ch)
		cnt = cnt +1
		if cnt > 4
			return True
		endIf
	else
		if cnt > 0 && !stringContains(digs, substring(s, i+1, 1))
			; Two separate numbers.
			return False
		endIf
	endIf
	i = i +1
endWhile
return False
endFunction

object function conv5__listByName(string listname)
; Return the list of given name or null.
if !conv5__applies()
	return
elif collectionItemExists(Skype__conv5___listCache, listname)
	return Skype__conv5___listCache[listname]
endIf
var object o = conv5___mainClient()
if !o
	return
endIf
var int childID = 1
var int n = o.accChildCount
while childID <= n
	if o.accRole(childID) == Role_System_List
	&& conv5___nameMatches(o.accName(childID), listname)
		if !Skype__conv5___listNames
			Skype__conv5___listCache = new collection
			Skype__conv5___listNames = new collection
		endIf
		var object oList = o.accChild(childID)
		listname = o.accName(childID)
		; Warning: Case is not significant here though it was before
		; use of JAWS 12+ collections.
		if collectionItemExists(Skype__conv5___listCache, listname)
			oList = Skype__conv5___listCache[listname]
		else
			Skype__conv5___listCache[listname] = oList
			Skype__conv5___listNames[o.accName(childID)] = listname
		endIf
		return oList
	endIf
	childID = childID +1
endWhile
return
endFunction

int function conv5___nameMatches(string name, string choices)
; Returns true if name is or is in choices.
if stringCompare(name, choices, False) == 0
	return True
elif stringContains("|"+choices+"|", "|"+name+"|")
	return True
endIf
return False
endFunction

string function conv5__focusedListName()
; Return the name of the focused list, or null if there isn't one.
var object o, int childID
o = getFocusObject(childID)
if !o
	return ""
endIf
var string name = o.accName(0)
if !name
	return ""
endIf
return Skype__conv5___listNames[name]
endFunction

string function conv5__focusShowingList()
; Focus the list that is already showing on screen.
; Returns True on success and False on failure.
var object o, int childID
o = conv5___mainClient()
if !o
	return False
endIf
var int n = o.accChildCount
childID = n
while childID
	if o.accRole(childID) == Role_System_List
		o.accSelect(SelFlag_TakeFocus, childID)
		return True
	endIf
	childID = childID -1
endWhile
return False
endFunction

object function conv5__chatContent(handle hwndChatOutput)
; Get the MSAA list object of chat messages from the given chat output window handle.
if !hwndChatOutput
	return
elif !conv5__applies()
	return
endIf
var int childID
var object o = getObjectFromEvent(hwndChatOutput, ObjID_Client, 0, childID)
if !o
	return
elif conv5___nameMatches(o.accName(1), conv5__ListChatOutput)
	return o.accChild(1)
elif conv5___nameMatches(o.accName(2), conv5__ListChatOutput)
	return o.accChild(2)
elif conv5___nameMatches(o.accName(3), conv5__ListChatOutput)
	return o.accChild(3)
endIf
return
endFunction

;=============================== End module conv5 ==============================


;================================= Module live5 ================================
; Code for handling the Skype 5.1-5.2 Live Call area.
;
; Author:  Doug Lee


collection function live5__getInfo(/*?*/ handle hwnd)
; Get an info collection describing the live area whose chat handle is given:
;	hwnd, obj - The window handle and MSAA Pane object for the area.
;	participants - A 1-based collection of participants.
;	muted - True if the call is muted, False if not.
;	held - "l" for local, "r" for remote, empty or missing otherwise.
;	videoOut, videoIn - "y" if video is being sent or received, "n" if it can be but isn't.
;	callQuality - An indicator of call quality issues.
;	duration - The current call duration as hh:mm:ss.
;	peopleSelector - The Add People popup area's Pane object if present.
;	toolbar - The area's toolbar's MSAA object.
; If hwnd is null or not given, the one containing the focused window is used.
var collection oInfo oInfo = new collection
var string rightClass = "TLiveConversation"
if !hwnd
	hwnd = getFocus()
endIf
hwnd = schat__getTopOfChat(hwnd)
if !hwnd
	return oInfo
endIf
hwnd = findWindow(hwnd, rightClass, "")
if !hwnd
	return oInfo
endIf
if getWindowClass(getFirstChild(hwnd)) == rightClass
	; JAWS matched TLiveConversationPanel with TLiveConversation.
	hwnd = getFirstChild(hwnd)
endIf
if getWindowClass(getParent(hwnd)) != rightClass
	return oInfo
endIf

var object oLive = MSAA__objectFromWindow(hwnd)
if !oLive
	return oInfo
endIf

oInfo.hwnd = hwnd
oInfo.obj = oLive
var collection cParticipants cParticipants = new collection
oInfo.participants = cParticipants

var object oToolbar
var string name, int iRole
var int callHeld = False
var int n = oLive.accChildCount
var int nparts = 0
; Count backward to find the CallToolBar before absorbing arbitrary accNames as participants.
var int childID = n
while childID
	name = oLive.accName(childID)
	iRole = oLive.accRole(childID)
	if iRole == Role_System_Pane
		if oToolbar
			nparts = nparts +1
			cParticipants[intToString(nparts)] = name
			; Skype 5.5 started putting the call duration for a non-conference
			; in the participant's pane.
			if childID == 1
				var object oPart = oLive.accChild(childID)
				var int nids = oPart.accChildCount
				var string maybeDuration = oPart.accName(nids)
				if oPart.accRole(nids) == Role_System_StaticText
				&& stringContains("0123456789", stringLeft(maybeDuration, 1))
					oInfo.duration = maybeDuration
				endIf
			endIf
		elif name == "CallToolbar"
			oToolbar = oLive.accChild(childID)
			oInfo.toolbar = oToolbar
		elif name == "LiveSelectContactsWidget"
			oInfo.peopleSelector = oLive.accChild(childID)
		endIf
	elif name == schat__scHold
		callHeld = True
	elif name == schat__scResume
		; Locally held call.
		oInfo.held = "l"
		callHeld = True
	elif name == schat__scSnapshot
		oInfo.videoIn = "y"
	else
		; Skip this item.
	endIf
	childID = childID -1
endWhile
if oToolbar
	live5___processToolbar(oInfo, oToolbar, callHeld)
endIf
if !collectionItemExists(oInfo, "videoIn")
	oInfo.videoIn = "n"
endIf
return oInfo
endFunction

int function live5___processToolbar(collection oInfo, object oToolbar, int callHeld)
; Absorb info from the toolbar.
; If this turns out not to be a live area containing a call,
; oInfo is emptied.
; Returns True on success and False on failure.
if !oToolbar
	return False
endIf
var int n = oToolbar.accChildCount
var int childID = 1
var string name, int iRole
var int isCall = False
while childID <= n
	name = stringTrimLeadingBlanks(oToolbar.accName(childID))
	iRole = oToolbar.accRole(childID)
	if iRole == Role_System_StaticText
	&& stringContains("0123456789", stringLeft(name, 1))
		oInfo.duration = name
	elif name == schat__scMute
		oInfo.muted = False
	elif name == schat__scUnmute
		oInfo.muted = True
	elif name == schat__scResume
		; Locally held call.
		oInfo.held = "l"
	elif name == schat__scEndCall
		isCall = True
	elif name == schat__scMyVideo
		if (oToolbar.accState(childID) & State_System_Pressed)
			oInfo.videoOut = "y"
		else
			oInfo.videoOut = "n"
		endIf
	else
		; Skip this item.
	endIf
	childID = childID +1
endWhile
if not isCall
	oInfo.removeAll()
	return True
endIf
if callHeld && !collectionItemExists(oInfo, "held")
	; Remotely held call.
	oInfo.held = "r"
endIf
return True
endFunction

;=============================== End module live5 ==============================


;============================= Module live5selector ============================
; Code for handling the Skype 5.x Live Call area's Add People contact selector.
;
; Author:  Doug Lee

int function live5selector__selected()
; Returns a status indicating whether the current item in the list is selected:
;	0:  Focus is not in a Live Call's Add People selector list.
;	1:  Focus is in one but not on a contact entry.
;	2:  Focus is in one but the current item is not selected.
;	3:  The current item is selected.
; An item is selected if its accName appears in the accValue list in the widget's Edit box.
; That list is delimited by semicolons (;).
; There are checkbox-like things to the right of each item in the selector list,
; but they semi-transparently overlay avatars, making color-based algorithms iffy at best.
var string sep = ";"

; Set up based on current focus position.
var object oSelector, int childID
oSelector = getFocusObject(childID)
if !live5selector___isSelector(oSelector)
	return 0
elif oSelector.accRole(childID) != Role_System_ListItem
	; Right list, but a subitem like a Skype button.
	return 1
elif !oSelector.accValue(childID)
	; Item like the "Other Item" one at the top:
	; Not a contact because contacts have accValue content.
	return 1
endIf
var string curItem = oSelector.accName(childID)

; Find the childID of the Edit containing the selected item list.
; This is an element, not a full-blown Accessible Object.
var object oParent = oSelector.accParent
var int idSelections = 0
var int n = oParent.accChildCount
childID = 1
while childID <= n && !idSelections
	if oParent.accRole(childID) == Role_System_Text
		; We assume just one Edit here and so don't check its accName.
		idSelections = childID
	endIf
	childID = childID +1
endWhile
if !idSelections
	; Treat like there are no selections.
	return 2
endIf

; Exact containment test.
var string sSelections = oParent.accValue(idSelections)
if !sSelections
	; Nothing is selected.
	return 2
endIf
if stringContains(sep+sSelections+sep, sep+curItem+sep)
	return 3
endIf
return 2
endFunction

int function live5selector___isSelector(object o)
; Returns True if o is the Add People selector in a live call.
if o.accRole(0) == Role_System_List
&& o.accName(0) == "LiveCheckBoxUserListPhone"
&& o.accParent.accName(0) == "LiveSelectContactsWidget"
	return True
endIf
return False
endFunction

;=========================== End module live5selector ==========================


;================================= Module conf =================================
; Code for handling conference participants.
;
; Author: Doug Lee

collection function conf__getSpeakers()
; Get a list of who is speaking in any active conference.
; On error, the collection contains .err as follows:
;	1 - Not in a chat.
;	2 - Not in a live conversation.
;	3 - Not in a conference.
return conf__getParticipants(True)
endFunction

collection function conf__getParticipants(int justSpeakers)
; Get a list of who is participating or speaking in any active conference.
; On error, the collection contains .err as follows:
;	1 - Not in a chat.
;	2 - Not in a live conversation.
;	3 - Not in a conference.
var int tc = getTickCount()
var collection speakers speakers = new collection
var handle hTop = schat__getTopOfChat(getFocus())
if !hTop
	speakers.err = 1
	return speakers
endIf
hTop = findWindow(hTop, "TLiveConversation", "")
if !hTop
	speakers.err = 2
	return speakers
endIf
if getWindowClass(getFirstChild(hTop)) == "TLiveConversation"
	hTop = getFirstChild(hTop)
endIf

var int ntries = 3
var int done = False
while ntries && !done
	speakers = new collection
	conf___fillSpeakers(speakers, hTop, justSpeakers)
	if speakers.count > 0
		done = True
	else
		ntries = ntries -1
		if !justSpeakers
			ntries = 0
		endIf
		pause()
	endIf
endWhile
speakers.tc = getTickCount() -tc
if !speakers.count
&& MSAA__objectFromWindow(hTop).accChildCount < 2
	speakers.err = 3
endIf
return speakers
endFunction

int function conf___fillSpeakers(collection speakers, handle hwnd, int justSpeakers)
; Fill the given (presumed empty) speakers collection.
; Includes all participants unless justSpeakers is true,
; which includes all people in the conference.
if !isWindowVisible(hwnd)
|| getWindowClass(hwnd) != "TLiveConversation"
	return False
endIf
var object oTop = MSAA__objectFromWindow(hwnd)
if !oTop
	beep()
	return False
endIf
var int n = oTop.accChildCount
var int i
var collection speaker
var int nspeakers = 0
for i = 1 to n
	if conf___isSpeaking(oTop, i, justSpeakers)
		nspeakers = nspeakers +1
		speaker = new collection
		speaker.name = oTop.accName(i)
		speaker.text = oTop.accChild(1).accName(0)
		speaker.chatname = schat__getChatName(hwnd, False)
		speakers[intToString(nspeakers)] = speaker
	endIf
endFor
speakers.count = nspeakers
return True
endFunction

int function conf___isSpeaking(object o, int childID, int justSpeakers)
; Returns True if the given participant appears to be speaking.

; First make sure this is actually a participant, not a toolbar.
if o.accRole(childID) != Role_System_Pane
|| o.accChild(childID).accRole(1) != Role_System_StaticText
	return False
elif !justSpeakers
	return True
endIf

; Plan: For any visible participant pane, pick a pixel to check for blueness.
; If it is more than 50 brighter Blue than Red, consider this a speaking participant.
; Currently we use a pixel 6 from left center.
var collection rect
if getJFWVersion() >= 1400000
	var object el = UIA__elementFromMSAA(o, childID)
	if el
		rect = UIA__boundingRectangle(el, True)
		if rect
			rect.method = "UIA"
		endIf
	endIf
endIf
if !rect.method
	var int left, int right, int top, int bottom
	if !MSAA__rect(o, childID, left, right, top, bottom)
		beep()
		return False
	endIf
	rect = new collection
	rect.left = left
	rect.right = right
	rect.top = top
	rect.bottom = bottom
	rect.method = "MSAA"
endIf
if !rect.method
	beep()
	return False
endIf
var int x = rect.left +6
var int y = (rect.top +rect.bottom) / 2
var int clr = getColorAtPoint(x, y)
var int r = (clr & 0xFF)
var int b = ((clr >> 16) & 0xFF)
if b > r+50 && b > 150
	return True
endIf
return False
endFunction

;=============================== End module conf ===============================


;================================ Module typing ================================
; Code for tracking typing indicators in Skype 6.21+
;
; Author: Doug Lee

globals
; Frequency of typing indicator polls, in tenths of a second.
	int Skype__typing___pollTime,
; ScheduleFunction handle for the typing indicator poller.
	int Skype__typing___pollSch,
; Info maintained about the active chat's typing indicator.
	collection Skype__typing___info

void function typing__startPolling()
var variant null
Skype__typing___info = null
Skype__typing___pollTime = 10
typing__schedulePoll()
endFunction

void function typing__stopPolling()
var variant null
Skype__typing___info = null
if Skype__typing___pollSch
	unscheduleFunction(Skype__typing___pollSch)
	Skype__typing___pollSch = 0
endIf
endFunction

void function typing__schedulePoll(optional int ntenths)
if !ntenths
	ntenths = Skype__typing___pollTime
endIf
Skype__typing___pollSch = scheduleFunction("typing___poller", ntenths)
endFunction

void function typing___poller()
var int announce = opt__getUserOption(opt__SV_TypingAnnounce)
var int soundLevel = opt__getUserOption(opt__SV_TypingSounder)
if !announce && !soundLevel
	typing__schedulePoll(50)
	return
endIf

var collection info = typing___getInfo()
if !info
	typing__schedulePoll(10)
	return
endIf
var string oldTxt = info.txt
var string txt
if !typing__getText(txt)
	typing__schedulePoll(10)
	return
endIf
; That updates info.txt also.
if stringCompare(txt, oldTxt) != 0
	if announce && stringLength(txt)
	&& !sayAllInProgress()
		sayUsingVoice(VCTX_Message, txt, OT_Status)
	endIf
endIf
if soundLevel
	if txt
		sutil__play(formatString(sutil__fNameSkypeTyping, intToString(soundLevel)))
	endIf
endIf
typing__schedulePoll(3)
endFunction

collection function typing___getInfo()
; Gets/updates Skype__typing___info for the current typing indicator.
; Returns null if no such indicator is showing.
; Fields managed here:
;	tick: The getTickCount() value when this structure was last built.
;	hwnd, obj: The indicator's window handle and MSAA object.
; Fields managed elsewhere:
;	txt: The last result from typing__getText().
if typing___infoValid(Skype__typing___info)
	return Skype__typing___info
endIf
if !Skype__typing___info
	Skype__typing___info = new collection
endIf
var collection info = Skype__typing___info
var handle hChat = appwin__getSkypeWindow(appwin__SWT_ChatOutput)
if !hChat
	return
endIf
; Skype 7 before 7.2.
var handle hwnd = findWindow(hChat, "TWidgetControl", "")
if !hwnd
	; Skype 7.2.
	hwnd = findWindow(getParent(hChat), "TTypingIndicatorPanel", "")
	if !hwnd
		return
	endIf
endIf
info.hwnd = hwnd
info.obj = MSAA__objectFromWindow(hwnd)
if !info.obj
	return
endIf
info.tick = getTickCount()
return info
endFunction

int function typing___infoValid(collection info)
; Returns True if info is valid for current focus and False if not.
if !info || !info.hwnd
|| !info.tick || getTickCount() -info.tick > 1000
|| !getWindowStyleBits(info.hwnd)
|| !isWindowVisible(info.hwnd)
|| isWindowObscured(info.hwnd)
	return False
endIf
var object o = info.obj
if !o || o.accChildCount != 1
	return False
endIf
return True
endFunction

int function typing__getText(string byRef txt)
; Returns any current typing indicator text.
; Manages info.txt also.
var collection info = Skype__typing___info
if !info
	return False
endIf
var object o = info.obj
info.txt = o.accName(1)
info.txt = stringTrimLeadingBlanks(stringTrimTrailingBlanks(info.txt))
txt = info.txt
return True
endFunction

;============================== End module typing ==============================


;================================== Module lbl =================================
; Code for labeling Skype fields.
;
; Author: Doug Lee

void function lbl__fix(collection props, string sOrigin, handle hwnd, object o, int childID)
if hwnd && !o
&& getWindowClass(hwnd) == "TAccessibleRadioButtonEx"
	hwnd = getParent(hwnd)
	o = MSAA__objectFromWindow(hwnd)
	childID = 0
	if o.accParent.accRole(0) == Role_System_Grouping
		lbl__setProp(props, "Name", o.accName(childID), True)
		lbl__setProp(props, "ContainerName", o.accParent.accName(0), True)
	endIf
endIf
if !(o.accState(childID) & State_System_Focused)
	childID = lbl___getFocusedID(o)
endIf
; Default names and sometimes values/tutor messages for things.
; The If avoids double title speaking on focus switch into Skype.
if sOrigin != "hwnd" || hwnd != getTopLevelWindow(getFocus())
	lbl__nameByMSAATree(props, hwnd, o, childID)
endIf
if o.accRole(childID) == Role_System_Pane
&& !jcp__get(props, "Name")
	props.Name = lbl__msgUnknownField
	props.type = " "
elif o.accRole(childID) == Role_System_ComboBox
&& childID && o.accChild(childID).accChildCount == 0
	props.subtypeCode = WT_ButtonMenu
endIf

if lbl___nameSkypeProfileFields(hwnd, props)
	return
elif lbl___handleTechInfoList(hwnd, props)
	return
elif lbl___handleUserCheckList(hwnd, props)
	return
elif lbl___handleLive5Selector(props)
	return
elif lbl___fixEmptyListBox(props, hwnd)
	return
elif getWindowClass(hwnd) != "TChatRichEdit"
&& lbl___fixIncompleteFocus(props, hwnd)
	return
elif lbl___handleChangePasswordForm(props, hwnd)
	return
endIf

var string wClass = getWindowClass(hwnd)
if wClass == "TContactProfile"
|| wClass == "TConversationHeader"
	var object oProf = MSAA__objectFromWindow(hwnd)
	var object profID = oProf.accFocus
	if oProf.accRole(profID) == Role_System_StaticText
	&& stringIsBlank(oProf.accName(profID))
	&& oProf.accRole(profID+1) == Role_System_StaticText
	&& !stringIsBlank(oProf.accName(profID+1))
	&& !stringIsBlank(oProf.accDescription(profID+1))
		; For the mood message field for the highlighted contact.
		lbl__setProp(props, "Name", oProf.accName(profID+1), False)
		lbl__setProp(props, "Type", " ", False)
	endIf
endIf

var int stc = getWindowSubtypeCode(hwnd)
var string pwClass = getWindowClass(getParent(hwnd))
var string twClass = getWindowClass(getTopLevelWindow(hwnd))
var string name

if getWindowTypeCode(hwnd) == WT_Button && getWindowName(hwnd) == "btn"
	name = stringChopLeft(getWindowName(hwnd), 3)
	lbl__setProp(props, "Name", name, False)
elif wClass == "TConversationsControl"
	var object CLObj, int CLID, string CLName
	CLObj = sutil__getFocusObject(CLID)
	if CLObj.accRole(CLID) == Role_System_ListItem
		CLName = CLObj.accName(CLID)
		if !CLName
			CLName = MSAA__rectText(CLObj, CLID)
		endIf
		if !stringIsBlank(CLObj.accDescription(CLID))
			CLName = formatString("%1, %2", CLName, CLObj.accDescription(CLID))
		endIf
		if (CLObj.accState(CLID) & State_System_Selected)
		&& (!stringIsBlank(CLObj.accName(CLID)))
			CLName = cmsgSelected +" " +CLName
		endIf
		lbl__setProp(props, "Name", " ", True)
		if CLObj.accRole(CLID) == Role_System_ListItem
			lbl__setProp(props, "Type", " ", True)
			lbl__setProp(props, "Position", " ", True)
		endIf
		lbl__setProp(props, "Value", CLName, True)
		if clObj.accRole(0) == Role_System_List
			lbl__setProp(props, "ContainerName", clObj.accName(0), False)
		else
			lbl__setProp(props, "ContainerName", clObj.accParent.accName(0), False)
		endIf
	endIf  ; listItem
elif wClass == "TNavigableTntEdit" && pwClass == "TConversationsControl"
	lbl__setProp(props, "Name", lbl___msgContactFilter, False)
endIf

; This is applied generically but was aimed specifically at sound widget combo boxes in Skype 3.x active call tabs.
; PositionInGroup() is overridden in this file to use MSAA to get position info for these controls.
if wClass == "TNavigableTntComboBox"
	lbl__setProp(props, "Position", positionInGroup(), False)
endIf

if stringContains(stringLower(getWindowName(getTopLevelWindow(hwnd))), lbl__scEditSkypeOutContact)
&& stc == WT_Edit
	; SkypeOut Contact list entry > Context menu > Edit Contact
	lbl__setProp(props, "Name", lbl__getNameNear("up", ""), False)
elif stc == WT_Edit && pwClass == "TConversationsControl"
	; This distinguishes the contact rename edit from the contact filter edit.
	if hwnd == getFirstChild(getParent(hwnd))
		; Contact list entry > Context > Rename, or F2
		; This is to prevent JAWS from using the previous contact's name as a name for this edit field.
		lbl__setProp(props, "Name", " ", False)
	else
		; The (I think unnamed) box for filtering contacts by substring.
		lbl__setProp(props, "Name", lbl__msgSearchContacts, False)
	endIf
elif (wClass == "TChatRichEdit" || wClass == "TskRichEdit") && pwClass == "TChatEntryControl"
	; Name conversation input boxes after the other party.
	lbl__setProp(props, "Name", schat__getChatName(hwnd, True), True)
	lbl__setProp(props, "subtypeCode", WT_Multiline_Edit, True)
elif wClass == "TChatContentControl"
	var int otc = getObjectTypeCode()
	if (otc == WT_Button || otc == WT_Link)
		lbl__setProp(props, "ContainerName", " ", False)
		lbl__setProp(props, "ContainerType", " ", False)
	elif otc == WT_ListBoxItem
		var object oList = appwin__getMSAAObject(appwin__SWT_ChatOutput)
		var int focusedMessageID = schat__getMessageIndex(o)
		var int cnt = schat__getMessageCount(oList)
		if focusedMessageID
			jcp__setItemPos(props, focusedMessageID, cnt)
			lbl__setProp(props, "name", oList.accName(0), True)
			if opt__getUserOption(opt__SV_ChatMessageTimes) == 0
				var string msgText = schat__getMessageText(oList, focusedMessageID, "")
				lbl__setProp(props, "value", msgText, True)
			endIf
		endIf
	endIf
elif wClass == "TMyselfControl"
&& getObjectSubtypeCode() == WT_Link
	; Skype 6.7 (and maybe older/newer 6.x) "Me" link at least.
	var object oMe, int idMe
	oMe = getFocusObject(idMe)
	if oMe
		var string meTxt = MSAA__rectText(oMe, idMe)
		if meTxt
			lbl__setProp(props, "Value", meTxt, True)
		endIf
	endIf
elif stringContains(wClass, "Edit") && pwClass == "tMyselfControl"
	lbl__setProp(props, "Name", lbl__msgMoodText, False)
elif lbl___fieldIsLink(hwnd)
	lbl__setProp(props, "SubtypeCode", WT_Link, False)
elif stringContains(stringLower(twClass), "options")
	; One of the Tools > Options pages is showing.
	if getObjectSubtypeCode() == WT_RadioButton
	&& getWindowClass(hwnd) == "TAccessibleRadioButtonEx"
	&& getWindowClass(getParent(hwnd)) == "TAccessibleRichRadioButton"
		; New in Skype 6.7.
		lbl__setProp(props, "Name", getWindowText(getParent(hwnd), False), True)
		lbl__setProp(props, "ContainerName", getWindowTextEx(getParent(getParent(hwnd)), False, False), True)
		lbl__setProp(props, "Position", lbl___getOrdinalPosition(getParent(hwnd)), False)
	endIf
	var
		string pageName
	pageName = lbl__getActiveOptionsPageName()
	if wClass == "TOptionsPanel"
		; Make SayLine work right after landing on the tree,
		; and add the page group name.
		lbl__setProp(props, "Value", pageName, True)
		lbl__setProp(props, "Name", " ", True)
		; Tree, not tree item.
		lbl__setProp(props, "SubtypeCode", WT_TreeView, True)
		; Position is wrong because this tree doesn't work normally, so leave it out.
		lbl__setProp(props, "Position", " ", True)
		; States like "opened" are superfluous here, so leave them out also.
		lbl__setProp(props, "State", " ", True)
		; PageUp/Down are not standard for a tree, so explain them to the user.
		lbl__setProp(props, "customTutor", lbl__msgOptionsTreeTutor, True)
	elif pagename == "RingtonesTab"
		if stringContains(stringLower(wClass), "list")
		|| stringContains(stringLower(wClass), "combo")
			; Using Name can upset JAWS' normal handling of listBoxes.
			lbl__setProp(props, "beforeField", lbl__getNameNear("up", ""), False)
		endIf
	elif pageName == "tsCallForwarding"
		if getWindowTypeCode(getParent(hwnd)) == WT_ComboBox
			lbl__setProp(props, "Name", formatString(lbl__msgEnterPhone, lbl___getOrdinalPosition(hwnd)), False)
		endIf
	elif pageName == "tsHotkeys" || pageName == "HotkeysTab"
		if wClass == "TTntListView" || wClass == "TAccessibleListView"
			var object HKObj, int HKID
			HKObj = getFocusObject(HKID)
			lbl__setProp(props, "Name", " ", True)
			lbl__setProp(props, "Value", hkObj.accName(hkID), False)
			lbl__setProp(props, "SubtypeCode", getObjectSubtypeCode(False, 1), False)
			lbl__setProp(props, "Attributes", getControlAttributes() & (~Ctrl_Selected), False)
		endIf
	endIf
elif stringContains(stringLower(twClass), "callquality")
&& (!getObjectName() || stringStartsWith(getObjectName(), "*"))
&& !stc && stringContains(stringLower(wClass), "button")
	; Call Quality Feedback buttons (set of five).
	var string cqName
	var object oBtn, int btnID
	oBtn = getObjectFromEvent(hwnd, -4, 0, btnID)
	cqName = oBtn.accDescription(1)
	if !cqName
		cqName = lbl___callQualityButtonNameByPosition(hwnd)
	endIf
	if !cqName
		saveCursor() invisibleCursor() saveCursor()
		routeInvisibleToPC()
		nextWord()
		cqName = getWindowText(getCurrentWindow(), Read_Everything)
		string__replace(cqName, " ", ": ", "a")
	endIf
	lbl__setProp(props, "Name", cqName, True)
elif twClass == "TskSendContacts" || twClass == "TskStartConference" || twClass == "TskSMSAdd"
	; Contacts > Send Contacts, Tools > Create a Conference Call, Add To SMS.
	if stringContains(stringLower(wClass), "userlist")
		; Setting Name changes how JAWS says this from "listview <line> <pos>" to "listbox item selected <pos>."  We don't want this.
		; See jcp__say() for the handling of beforeField.
		lbl__setProp(props, "BeforeField", lbl__getNameNear("up", ""), False)
	elif stc == WT_Edit && (twClass == "TskStartConference" || twClass == "TskSMSAdd")
		lbl__setProp(props, "Name", lbl__getNameNear("up", ""), False)
	elif stc == WT_EditCombo
		; JAWS speaks this control fine until something gets put in it.
		; This would fix that but also causes JAWS to call it an Edit instead of an Edit Combo.
		;lbl__setProp(props, "Name", lbl__msgSendContacts, False)
	endIf
elif twClass == "TskSendSMS"
	if !stringCompare(wClass, "TRichViewEdit", False)
		lbl__setProp(props, "Name", lbl__msgSMSSend, False)
	endIf
elif twClass == "TskBlockUsers"
	if stringContains(stringLower(wClass), "list")
		lbl__setProp(props, "BeforeField", lbl__getNameNear("up", ""), False)
	endIf
elif pwClass == "tDialPad"
|| pwClass == "TLoginControl"
|| getWindowClass(getParent(getParent(hwnd))) == "TLoginControl"
	if getWindowTypeCode(hwnd) == WT_Edit || getWindowTypeCode(hwnd) == WT_ComboBox || getWindowTypeCode(hwnd) == WT_EditCombo
		lbl__setProp(props, "Name", lbl__getNameNear("up", ""), False)
	endIf
endIf
endFunction

void function lbl__setProp(collection props, string whichProp, variant val, int force)
; Property setter() that allows earlier property values to override later ones.
if !force && collectionItemExists(props, whichProp)
	return
endIf
props[whichProp] = val
endFunction

int function lbl__nameByMSAATree(collection props, handle hwnd, object o, int childID)
; Name fields by MSAA information.

; First deal with an anomaly found in Skype 6 profile inserts:
; Role_System_StaticText controls without State_System_Focused set,
; causing getFocusObject and accFocus to return the parent Pane object.
var int iRole = o.accRole(childID)
if iRole == Role_System_Pane
	var object o1, int childID1
	if !isPCCursor()
		saveCursor() PCCursor()
	endIf
	o1 = getObjectAtPoint(childID1, getCursorCol(), getCursorRow())
	if o1
		if !(o1.accState(childID1) & State_System_Focusable) && (
			o1.accRole(childID1) == Role_System_StaticText
			|| o1.accRole(childID1) == Role_System_Link
		)
			o = o1
			childID = childID1
			lbl__setProp(props, "SubtypeCode", WT_Link, True)
		endIf
	endIf
endIf

if lbl___nameFromMSAA(props, hwnd, o, childID)
	return True
endIf

; Noticed first in Skype beta build 6.3.0.529:
; Many fields' MSAA info can best be found by chasing .accFocus from
; the ObjID_Client object drawn from the window parenting the one in focus.
o = MSAA__objectFromWindow(getParent(hwnd))
if !o
	return False
endIf
childID = 0
var int safety = 8
var int found = False
var variant v = o.accFocus
while !found && safety && v
	if intToString(v)
		childID = v
		found = True
	else
		o = v
		childID = 0
		if o.accChildCount
			v = o.accFocus
		else
			found = True
		endIf
	endIf
	safety = safety -1
endWhile
if !found
	return False
endIf
return lbl___nameFromMSAA(props, hwnd, o, childID)
endFunction

int function lbl___nameFromMSAA(collection props, handle hwnd, object o, int childID)
; Name the given item via MSAA if possible.
; Helper for lbl__nameByMSAATree().
var int iRole = o.accRole(childID)
var string val
if iRole == Role_System_ListItem
|| iRole == Role_System_OutlineItem
|| iRole == Role_System_MenuItem
	val = o.accName(childID)
	if val
		lbl__setProp(props, "Value", val, False)
	endIf
	childID = 0
elif iRole == Role_System_Pane && !o.accChild(childID).accChildCount
	lbl__setProp(props, "SubtypeCode", WT_Link, False)
elif iRole == Role_System_StaticText
	lbl__setProp(props, "Type", " ", False)
endIf
val = o.accName(childID)
if !val
	val = o.accHelp(childID)
endIf
if !val
	return False
endIf
lbl__setProp(props, "Name", val, False)
val = o.accDescription(childID)
if val && stringLeft(val, 6) != "skype:"
	lbl__setProp(props, "customTutor", val, False)
endIf
return True
endFunction

int function lbl___nameSkypeProfileFields(handle hwnd, collection props)
; Name the fields in a Skype profile screen.
if getWindowClass(hwnd) != "TMyselfProfileControl"
	return False
endIf
var object o, int childID
o = sutil__getFocusObject(childID)
if !o
	return False
endIf
var int iRole = o.accRole(childID) +0
var int i, string name
if iRole == Role_System_Link
	; Name by previous text item if any.
	if o.accRole(childID-1) == Role_System_StaticText
		lbl__setProp(props, "Name", o.accName(childID-1), True)
		; Move the original name to the value.
		lbl__setProp(props, "Value", o.accName(childID), True)
		lbl__setProp(props, "customTutor", lbl___msgProfileFieldTutor, False)
		return True
	endIf
elif iRole == Role_System_ComboBox
	; These are really irregular combo boxes:
	; No descendants, and must be clicked or Spaced on to open.
	; They do have names, but sometimes incomplete ones and sometimes
	; better ones than their preceding text items.
	; This algorithm nicely names the three Birthdate fields,
	; includes the current time in the Time field name,
	; and avoids repetition when text and combo names mean the same thing.
	; [DGL, 2011-01-23, Skype 5.1.0.104]
	name = o.accName(childID)
	i = childID -1
	while childID-i < 5 && o.accRole(i) == Role_System_ComboBox
		i = i -1
	endWhile
	if o.accRole(i) == Role_System_StaticText
		var string textName = o.accName(i)
		if stringChopRight(textName, 1) != name
			name = textName +" " +name
			lbl__setProp(props, "Name", name, True)
		endIf
		; Let the users know how to use these unconventional things.
		lbl__setProp(props, "SubtypeCode", WT_ButtonDropDown, True)
		lbl__setProp(props, "customTutor", lbl___msgProfileFieldTutor, False)
		return True
	endIf
elif iRole == Role_System_StaticText
	; Applies to mood text.
	lbl__setProp(props, "customTutor", lbl___msgProfileFieldTutor, True)
elif iRole == Role_System_ButtonDropdown
	if !childID
		; Skype 5.1 returns these as full-blown Accessible Objects.
		; For naming, we need this object's position among the others.
		o = o.accParent
		childID = lbl___getFocusedID(o)
	endIf
	i = childID -1
	var int textSeen = False
	while i > 0 && (!textSeen || childID -i < 5 && o.accRole(i) == Role_System_StaticText)
		if o.accRole(i) == Role_System_StaticText
			textSeen = True
		endIf
		i = i -1
	endWhile
	i = i +1
	name = ""
	while i < childID
		if o.accRole(i) == Role_System_StaticText
			name = formatString("%1 %2", name, o.accName(i))
		endIf
		i = i +1
	endWhile
	name = stringChopLeft(name +" " +lbl___msgVisibility +":", 1)
	name = formatString("%1 %2", name, o.accName(childID))
	lbl__setProp(props, "Name", name, True)
	return True
endIf
return False
endFunction

int function lbl___getFocusedID(object o)
; Get and return the childID with State_System_Focused set.
var int childID
childID = o.accFocus
if childID && (o.accState(childID) & State_System_Focused)
	return childID
endIf
var int n = o.accChildCount
childID = 1
while childID <= n
	if o.accState(childID) & State_System_Focused
		return childID
	endIf
	childID = childID +1
endWhile
return 0
endFunction

string function lbl___callQualityButtonNameByPosition(handle hwnd)
; Name the five Call Quality dialog buttons.
; Returns the applicable name if found, or the null string otherwise.
; Tested in Skype 6.16.0.105.
var int top = getWindowTop(hwnd)
var int bottom = getWindowBottom(hwnd)
hwnd = getFirstWindow(hwnd)
while hwnd
	var int cy = (getWindowTop(hwnd) +getWindowBottom(hwnd)) /2
	if getWindowClass(hwnd) == "TRichLabel"
	&& cy > top
	&& cy < bottom
		var object o, int childID
		o = getObjectFromEvent(hwnd, 0, ObjID_Client, childID)
		var int n = o.accChildCount
		childID = 1
		var string buf = ""
		while childID <= n
			if buf
				buf = buf +"\10"
			endIf
			buf = formatString("%1%2", buf, o.accName(childID))
			childID = childID +1
		endWhile
		return buf
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return ""
endFunction

int function lbl___handleTechInfoList(handle hwnd, collection props)
; Handle the announcement of items in the Skype 5.1 Technical Call Info screen.
if getWindowClass(hwnd) != "TWidgetControl"
|| getWindowClass(getParent(hwnd)) != "TCallTechInfoForm"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
if !childID
	return False
endIf
var int childID0 = childID
var int isHeader = lbl__isTCIHeaderItem(o, childID)
var int n = o.accChildCount
var string txt = ""
childID = childID +1
while childID <= n && o.accRole(childID) == Role_System_StaticText
	txt = string__join(txt, " ", o.accName(childID))
	childID = childID +1
endWhile
if !txt
	return False
endIf
; It's not a tree but it sort of acts like one.
lbl__setProp(props, "subtypeCode", WT_TreeView, True)
; Block out the MSAA "name" that is more like an item type.
lbl__setProp(props, "Name", " ", True)
lbl__setProp(props, "value", txt, True)
var int isExpanded = False
if isHeader
	lbl__setProp(props, "level", 0, True)
	isExpanded = (childID < n && !lbl__isTCIHeaderItem(o, childID))
	if isExpanded
		lbl__setProp(props, "attributes", Ctrl_Expanded, True)
	else
		lbl__setProp(props, "attributes", Ctrl_Collapsed, True)
	endIf
else
	lbl__setProp(props, "level", 1, True)
endIf
if !lbl___setTCIPosInfo(props, o, childID0, isHeader, isExpanded)
	; Avoid saying things like "1 of 37" when we can't get it right.
lbl__setProp(props, "Position", " ", True)
	lbl__setProp(props, "position", " ", True)
endIf
return True
endFunction

int function lbl__isTCIHeaderItem(optional object o, int childID)
; Returns True if the focused or given item from a Technical Call Info list is a header.
if !o && !childID
	o = getFocusObject(childID)
endIf
if stringContains(o.accName(childID), "Separator")
	; That should not need localization.
	return True
endIf
return False
endFunction

int function lbl___setTCIPosInfo(collection props, object o, int childID, int isHeader, int isExpanded)
; Set position info for the given item.
var int pos = 0, int n = 0, int i = 0
if isExpanded  ; implies isHeader also
	; Count subitems under this header.
	n = lbl___countTCIItems(o, childID, 1)
	pos = 0
elif isHeader
	; Count headers.
	; This only works in JAWS 14 and later.
	if getJFWVersion() < 1400000
		return False
	endIf
	var object api = UIA__getScriptAPI()
	if !api
		return False
	endIf
	var object el = UIA__elementFromMSAA(o, 0)
	if !el
		return False
	endIf
	; 2 is TreeScope_Children.
	var collection iter = UIA__elementIterator(UIA__findAll(el, 2,
		"name 'ListSeparatorItem' name 'ListSeparatorTopItem' or"
	))
	if !iter
		return False
	endIf
	var object el0 = UIA__elementFromMSAA(o, childID)
	if !el0
		return False
	endIf
	n = iter.count
	i = 0
	while !pos && UIA__nextElement(iter)
		i = i +1
		el = iter.currentElement
		if api.compareElements(el, el0)
			pos = i
		endIf
	endWhile
	if !pos
		return False
	endIf
else  ; regular item
	; Position among descendants of the previous header.
	pos = lbl___countTCIItems(o, childID, -1) +1
	n = pos +lbl___countTCIItems(o, childID, 1)
endIf
if n
	jcp__setItemPos(props, pos, n)
	return True
endIf
return False
endFunction

int function lbl___countTCIItems(object o, int childID, int delta)
var int n = 0
var int nids = o.accChildCount
while True
	childID = childID +delta
	if childID < 1 || childID > nids
		return n
	; Text items are quietly skipped.
	elif o.accRole(childID) != Role_System_StaticText
		if lbl__isTCIHeaderItem(o, childID)
			return n
		endIf
		n = n +1
	endIf
endWhile
; Never reached.
endFunction

int function lbl___handleUserCheckList(handle hwnd, collection props)
; Handle the otherwise unreported checkbox states in the Send Contacts list.
; In Skhype 5.1: Conversations or Contacts entry, Context menu, Send Contacts;
; the list in that dialog.
if getWindowClass(hwnd) != "TCheckBoxUserListControl"
	return False
endIf
var object o, int childID
o = getFocusObject(childID)
if childID == 0
	; No item selected yet, at least according to JAWS.
	return False
endIf
; MSAA__rectText() should work here,
; but Gianluca discovered that it can fail to include the checkmark/corner for some reason.
var string txt = getTextBetween(getWindowLeft(hwnd), getWindowRight(hwnd))
lbl__setProp(props, "Value", txt, False)
lbl__setProp(props, "Name", " ", False)
return True
endFunction

int function lbl___handleLive5Selector(collection props)
; Handle the otherwise unreported checkbox states in the Live Call area Add People list.
var int stat = live5selector__selected()
if !stat
	return False
endIf
if stat == 1
	lbl__setProp(props, "Name", getObjectName(), False)
else
	lbl__setProp(props, "Name", " ", False)
	lbl__setProp(props, "Value", getObjectName(), False)
endIf
if stat == 1
	; Right list, not a contact entry.
	lbl__setProp(props, "customTutor", lbl___msgPressForMenu, False)
	return True
endIf
var int attribs = getControlAttributes()
if stat == 2
	attribs = attribs | Ctrl_Unchecked
else
	attribs = attribs | Ctrl_Checked
endIf
lbl__setProp(props, "Attributes", attribs, False)
return True
endFunction

int function lbl__getCheckStat(int offset)
var string sColor = ColorToRGBString(getColorAtPoint(getCursorCol()-offset, getCursorRow()))
; Black-on-white checkmarks (seen on Sarah's machine) (Skype 3.x though).
if sColor == "000000000"
	return Ctrl_Checked
elif sColor == "255255255"
	return Ctrl_Unchecked
endIf

; Blue checkmarks on gray background (my machine (DGL)).
var int r, int g, int b
r = stringToInt(stringLeft(sColor, 3))
g = stringToInt(substring(sColor, 4, 3))
b = stringToInt(substring(sColor, 7, 3))
if r-g == 0 && g-b == 0
	return Ctrl_Unchecked
elif b > r && b > g
	return Ctrl_Checked
endIf
return 0
endFunction

int function lbl___fixEmptyListBox(collection props, handle hwnd)
; Fix announcement of empty listboxes where getFocusObject returns an invalid childID of 1.
var object o = MSAA__objectFromWindow(hwnd)
if !o
	return False
elif o.accChildCount != 0
	return False
endIf
var int childID = o.accFocus
if childID != 1
	return False
endIf
; childID 1 and accChildCount 0.
lbl__setProp(props, "Origin", formatString("Point%1,%2",
	intToString(getCursorCol()),
	intToString(getCursorCol())
), False)
return True
endFunction

int function lbl___fixIncompleteFocus(collection props, handle hwnd)
; Fix announcement of controls where getFocusObject doesn't go far enough.
var object o, int childID, int childID1
o = getFocusObject(childID)
if !o
	return False
elif o.accChildCount == 0 || childID
	return False
endIf
childID1 = o.accFocus
if !childID1
	return False
endIf
o.accSelect(SelFlag_TakeFocus, 0)
o.accSelect(SelFlag_TakeFocus, childID1)
return True
endFunction

int function lbl___handleChangePasswordForm(collection props, handle hwnd)
; Handle fields in the Skype 5 Change Password form.
; Written against Skype 5.9.
if getWindowClass(getTopLevelWindow(hwnd)) != "TChangePasswordForm"
|| getObjectTypeCode() == WT_Button
	return False
endIf

; Name is to the left, hint is below.
var int left, int right, int top, int bottom
if !getWindowRect(hwnd, left, right, top, bottom)
	return False
endIf
saveCursor()  invisibleCursor()  saveCursor()
moveTo(left, (top+bottom)/2)
priorWord()
if getCursorRow() < top
	return False
endIf
var string name = getChunk()
moveTo(left+7, bottom+7)
var object o, int childID
o = getCurrentObject(childID)
if o.accRole(childID) == Role_System_StaticText
	name = formatString("%1 (%2)", name, o.accName(childID))
endIf
lbl__setProp(props, "Name", name, False)
return True
endFunction

string function lbl__getNameNear(string dir, string substr)
; Gets and returns the text (chunk) in the given direction from the current position as field name text.
; Dir can be left, right, up, or top.
; Top is special and uses the top line of the current window (for StringGrids like in Options > Hotkeys).
; If substr is given, the name is returned only if it contains substr.
var
	int x, int y,
	handle hwnd,
	int left, int right, int top, int bottom,
	int oldRestriction,
	int safetyCount,
	string name
hwnd = getCurrentWindow()
; X/Y used so we need not know which cursor is active.
x = getCursorCol()
y = getCursorRow()
top = getLineTop() -3
bottom = y+4
left = getWindowLeft(getTopLevelWindow(hwnd))
right = getWindowRight(getTopLevelWindow(hwnd))
saveCursor() invisibleCursor() saveCursor()
oldRestriction = getRestriction()
moveTo(x, y+2)
setRestrictionToRect(left, top, right, bottom)
if dir == "left"
	right = getCursorCol()
	; PriorChunk works for speech but not for Braille.  [DGL, 2007-03-01]
	priorWord()
	if getCurrentWindow() == hwnd
		moveTo(getWindowLeft(hwnd), getCursorRow())
		right = getCursorCol()
		priorWord()
	endIf
elif dir == "right"
	moveTo(getWindowRight(hwnd), getCursorRow())
	nextWord()
elif dir == "top"
	setRestriction(oldRestriction)
	moveToWindow(getCurrentWindow())
else  ; "up"
	setRestriction(oldRestriction)
	moveToWindow(getCurrentWindow())
	priorLine()
	if getWindowSubtypeCode(hwnd) == WT_RadioButton || getWindowSubtypeCode(hwnd) == WT_Checkbox
		safetyCount = 0
		while (stringIsBlank(getChunk()) || (getWindowSubtypeCode(getCurrentWindow()) == WT_RadioButton || getWindowSubtypeCode(getCurrentWindow()) == WT_CheckBox)) && safetyCount < 10
			priorLine()
			safetyCount = safetyCount +1
		endWhile
		if safetyCount >= 10
			log__write("Loop ran out.")
		endIf
	endIf
endIf
if getWindowSubtypeCode(hwnd) == WT_RadioButton || getWindowSubtypeCode(hwnd) == WT_Checkbox
	if dir != "top" && (getCurrentWindow() != getParent(hwnd) && getCurrentWindow() != getParent(getParent(hwnd)))
		; Avoid using another control as a label for this one.
		; Combos end up two levels below the containing window, hence the second condition above.
		setRestriction(oldRestriction)
		return ""
	endIf
endIf
var int right1
; In some non-English JAWS versions, hjconst.jsh is missing const IT_Chunk=1.
if !getItemRect(getCursorCol(), getCursorRow(), left, right1, top, bottom, 1)
	return ""
endIf
right = min(right, right1)
setRestrictionToRect(left, top, right, bottom)
;name = getTextInRect(left, top, right, bottom)
name = getChunk()
setRestriction(oldRestriction)
if !stringLength(substr)
	return name
elif stringContains(substr, "|")
	if stringContains("|"+stringLower(substr)+"|", "|"+stringLower(name)+"|")
		return name
	endIf
else
	if stringContains(stringLower(name), stringLower(substr))
	return name
	endIf
endIf
return ""
endFunction

int function lbl___fieldIsLink(handle hwnd)
if stringCompare(getWindowClass(hwnd), "TRichView", False)
	return False
endIf
if getCurrentWindow() != hwnd || getCurrentWindow() != getWindowAtPoint(getCursorCol(), getCursorRow())
	; The first condition covers calls with hwnd being a non-current window.
	; The second condition has been seen on Options > General's first link and Options > Call Forwarding & Voicemail > Basic/Advanced Settings link:
	; The PC cursor can lag behind JAWS' idea of the focused and current window by up to a couple of seconds.
	; [DGL, 2007-03-04, JAWS 8.0]
	saveCursor() invisibleCursor() saveCursor()
	moveToWindow(hwnd)
endIf
return (getCharacterAttributes() & Attrib_Underline)
endFunction

string function lbl__getActiveOptionsPageName()
var
	handle hwnd
hwnd = getFocus()
if getWindowClass(hwnd) == "TOptionsPanel"
	var object o, int childID
	o = getObjectFromEvent(hwnd, ObjID_Client, 0, childID)
	if o
		var int i, string name
		i = o.accFocus +0
		name = o.accName(i)
		while i > 1
			i = i -1
			if o.accValue(i) == 0
				name = formatString("%1, %2", o.accName(i), name)
				return name
			endIf
		endWhile
		return name
	endIf
endIf
while hwnd
	if stringContains(stringLower(getWindowClass(hwnd)), "tabsheet")
		return getWindowName(hwnd)
	endIf
	hwnd = getParent(hwnd)
endWhile
return ""
endFunction

string function lbl___getOrdinalPosition(handle hwnd)
; Gets the ordinal position of hwnd in the set of windows at its level with the same subtype code.
; The degenerate case "1 of 1" returns null here.
; Counts from the end because windows seem to be created in reverse in this application.
var
	int stc,
	handle h1,
	int pos, int count
if getWindowTypeCode(getParent(hwnd)) == WT_ComboBox
	hwnd = getParent(hwnd)
endIf
stc = getWindowSubtypeCode(hwnd)
h1 = getLastWindow(hwnd)
count = 0
pos = 0
while h1
	if h1 == hwnd || (isWindowVisible(h1) && getWindowSubtypeCode(h1) == stc)
		count = count +1
	endIf
	if h1 == hwnd
		pos = count
	endIf
	h1 = getPriorWindow(h1)
endWhile
if !count || !pos || (count == 1 && pos == 1)
	return ""
endIf
return formatString(cmsgMyPosInGroup1, intToString(pos), intToString(count))
endFunction

string function lbl___getTextWithAttributesFromWindow(handle hwnd, int attribs)
; Return the text of the given attribute or attribute combination from the given window.
var
	int left, int top, int right, int bottom
if !getWindowRect(hwnd, left, right, top, bottom)
	return ""
endIf
return getTextInRect(left, top, right, bottom, attribs)
endFunction

;================================ End module lbl ===============================


;=============================== Module incsearch ==============================
; Code to allow incremental search in listboxes that don't allow it,
; including JAWS pop-up lists. Also makes Home/End work where necessary.
;
; Usage: When a control needing this support is in focus,
; call incsearch__gotoNext(strKeyName) from KeyPressedEvent.
; Call incsearch__setFlags() from autoStartEvent if necessary.
;
; Author:  Doug Lee of Level Access

const
; These are not defined by older JAWS versions but are by newer ones,
; so to avoid compilation problems in all cases, we define local versions here.
	incsearch___SelFlag_TakeFocus = 1,
	incsearch___SelFlag_TakeSelection = 2

globals
; The search string entered so far at any given time.
	string Skype__incsearch___search,
; Flags that can be set for this system. See incsearch__setFlags() for details.
	string Skype__incsearch___flags

void function incsearch__setFlags(string flags)
; Set flags that govern how this system works. Flags defined:
;	l: Use ListView focus instead of MSAA focus to focus an item.
;	h: Implement Home/End navigation (for lists that don't already).
; Must call from autoStartEvent if any special flags are needed.
; Call from autoFinishEvent also to clear flags if any are set and
; this module is not given a unique namespace.
Skype__incsearch___flags = stringLower(flags)
return Skype__incsearch___flags
endFunction

void function incsearch___clear()
; Ends an existing search.
Skype__incsearch___search = ""
endFunction

void function incsearch__gotoNext(string keyname)
; Implements the incremental search feature using MSAA.
; Also implements Home=top and End=bottom navigations.
var
	object o, int childID,
	int i, int n, int n1
var int nav = 0
if stringLength(keyname) > 1
	; End searches when a non-alphanumeric key (e.g., Tab or Ctrl+A) is typed.
	incsearch___clear()
	if stringCompare(keyName, "home", False) == 0
		nav = 1
	elif stringCompare(keyName, "end", False) == 0
		nav = 2
	endIf
	if nav && stringContains(Skype__incsearch___flags, "h")
		PCCursor()
		o = getObjectAtPoint(childID, getCursorCol(), getCursorRow())
		n = o.accChildCount
		if nav == 1
			incsearch___focusItem(o, 1)
		else  ; 2
			incsearch___focusItem(o, n)
		endIf
	endIf
	return
endIf

; If the key typed is not the same as the last key typed,
; or there is already a multi-character search underway,
; add the new key to the search.
if stringCompare(keyname, Skype__incsearch___search, False)
	Skype__incsearch___search = stringChopLeft(" " +Skype__incsearch___search +keyname, 1)  ; stringChopLeft avoids arithmetic addition
endIf
scheduleFunction("incsearch___clear", 10)  ; listViews seem to allow about 1 second too.

; Search for the string entered thus far, starting at the current list item.
PCCursor()
o = getObjectAtPoint(childID, getCursorCol(), getCursorRow())
n = o.accChildCount
if stringLength(Skype__incsearch___search) == 1
	if inHJDialog()
		; HJ dialogs already do single-key searching, so don't mess with that.
		return
	endIf
	; For a single-key search, start on the next item.
	i = childID +1
else
	; For an already multi-key search, allow the current item to continue matching.
	i = childID
endIf
n1 = 0
while n1 <= n
	; Search all items once, wrapping as necessary.
	if i > n
		i = 1
	endIf
	if stringStartsWith(o.accName(i), Skype__incsearch___search)
		incsearch___focusItem(o, i)
		return
	endIf
	i = i +1
	n1 = n1 +1
endWhile
; Beep to indicate the search failed.
beep()
endFunction

void function incsearch___focusItem(object o, int childID)
; Select/focus the indicated item.
if childID == o.accFocus
	; Repeat this item to let the user know we're still here.
	scheduleFunction("sayObjectActiveItem", 0)
else
	; Select/focus the found item.
	if inHJDialog()
		o.accSelect(incsearch___SelFlag_TakeSelection, childID)
	elif stringContains(Skype__incsearch___flags, "l")
		setCurrentItem(getFocus(), childID)
	else
		o.accSelect(incsearch___SelFlag_TakeFocus+incsearch___SelFlag_TakeSelection, childID)
	endIf
			refreshWindow(getFocus())
		endIf
endFunction

;============================= End module incsearch ============================


;================================ Module saysel ================================
; Code to make the SaySelectedText script work better in some MSAA-capable list controls.
;
; Author: Doug Lee

globals
	int Skype__saysel___vpcFlag

string function getSelectedText(optional int bWantMarkUp, int bWantAllListViewItemText)
; Override to make saySelectedText work.
var
	handle hwnd,
	object o, int childID, int n,
	int tc,
	string buf
if !Skype__saysel___vpcFlag
	return getSelectedText(bWantMarkUp, bWantAllListViewItemText)
endIf
hwnd = getFocus()
tc = getWindowTypeCode(hwnd)
if tc == WT_Edit || tc == WT_MultiLine_Edit ||tc == WT_ComboBox
|| !sutil__isMSAASupport(hwnd)
	return getSelectedText(bWantMarkUp, bWantAllListViewItemText)
endIf
o = getObjectFromEvent(hwnd, -4, 0, childID)
childID = 0
n = o.accChildCount
while childID < n
	childID = childID +1
	if o.accState(childID) & State_System_Selected
		buf = string__join(buf, "\r\b", string__ref(o.accName(childID)))
	endIf
endWhile
return buf
endFunction

int function isVirtualPCCursor()
if Skype__saysel___vpcFlag
	return True
endIf
return isVirtualPCCursor()
endFunction

script saySelectedText()
Skype__saysel___vpcFlag = True
performScript saySelectedText()
Skype__saysel___vpcFlag = False
endScript

;============================== End module saysel ==============================


;================================== Module nav =================================
; Code for moving among Skype control groups.

script NextDocumentWindow()
util__waitForKeysToClear(15)
nav__switchDocumentWindows(True)
endScript

script PreviousDocumentWindow()
util__waitForKeysToClear(15)
nav__switchDocumentWindows(False)
endScript

script nextControlGroup()
; Switches to the next major control group in the Skype main window or in the currently focused conversation.
util__waitForKeysToClear(15)
nav__switchControlGroups(True)
endScript

script PreviousControlGroup()
; Switches to the previous major control group in the Skype main window or in the currently focused conversation.
util__waitForKeysToClear(15)
nav__switchControlGroups(False)
endScript

void function nav__switchDocumentWindows(int isNext)
; Implements round-robin conversation window switching.
var
	int doChats
if util__isSpecialFocus()
	doChats = False
elif schat__isChatWindow(getFocus(), False)
	doChats = True
elif appwin__isMainSkypeWindow(getTopLevelWindow(getFocus()))
	; Allows Ctrl+[Shift+]Tab to jump from main window to a conversation.
	doChats = True
else
	doChats = False
endIf

if !doChats
	if isNext
		performScript nextDocumentWindow()
	else
		performScript previousDocumentWindow()
	endIf
	return
endIf

if isNext
	isNext = 1
else
	isNext = -1
endIf
schat__switchToChat(False, isNext)
endFunction

void function nav__switchControlGroups(int isNext)
; Implements round-robin switching among control groups.
if util__isSpecialFocus()
	; Do nothing.
	beep()
	return
endIf
speechOff()
; If the chat entry box is focused, move focus to its parent first.
; Otherwise, Shift+F6 (previous group) gets stuck in that box.
var handle hwnd = getFocus()
if getWindowClass(getParent(hwnd)) == "TChatEntryControl"
	setFocus(getParent(hwnd))
	pause()
endIf
var int done = False
var int safety
while !done
	hwnd = getFocus()
	if isNext
		util__setFocus(0, "n")
	else
		util__setFocus(0, "p")
	endIf
	safety = 10
	while safety && getFocus() == hwnd
		pause()
		safety = safety -1
	endWhile
	hwnd = getFocus()
	if nav___isFocusWorthy(hwnd)
		done = True
	else
		stopSpeech()
	endIf
endWhile
if getWindowClass(hwnd) == "TChatEntryControl"
	; Focus the edit part of this one.
	var handle hwnd1
	hwnd1 = findWindow(hwnd, "TChatRichEdit", "")
	if !hwnd1
		hwnd1 = findWindow(hwnd, "TChatRichEdit.UnicodeClass", "")
	endIf
	if hwnd1
		setFocus(hwnd1)
		pause()
	else
		beep()
	endIf
else
	; Speak a name when possible and make sure focus is somewhere sensible.
	var object o = MSAA__objectFromWindow(hwnd)
	if o
		speechOn()
		sayMessage(OT_Dialog_Name, o.accName(0))
		speechOff()
		var variant v = o.accFocus
		if o == v || v == 0
			var int i = 1
			var int n = o.accChildCount
			while i <= n
				if o.accState(i) & State_System_Focusable
					o.accSelect(SelFlag_TakeFocus, i)
					pause()
					i = n
				endIf
				i = i +1
			endWhile
		endIf
	endIf
endIf
speechOn()
scheduleFunction("sayObjectTypeAndText", 1)
endFunction

int function nav___isFocusWorthy(handle hwnd)
; Returns True if hwnd appears to be a window worthy of receiving focus.
if getWindowRight(hwnd) -getWindowLeft(hwnd) < 5
|| getWindowBottom(hwnd) -getWindowTop(hwnd) < 5
	return False
endIf
return True
endFunction

;================================ End module nav ===============================


;================================ Module sayall ================================
; Support for SayAll in Skype chat message lists.
;
; Author: Doug Lee

globals
; Info used by the SayAll code in this module while a SayAll is in progress.
	collection Skype__sayall___data

void function sayAllStarted()
; This function is called by the built in script function SayAll just before
; the Speak method of the SAYALL object is called for the first time.  It is
; intended that this function be overridden in a given application's scripts
; to provide a method of performing actions (such as setting JCF options)
; just before a Say All is started.  This complements the preexisting
; SayAllStoppedEvent  which is called whenever a Say All process is stopped.
; (This text was drawn from default.jss in JAWS 11-15.)
var variant null
var handle hwnd = getFocus()
if getWindowClass(hwnd) != "TChatContentControl"
	Skype__sayall___data = null
	return
endIf
Skype__sayall___data = new collection
var collection data = Skype__sayall___data
data.hwnd = hwnd
var object oMsg, int childID
oMsg = getFocusObject(childID)
if childID
	; Unexpected; oMsg should be a message object.
	Skype__sayall___data = null
	return
endIf
Skype__sayall___data.msgs = oMsg.accParent
data.index = schat__getMessageIndex(oMsg)
if !data.index
	; If we don't know where to start, we don't know where to go.
	Skype__sayall___data = null
	return
endIf
data.count = schat__getMessageCount(data.msgs)
; That count may increase as messages are read.
data.sayAllBySetting = getJCFOption(Opt_Say_All_Mode)
data.linePauseSetting = getJCFOption(Opt_Line_Pauses)
setJCFOption(Opt_Line_Pauses, 1)
setJCFOption(Opt_Say_All_Mode, SaSayAllByLine)
endFunction

int function SayAllSpeakUnit(int iSayAllMode, int iSpeakNext)
if !Skype__sayall___data
	; This is not our area, so let internal code handle this.
	return SayAllSpeakUnit(iSayAllMode, iSpeakNext)
endIf
var collection data = Skype__sayall___data
if iSpeakNext
	data.index = data.index +1
	if data.index > data.count
		; In case messages came in during reading.
		data.count = schat__getMessageCount(data.msgs)
	endIf
	if data.index > data.count
		return saSayAllEndOfDocument
	endIf
	; This should not be necessary but is; see SayAllRelocateCursor().
	typeKey("downArrow")
endIf
sayall___speakMessage()
return saSayAllUnitSpoken
endFunction

void function SayAllRelocateCursor(int iSayAllMode,int iUnitsToRetreat,int iWordsSpokenInCurrentUnit)
if !Skype__sayall___data
	; This is not our area, so let internal code handle this.
	return SayAllRelocateCursor(iSayAllMode,iUnitsToRetreat,iWordsSpokenInCurrentUnit)
endIf
var collection data = Skype__sayall___data
if iUnitsToRetreat
	data.index = data.index -iUnitsToRetreat
	; Focusing this should be enough, but even though the focus takes,
	; the next Up/Down moves relative to the previous focus.
	; Selecting does not seem to help.
	; On high-DPI machines, clicking is not safe either.
	; We therefore have to send arrows to move focus.
	; [DGL, 2014-09-14, Skype 6.18.59.106.]
	speechOff()
	var int i = iUnitsToRetreat
	while i
		typeKey("upArrow")
		i = i(-1)
	endWhile
	pause()
	speechOn()
endIf
var object msgs = data.msgs
;msgs.accSelect(SelFlag_TakeFocus, data.index)
; speak message on which cursor lands when SayAll is stopped by user.
sayall___speakMessage()
;sayInteger(iUnitsToRetreat)
;sayInteger(iWordsSpokenInCurrentUnit)
endFunction

void function SayAllStoppedEvent()
var variant null
setJCFOption(Opt_Line_Pauses, Skype__sayall___data.linePauseSetting)
setJCFOption(Opt_Say_All_Mode, Skype__sayall___data.sayAllBySetting)
Skype__sayall___data = null
sayAllStoppedEvent()
endFunction

void function sayall___speakMessage()
var collection data = Skype__sayall___data
; This honors timestamp user setting.
var string txt = schat__getMessageText(data.msgs, data.index, "a")
sayMessage(OT_SayAll, txt)
endFunction

;============================== End module sayall ==============================


;======================== Module sutil (from skypeUtil) ========================
; Utility code for the Skype scripts.
;
; Author:  Doug Lee

object function sutil__getFocusObject(int byRef childID)
; A GetFocusObject wrapper that accounts for Skype's sometimes returning the wrong thing.
var
	object o,
	variant v
o = getFocusObject(childID)
v = o.accFocus
if !stringLength(v) && v
	o = v
	childID = 0
elif v
	childID = v +0
endIf
return o
endFunction

int function sutil__isSkypeWatchLoaded()
; Returns True if SkypeWatch is running:
;	- 1 if SkypeWatch is running via the old (Use "SkypeWatch.jsb" in default.jss) installation method.
;	- 2 if SkypeWatch is running via the new method (default.jss is loader of SkypeWatch and default_fs.jsb).
; The two are distinguished from one another by whether SkypeWatche's WindowCreatedEvent has been called since the last SkypeWatch restart.
; SkypeWatch_lastTick is updated by SkypeWatch::SkypeWatcher(), which runs once every five seconds and once two seconds after SkypeWatch first loads.
; If it hasn't been updated in 15 seconds, SkypeWatch is surely not around to do it.
; For some reason though, the delay has been seen to exceed 8 seconds. [DGL, 2005-02-08]
var
	int diff
diff = getTickCount() -SkypeWatch_lastTick
;sayInteger(diff)
if diff < 15000
	if SkypeWatch_WCECalled
		return 2
	endIf
	return 1
endIf
return 0
endFunction

globals int Skype__sutil___saidUnsupported
void function sutil__sayUnsupportedSkype()
; Announce once that this Skype version is not supported by these scripts.
; A beep is produced on every call regardless of maxTimes.
if !Skype__sutil___saidUnsupported
	sayMessage(OT_Error, sutil___msgNotSupported)
endIf
if Skype__sutil___saidUnsupported < 4
	beep()
endIf
Skype__sutil___saidUnsupported = Skype__sutil___saidUnsupported +1
endFunction

void function sutil__sayMessageToUserBuffer(string s, int format)
var
	int ot
s = s +"\13\10\13\10" +cMsgBuffExit
ot = OT_User_Buffer
if format
	sayFormattedMessage(ot, s)
else
	sayMessage(ot, s)
endIf
endFunction

void function sutil__sayURMessage(string sLong, string sShort, int format)
; Similar to sayMessageToUserBuffer:  Allows JAWS 4.x/5.x/6.x to do things the same way.
; A message sent this way will flash in Braille in JAWS 6+.
var
	int wasSpeechOff,
	int ot
wasspeechOff = isSpeechOff()
if wasSpeechOff
	speechOn()
endIf
ot = OT_User_Requested_Information
if sShort
	if format
		sayFormattedMessage(ot, sLong, sShort)
	else
		sayMessage(ot, sLong, sShort)
	endIf
else
	if format
		sayFormattedMessage(ot, sLong)
	else
		sayMessage(ot, sLong)
	endIf
endIf
if wasSpeechOff
	speechOff()
endIf
endFunction

int function sutil__isMSAASupport(handle hwnd)
; Determines whether the given window provides MSAA support.
var
	object o, int childID, int iRole
o = getObjectFromEvent(hwnd, -4, 0, childID)
if !o
	return False
endIf
iRole = o.accRole(childID)
if iRole == Role_System_Client || iRole == Role_System_Window
	return False
endIf
return True
endFunction

string function sutil__getToolBarItemDesc(string givenName, object o, int childId, int includeSeps)
; Construct and return the displayable name/description for the indicated toolbar item.
; GivenName, if non-null, will be used as the name instead of whatever is found on screen or in MSAA.
; As a special case, if givenName contains a slash and the button is a check box,
; the part left of the slash will be used as the name if the box is unchecked, and the right part if checked.
; includeSeps determines whether separators will return something meaningful or a null string.
; Author:  Doug Lee
var
	int iRole, int iState,
	int left, int right, int top, int bottom,
	string name, string states,
	int altnames
iRole = o.accRole(childId)
if iRole == Role_System_Window
	return ""
elif iRole == Role_System_Separator
	if includeSeps
		if givenName
			return givenName
		else
			return sutil__msgSep
		endIf
	else
		return ""
	endIf
endIf
states = ""
name = givenName
if !name
	name = o.accName(childId)
endIf
if !name
	; Example:  WordPad:  No accName but labeling graphics puts names in the object's rectangle.
	if MSAA__rect(o, childId, left, right, top, bottom)
		name = getTextInRect(left, top, right, bottom)
	endIf
endIf
if !name
	; TODO:  When separators are skipped above, item numbers here will skip also.
	name = formatString(sutil__msgUnnamedItem, intToString(childId))
	states = states +" " +sutil__msgUnnamed
endIf
name = name +" " +getRoleText(iRole)
iState = o.accState(childId)
if iState & State_System_Unavailable
	states = states + " " +cMsgNotavailable
endIf
if iState & State_System_OffScreen
	states = states + " " +sutil__msgOffScreen
elif iState & State_System_Invisible
	states = states + " " +sutil__msgInvisible
endIf
if iRole == Role_System_CheckButton || iRole == Role_System_RadioButton
	if iRole == Role_System_CheckButton
		if stringContains(name, "/")
			; When given alternate names, we'll treat it like a toggle button, not a check box.
			altnames = True
		else
			states = " " +sutil__msgCheckBox +states
		endIf
	elif iRole == Role_System_CheckButton
		states = " " +sutil__msgRadioButton +states
	endIf
	if iState & State_System_Checked
		if altnames
			name = stringSegment(name, "/", 2)
		else
			states = states +" " +cMsg_Checked
		endIf
	else
		if altnames
			name = stringSegment(name, "/", 1)
		else
			states = states +" " +cMsg_NotChecked
		endIf
	endIf
elif iRole == Role_System_Separator
endIf
if states
	name = name +" -" +states
endIf
return name
endFunction

string function sutil__myGetWindowText(handle hwnd)
; Returns the text in the given window with line breaks and without graphics.
var
	int oldG,
	string buf
oldG = getJCFOption(Opt_Include_Graphics)
setJCFOption(Opt_Include_Graphics, 0)
buf = GetTextInRect (getWindowLeft(hwnd), getWindowTop(hwnd), getWindowRight(hwnd), getWindowBottom(hwnd), 0, IgnoreColor, IgnoreColor, True)
setJCFOption(Opt_Include_Graphics, oldG)
return buf
endFunction

int function sutil__versionCompare(string vcode)
; Return an integer indicating how the actual Skype version compares to the given one:
; -1 means Skype is older, 0 means they're equal, and 1 means Skype is newer.
; -2 means the real Skype version code is not known.
; vcode: A string Skype version, like 5.10.
if stringLeft(vcode, 1) != "v"
	vcode = formatString("v%1", vcode)
endIf
var string real = sutil__getSkypeVersionCode(False)
if !real
	return -2
elif stringCompare(real, vcode, True) == 0
	return 0
endIf
vcode = stringChopLeft(vcode, 1)
real = stringChopLeft(real, 1)
var int n = max(
	string__segmentCount(real, "."),
	string__segmentCount(vcode, ".")
)
var int i = 1
var string vseg, string rseg, int vint, int rint
while i <= n
	vseg = stringSegment(vcode, ".", i)
	if !stringLength(vseg)
		; This means versionCompare("5.9") returns equal for Skype 5.9.x.
		return 0
	endIf
	rseg = stringSegment(real, ".", i)
	if !stringLength(rseg)
		; versionCompare("5.9.1.234") returns -1 if Skype's version code only comes back "5.9."
		return -1
	endIf
	vint = stringToInt(vseg)
	rint = stringToInt(rseg)
	if vint > rint
		return -1
	elif vint < rint
		return 1
	; Else equal, so keep looking.
	endIf
	i = i +1
endWhile
; And if the loop ran out, all is equal.
return 0
endFunction

globals
	int Skype__sutil___vcLastTicks,
	string Skype__sutil___vcLastVer
string function sutil__getSkypeVersionCode(int force)
; Returns a code that indicates what Skype version is running on this machine.
; Codes returned will be exact (if possible) or approximate version numbers of Skype.exe.
; Approximations will always be greater than or equal to the actual version number of skype.exe and will be suffixed with "-" to indicate that they are approximations.
; Returns the null string if the version code can't be obtained or approximated.
; The returned code is prefixed with "v" because this makes the < and > operators behave reasonably enough with version codes.
; WARNING:  "Reasonably enough" means functional but still odd: "v1.3" == "v1.3.2"; == is apparently tried in JAWS before < and >.
; Note: JAWS' GetVersionInfoStrings function is unable to return the real version code from some Skype versions.
; In such a case, the Scripting.FileSystemObject is used to get it.
; Note:  In JAWS versions older than 7.10, the full version code can't be obtained except when Skype is in focus.
; This code is also used by SkypeWatch, so that's not practical in all cases.
; The plan is to get the real code when possible and cache it wisely (I hope).
var
	handle hwnd,
	int ticks, int tickDiff,
	string appPath,
	object o

; Only actually check for the version every 10 seconds, to minimize work.
; This function can be called very frequently (e.g., if used from FindCaret), so this removes the need to worry about that.
ticks = getTickCount()
tickDiff = ticks -Skype__sutil___vcLastTicks
if !force && tickDiff <= 10000 && stringLength(Skype__sutil___vcLastVer)
	return Skype__sutil___vcLastVer
endIf
Skype__sutil___vcLastTicks = ticks

; If possible, get the application path, which allows us to get the real version number.
; This requires some part of Skype to be in focus.
appPath = ""
appPath = string__ref(getWindowOwner(appwin__getSkypeWindow(appwin__SWT_Main)))
if !appPath
	appPath = getAppFilePath()
	if !stringContains(stringLower(appPath), "skype.exe")
		appPath = ""
	endIf
endIf

; If we have a path, use it.
if appPath
	; First try JAWS' way, since it seems to work in later Skypes again.
	Skype__sutil___vcLastVer = getVersionInfoString(appPath, "FileVersion")
	if stringLength(Skype__sutil___vcLastVer) > 5
		Skype__sutil___vcLastVer = "v" +Skype__sutil___vcLastVer
		return Skype__sutil___vcLastVer
	endIf
	; Get the actual version code from the fileSystemObject if it feels like cooperating.
	o = util__getFSObject()
	if o
		Skype__sutil___vcLastVer = o.getFileVersion(appPath)
		if stringLength(Skype__sutil___vcLastVer)
			Skype__sutil___vcLastVer = "v" +Skype__sutil___vcLastVer
			return Skype__sutil___vcLastVer
		endIf
	endIf
endIf

; Favor the newest Skype available when we can't tell what it really is.
; Also force an update until we get a better answer though.
Skype__sutil___vcLastVer = "v999-"
Skype__sutil___vcLastTicks = 0
return Skype__sutil___vcLastVer
endFunction

string function sutil__getSkypeUserName()
; Returns the Skype ID of the current Skype user.
var
	handle hwnd,
	int pos,
	string buf
hwnd = appwin__getSkypeWindow(appwin__SWT_Main)
if !hwnd
	return ""
endIf
buf = getWindowName(hwnd)
pos = stringContainsFromRight(buf, " - ")
if !buf
	return ""
endIf
buf = stringChopLeft(buf, pos+2)
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(buf))
endFunction

string function sutil__gotoSkypeScriptPage(string page, int deferred)
; "Runs" or returns the run command for the given page.
; Returns the null string if deferred is False and the command to run if deferred is True.
; page: The page to launch.
; deferred: False to run, True to return the command to be run as a string.
var
	string strURLString,
	string strBrailleDriver, string strBrailleDisplay, string strBraillePort,
	string strSkypeProgramPath,
	string strSkypeProgramDate

if deferred
	var string cmd
	cmd = "sutil__gotoSkypeScriptPage(\"" +page +"\", 0)"
	return cmd
endIf

strSkypeProgramPath = getAppFilePath()
strSkypeProgramDate = IntToString(GetFileDate (strSkypeProgramPath))
BrailleGetDeviceInfo(strBrailleDriver, strBrailleDisplay, strBraillePort)

strURLString = sutil___url +page
+"?JFWVersion="+IntToString(getJFWVersion())
+"&ScriptVersion="+sutil__Script_Version
+"&ScriptRevision="+sutil__Script_Revision
+"&BrailleDisplay="+strBrailleDisplay
+"&SkypeProgramDate="+strSkypeProgramDate
+"&SkypeProgramVersion=" +stringChopLeft(sutil__getSkypeVersionCode(True), 1)
+"&WindowsOSVersion="+IntToString(GetWindowsOS())
+"&SkypeWatch="+intToString(sutil__isSkypeWatchLoaded())
+"&SkypeUserName="+sutil__getSkypeUserName()
strURLString = stringReplaceSubstrings(strURLString, " ", "%20")

if UserBufferIsActive () UserBufferDeactivate() EndIf
SayFormattedMessage (OT_Message, sutil__msgSkypeScriptsHomePage_L, sutil__msgSkypeScriptsHomePage_S)
Run("\"" +strURLString +"\"")
endFunction

int function sutil___splitContactInfo(string buffer, string byRef name, string byRef id, string byRef stat)
; Returns by reference the contact name, Skype ID, and status from the given buffer.
; The buffer should have come directly from MSAA's AccHelp property.
; Buffer format: <contactName>[ (<SkypeID>)] - <statusInfo>
; Returns True on success and False on failure.
; Author:  Doug Lee
var
	int pos

if stringIsBlank(buffer)
	; Happens on contact list entries for multichats.
	beep()
	return False
endIf

; Start by removing and collecting the status text (Online, Offline and last-seen time, etc.).
pos = stringContainsFromRight(buffer, "/>")  ; should always be found
if !pos
	return False
endIf
stat = stringChopLeft(buffer, pos+3)
buffer = stringLeft(buffer, stringContainsFromRight(buffer, "<")-2)
; Current buffer format: <contactName>[ (<SkypeID>)].
buffer = stringSegment(buffer, "\13", 1)

; Find out if there's a second name, and set name and id accordingly.
; If there's not, all we have is the Skype ID, which can't contain parentheses.
; If the display name differs from the Skype ID, we'll have both here.
pos = stringContainsFromRight(buffer, "(")  ; only found if second name is present
if pos
	name = stringLeft(buffer, pos-2)
	id = substring(buffer, pos+1, stringLength(buffer)-pos-1)
else
	name = buffer
	id = buffer
endIf
return True
endFunction

int function sutil___isPhoneNumber(string s)
; Returns True if the given string appears to be a phone number as might be found in Skype.
var
	int i, int n, string ch
s = stringTrimLeadingBlanks(stringTrimTrailingBlanks(s))
n = stringLength(s)
if n > 50
	; 50 is a pretty liberal limit for the length of a phone number...
	return False
endIf
i = 1
if stringLeft(s, 1) == "+"
	i = 2
endIf
while i <= n
	ch = substring(s, i, 1)
	if !stringIsBlank(ch) && !stringContains(",-.0123456789", ch)
		return False
	endIf
	i = i +1
endWhile
return True
endFunction

string function sutil__removeHTML(string s)
; Remove HTML tags and (some) character names from s.
var
	string lhs, string rhs,
	int pos,
	string result
pos = stringContains(s, "<")
while pos
	result = result +" " +stringLeft(s, pos-1)
	s = stringChopLeft(s, pos)
	pos = stringContains(s, ">")
	if pos
		s = stringChopLeft(s, pos)
	else
		s = ""
	endIf
	pos = stringContains(s, "<")
endWhile
result = stringChopLeft(result +" " +s, 1)
string__replace(result, "&nbsp;", " ", "e")
string__replace(result, "&lt;", "<", "e")
string__replace(result, "&gt;", ">", "e")
string__replace(result, "&lquot;", "\"", "e")
string__replace(result, "&rquot;", "\"", "e")
return result
endFunction

collection function sutil__getSkypeContactInfoFromMSAA(object o, int childID)
; Get as much information from MSAA about the indicated contact as possible.
; Currently this includes name, ID, status, ExtStatus, SpeedDial, and MoodText.
var
	string buf, string name, string ID, string stat, string extstat, string mood,
	int pos,
	collection cContact
name = ""
id = ""
stat = ""
extstat = ""

buf = sutil__removeHTML(o.accHelp(childID))
if stringLength(buf) && sutil___splitContactInfo(buf, name, id, stat)
	; Separate "Last seen: ..." from "Offline" when found.
	pos = stringContainsFromRight(stat, "(")
	if pos
		extstat = stringChopLeft(stat, pos)
		extstat = stringChopRight(extstat, 1)  ; chops off ')'
		stat = stringTrimTrailingBlanks(stringLeft(stat, pos-1))
	endIf
else
	; Use accValue when there's no accHelp.
	; AccHelp for contacts vanished around Skype 4.0 beta 3.
	buf = o.accValue(childID)
	if stringLength(buf) && stringRight(buf, 1) == ")"
		pos = stringContainsFromRight(buf, "(")
		if pos
			name = stringTrimTrailingBlanks(stringLeft(buf, pos-1))
			id = stringChopRight(stringChopLeft(buf, pos), 1)
		endIf
	endIf
endIf

cContact = new collection
cContact.Name = name
cContact.ID = id
cContact.Status = stat
if extstat
	cContact.ExtStat = extstat
endIf
; Include extra stuff from accDescription if found (Speed dial for example).
; Known format as of Skype 2.0.0.78: <status>[, <moodText>][ Speed dial <speedDialNumber>]
; TODO:  See if "Speed dial" is localized in accDescription.
buf = stringTrimTrailingBlanks(o.accDescription(childID))
if buf
	if buf == stat || !stringLength(stat)   ; match up to the end of status text
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(stat)))
	else
		; accDescription should always begin with status text.
		beep()
	endIf
	mood = ""
	if buf == ", "
		; There is mood text.
		buf = stringChopLeft(buf, 2)
		pos = stringContainsFromRight(buf, sutil__MSAASpeedDial)
		if pos
			mood = stringLeft(buf, pos-1)
			buf = stringChopLeft(buf, pos)
		else
			mood = buf +""
			buf = ""
		endIf
	endIf
	if mood
		cContact.MoodText = mood
	endIf
	if buf == sutil__MSAASpeedDial
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(sutil__MSAASpeedDial)))
		cContact.SpeedDial = intToString(stringToInt(buf))
		buf = stringTrimLeadingBlanks(stringChopLeft(buf, stringLength(intToString(stringToInt(buf)))))
	endIf
	if !stringIsBlank(buf)
		cContact.Other = buf
	endIf
endIf  ; non-empty buf from accDescription
return cContact
endFunction

int function sutil___refreshSkypeWindow(handle hwnd)
if False && getWindowClass(getFirstChild(hwnd)) == "TScrollBar"
	;speechOff()
	setFocus(getFirstChild(hwnd))
	pause()
	setFocus(hwnd)
	pause()
	speechOn()
	return True
endIf
;sayInteger(sendMessage(hwnd, 14, 0, 0))
return False
endFunction

collection function sutil__getCurrentContactInfo()
; Returns a collection consisting of all retrievable fields about the current contact, with field names.  Both MSAA and screen text are used.
; When a field appears on screen but its name can't be determined, it is named "Field" plus its ordinal number in the field list (e.g., "Field5").
; Currently recognized fields are Name, ID, Status, ExtStatus, SpeedDial, MoodText, Location, Time, TimeZone, and Phone1-3.
var
	handle hContactList,
	object o, int childID,
	int tries,
	int left, int right, int top, int bottom,
	int oldRestriction,
	int i, int ibase, int phoneCount, int done,
	string buf, int pos,
	string mood, string stat, string id,
	collection cContact

; First get the MSAA information.
o = sutil__getFocusObject(childID)
if !o
	o = appwin__getMSAAObject(appwin__SWT_ContactList)
endIf
if !o
	return ""
endIf
childID = o.accFocus
if !childID
	return ""
endIf
cContact = sutil__getSkypeContactInfoFromMSAA(o, childID)

; Now restrict the invisible cursor to the contact's rectangle as indicated by its MSAA coordinates.
; Obscured windows are skipped because that means the information from the screen might not be valid anyway.
; This whole step is also skipped if focus is not in the contact list, because even other contact lists with the same MSAA structure do not share the same on-screen format.
; Example:  the participant list in a chat window (Skype 3-).
hContactList = appwin__getSkypeWindow(appwin__SWT_ContactList)
if !hContactList
	; TODO: Exits here for Skype 5 because there is no contact list window.
	return cContact
endIf
saveCursor() invisibleCursor() saveCursor()
oldRestriction = getRestriction()
tries = 2
done = False
while !done
	if !tries || getFocus() != hContactList || isWindowObscured(hContactList) || !MSAA__rect(o, childID, left, right, top, bottom)
		;beep()
		;sayString("Rect " +getTextInRect(left, top, right, bottom, 0, IgnoreColor, IgnoreColor, True))
		setRestriction(oldRestriction)
		return cContact
	endIf
	moveTo((left+right)/2, (top+bottom)/2)
	setRestrictionToRect(left, top, right, bottom)
	; Make sure we're on the first text chunk in the rectangle.
	JAWSPageUp() JAWSHome() nextChunk() priorChunk() priorChunk()
	if stringContains(getLine(), cContact.Name)
		done = True  ; causes the loop to exit
	else
		if tries == 2
			sutil___refreshSkypeWindow(hContactList)
		endIf
		pause()
		;sayString("retry")
	endIf
	tries = tries -1
endWhile

; Collect the data one text chunk at a time from the screen.
; The first line is skipped in all cases because it is just the contact name, which we already got from MSAA.
; For SkypeOut contacts, the phone number appears on the second line, but we already have that from MSAA also.
; Field names can be filled in retroactively by this code as context is discovered.
if nextLine()
	JAWSHome()
	stat = cContact.Status
	id = cContact.ID
	if stringContains(stringLower(stat), sutil__MSAASkypeOut) && getChunk() == ID
		if !nextChunk()
			setRestriction(oldRestriction)
			return cContact
		endIf
	endIf
	done = False
	ibase = collectionItemCount(cContact)
	i = ibase
	mood = cContact.MoodText
	while !done && i < 50
		buf = getChunk()
		if stringCompare(buf, sutil__scMenu, True) && stringCompare(buf, mood, True)
		pos = stringContainsFromRight(buf, "("+sutil__scGMT)
		if pos
			; Skype 3.x structure: <time> (<tz>)[, <location>]
			ccontact.Time = stringTrimTrailingBlanks(stringLeft(buf, pos-1))
			buf = stringChopLeft(buf, pos)  ; also chops off '('
			pos = stringContains(buf, ")")
			cContact.TimeZone = stringLeft(buf, pos-1)
			buf = stringChopLeft(buf, pos)
			if buf == ","
				cContact.Location = stringTrimLeadingBlanks(stringChopLeft(buf, 1))
			endIf
		else  ; not a time/location line
			if sutil___isPhoneNumber(buf)
				phoneCount = phoneCount +1
				; This arranges for a single phone number to be called "Phone" but multiple phone numbers to be called "Phone1," "Phone2," etc.
				if phoneCount == 1
					cContact.Phone = buf
				elif phoneCount == 2
					cContact["Phone"+intToString(phoneCount)] = buf
					cContact.Phone1 = cContact.Phone
					collectionItemRemove(cContact, "Phone")
				else
					cContact["Phone"+intToString(phoneCount)] = buf
				endIf
			else
				; We don't know the name of this field, so we'll assign a generic one (and hope a better one is found above soon).
				cContact["Field"+intToString(i/2+1)] = buf
			endIf
			i = i +2
		endIf
		endIf  ; chunk is "Menu"
		done = !nextChunk()
	endWhile
endIf
setRestriction(oldRestriction)
return cContact
endFunction

globals string Skype__sutil___lastStat
string function sutil__getMyOnlineStatus()
; Returns this user's online/offline/etc. status as a string.
; In Skype 3.2+, the return string is already localized, as it comes directly from MSAA.
; Author:  Doug Lee
var
	string sVer,
	object o, int childID, int n,
	int found,
	string stat
o = appwin__getMSAAObject(appwin__SWT_Myself)
if o
	n = o.accChildCount
	childID = 1
	found = False
	while childID <= n
		if o.accRole(childID) == Role_System_ButtonDropDown
			stat = o.accName(childID)
			if !stat
				stat = sutil__msgStatUnknown
			endIf
			Skype__sutil___lastStat = stringLower(stat)
			found = True
		endIf
		childID = childID +1
	endWhile
endIf
return Skype__sutil___lastStat
endFunction

string function sutil__speakableSkypeID(string id)
; Speakable version of the given Skype ID:
;	- Conference IDs say "conference."
;	- US phone numbers are formatted as +1 nnn-nnn-nnnn.
if id == "#"
	id = sutil__msgConference
elif id == "+" && stringLength(id) == 12
	id = formatString("%1 %2-%3-%4",
		stringLeft(id, 2),
		substring(id, 3, 3),
		substring(id, 6, 3),
		stringRight(id, 4)
	)
endIf
return id
endFunction

int function sutil__play(string sound)
; Play the indicated sound file.
var
	string spath
;spath = findJAWSSoundFile(sound)
spath = getJAWSSettingsDirectory() +"\\sounds\\" +sound
if !fileExists(spath)
	spath = getJAWSSettingsDirectory() +"\\" +sound
endIf
if !fileExists(spath)
	spath = findJAWSSoundFile(sound)
endIf
if spath
	playSound(spath)
endIf
endFunction

string function sutil__getWindowClassPath(handle hwnd)
var
	string path,
	string sep
sep = "/"
while hwnd
	path = getWindowClass(hwnd) +"/" +path
	hwnd = getParent(hwnd)
endWhile
return stringChopRight(path, 1)
endFunction

int function sutil__isUnobscured(object o, int childID)
; Returns True if the indicated MSAA element appears to be unobscured by something else.
; Uses GetObjectAtPoint, so requires .AccHitTest to work.
var
	int left, int right, int top, int bottom,
	object o1, int childID1
if !MSAA__rect(o, childID, left, right, top, bottom)
	; If it doesn't have screen space, it's sort of obscured by default.
	return False
endIf
o1 = getObjectAtPoint(childID1, (left+right)/2, (top+bottom)/2)
if childID1 == childID && o.accName(0) == o1.accName(0)
	return True
endIf
return False
endFunction

int function sutil__isAncestorOf(handle hwndChild, handle hwndAncestor)
; Returns True if and only if hwndChild descends at some level from hwndAncestor.
; True is also returned if the two are equal.
while hwndChild
	if hwndChild == hwndAncestor
		return True
	endIf
	hwndChild = getParent(hwndChild)
endWhile
return False
endFunction

int function sutil__getMSAAItemDim(object o, int childID, string which)
; Return left, right, top, bottom, width, or height for the given MSAA item.
var
	int left, int top, int width, int height
o.accLocation(intRef(left), intRef(top), intRef(width), intRef(height), childId)
if !left && !top && !width && !height
	return 0
endIf
left = left +1
top = top +1
if which == "l" return left
elif which == "t" return top
elif which == "w" return width
elif which == "h" return height
elif which == "r" return left+width
elif which == "b" return top+height
else return 0
endIf
endFunction

int function sutil__findNthColorTransition(handle hwnd, int delta, int n)
; Finds the nth color transition in delta's direction from the current active cursor's location.
; Returns True on success and False on failure.
; The active cursor is moved to the found spot only if a spot is found.
; The search is limited by the edge of hwnd in delta's direction if hwnd is not null.
; Author:  Doug Lee
var
	int x, int y,
	int limit, int tooFar,
	int clr, int clr0
if hwnd
	if delta > 0
		limit = getWindowRight(hwnd)
	else
		limit = getWindowLeft(hwnd)
	endIf
else
	if delta > 0
		limit = screenGetWidth() +1
	else
		limit = 0
	endIf
endIf
x = getCursorCol() +(delta *3)  ; arbitrary safety margin to avoid edge of text characters
y = getCursorRow()
clr0 = getColorAtPoint(x, y)
tooFar = False
while n && !tooFar
	clr = getColorAtPoint(x, y)
	if clr != clr0
		n = n -1
		clr0 = clr
	endIf
	x = x +delta
	if delta > 1
		tooFar = (x > limit)
	else
		tooFar = (x < limit)
	endIf
endWhile
if tooFar
	return False
endIf
moveTo(x, y)
return True
endFunction

;====================== End module sutil (from skypeUtil) ======================


;================================== Module log =================================
; Allow logging to a text file.
;
; Interface:  log__open, log__isOpen, log__close, log__write, log__setUserID.
; Basic usage: log__open() from AutoStartEvent, log__write("...") to log something.
; Usage:
;	- Call log__open() from AutoStartEvent.
;	- Logging will occur by default only if the log file already exists (good for debugging during development).
;	- Call as log__open("", True) if you want logging to occur even if the log file doesn't already exist.
;	- Pass a file name instead of "" for a custom file name (rarely needed).
;	- Call log__write("Oops I saw an error") to log something.
;	- Use log__write("Oops this error needs to speak too", True) to speak what is being logged.
;	- Call log__close() to disable logging (rarely needed).
;
; Author:  Doug Lee of Level Access

; Constants for log__write and related functions.
; Not localized because log text is not localized.

; Used as a user ID when none is found.
const log___msgNoName = "<no name>"
; What "name" is used for the caller string when JAWS is too old for getScriptCallStack().
; %1 is the JAWS version.
const log___msgJAWSTooOld = "(caller unknown; JAWS version (%1) too old)"
; Name used for the caller when it can't be determined for an unknown reason,
; probably a getScriptCallStack return value format surprise.
const log___msgNoCaller = "(caller unknown)"

globals
; FileSystemObject and log stream object obtained from it.
	object Skype__log___fso,
	object Skype__log___logstream,
; Name of file.
	string Skype__log___fileName,
	int Skype__log___canCreate,
	string Skype__log___userID,
	string Skype__log___machineID

int function log___setup()
; Internal function called to set things up when necessary before any log text is generated.
if !stringIsBlank(Skype__log___userID)
	return True
endIf
Skype__log___fso = createObjectEx("Scripting.FileSystemObject", False)
if !Skype__log___fso
	return False
endIf
Skype__log___machineID = ""
var object oNet = createObjectEx("WScript.Network", False)
if oNet
	Skype__log___machineID = oNet.ComputerName
endIf
if stringIsBlank(Skype__log___machineID)
	Skype__log___machineID = log___msgNoName
endIf
Skype__log___userID = getJAWSUserName()
if stringIsBlank(Skype__log___userID) && oNet
	Skype__log___userID = oNet.UserName
endIf
if stringIsBlank(Skype__log___userID)
	Skype__log___userID = log___msgNoName
endIf
return True
endFunction

int function log__setUserID(string id)
; Optional means of setting a custom user ID.
; If not called, the value of getJAWSUserName() is used.
Skype__log___userID = id
return True
endFunction

int function log__close()
; Close the log file.
log___close()
Skype__log___fileName = ""
return True
endFunction

int function log___setFileName(string name)
; Set the log file name.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Helper for log__open().
var string basename
basename = name
if stringRight(basename,4) == ".log"
	basename = stringChopRight(basename, 4)
endIf
if !basename
	basename = getActiveConfiguration()
endIf
if !basename
	basename = getAppFileName()
	var int pos = stringContainsFromRight(basename, ".")
	if pos > 0
		basename = stringLeft(basename, pos-1)
	endIf
endIf
Skype__log___fileName = getJAWSSettingsDirectory() +"\\" +basename
if stringRight(Skype__log___fileName, 4) != ".log"
	Skype__log___fileName = Skype__log___fileName +".log"
endIf
return True
endFunction

int function log__open(/*?*/ string name, int canCreate)
; Open and optionally create the log file.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Call from autoStartEvent, or logging will not occur.
; Returns True if the file is opened and False if not.
if !log___setup()
	return False
endIf
Skype__log___canCreate = canCreate
log__close()
if !log___setFileName(name) || !log___open()
	Skype__log___fileName = ""
	return False
endIf
return True
endFunction

int function log___open()
; Raw open/create.  Sets Skype__log___logstream.
; Returns True on success and False on failure.
if Skype__log___logstream
	return True
endIf
if fileExists(Skype__log___fileName)
	Skype__log___logstream = Skype__log___fso.openTextFile(Skype__log___fileName, 8)  ; Append
elif Skype__log___canCreate
	Skype__log___logstream = Skype__log___fso.CreateTextFile(Skype__log___fileName)
endIf
if !Skype__log___logstream
	return False
endIf
scheduleFunction("log___close", 30)
return True
endFunction

int function log___close()
; Raw cloase; unsets Skype__log___logstream.
var variant null
if !Skype__log___logstream
	return True
endIf
Skype__log___logstream.Close()
Skype__log___logstream = null
return True
endFunction

int function log___isOpen()
; Returns True if the log file is actually open and False if not.
if Skype__log___logstream
	return True
endIf
return False
endFunction

int function log__isOpen()
; Returns True if the log file is open and False if not.
; "Open" here means can be written to but may not necessarily mean the file is physically opened at the moment.
if Skype__log___fileName
	return True
endIf
return False
endFunction

string function log___getStamp()
; Returns a formatted date/time stamp for the log file.
var string tm = sysGetTime("HH:mm:ss")
var string dt = sysGetDate("ddd MMM dd yyyy")
return formatString("%1 %2", tm, dt)
endFunction

int function log__write(string s, /*?*/ int alsoSpeak)
; Write a string to the log file, and also optionally speak it.
if alsoSpeak
	sayMessage(OT_Error, s)
endIf
if !log__isOpen()
	return False
endIf
if !log___open()
	return False
endIf
var string buf = formatString(
	"*** %1 %2@%3 (%4)\n%5:\n",
	log___getStamp(), Skype__log___userID, Skype__log___machineID, getAppFileName(),
	log___getCaller()
) +s +"\n\n"
; Standardize any CR/LF characters in buf before writing.
buf = stringReplaceChars(buf, "\r\n", "\n")
buf = stringReplaceSubstrings(buf, "\n", "\r\n")
Skype__log___logstream.write(buf)
return True
endFunction

string function log___getCaller()
; Get and return the caller of this logging system.
; Helper for log__write().
if getJFWVersion() < 1000000
	var string jver = stringSegment(util__getJAWSInfo(), "|", 1)
	return formatString(log___msgJAWSTooOld, jver)
endIf
var string stack = util__callStackInfo(2, "s", 2)
if !stringIsBlank(stack)
	return stack
endIf
return formatString("%1\nStack begin:\n%2\nStack end.",
	log___msgNoCaller,
	util__vcast(getScriptCallStack())
)
endFunction

string function log___splitOff(string byRef s, string sep, /*?*/ int fromRight)
; Split what's before sep off and return it.
; If fromRight is True, split everything before the last sep off and return that.
; Returns by reference s with both that and sep removed.
var string s1
var int pos
if fromRight
	pos = stringContainsFromRight(s, sep)
else
	pos = stringContains(s, sep)
endIf
if pos
	s1 = stringLeft(s, pos-1)
	s = stringChopLeft(s, pos+stringLength(sep)-1)
else
	s1 = s
	s = ""
endIf
return s1
endFunction

;================================ End module log ===============================


;================================= Module mkcs =================================
; This module implements a KeymapChangedEvent-based multi-key command sequence system.
;
; Usage:
; Include this module in the main jss.
; Create a message constant for the MKCS help text (see the mkcs__help script for format info).
; Load the help text from AutoStartEvent with a call to mkcs__setHelpText().
; Set the command key (and again if it changes) by calling mkcs__setCommandKey().
; Write key mappings in the jkm file using "&" between layers on each line.
; For each command level, make the following additional mappings (the examples are for the top level with [ as the command key):
;	[&Escape=mkcs__exit
;	[&Tab*=mkcs__help("next")
;	[&Shift+Tab*=mkcs__help("prev")
;	[&JAWSKey+Tab*=mkcs__help("cur")
;	[&Insert+Tab*=mkcs__help("cur")
;	[&Enter=mkcs__help("exec")
; NOTE: The first line above is mandatory, as it is also used by this code
; to find the original command key when changing it to a new one.
; See mkcs__setCommandKey().
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

include "hjglobal.jsh"

; For support of JAWS versions before 14.0.
prototype string function getScriptSynopsis(string scriptName, optional int useDefault)
prototype string function getScriptDescription(string scriptName, optional int useDefault)
; And for JAWS versions before 17.0...
prototype string function getActiveConfiguration(optional int wantDomain)

globals
; The path to the sound file to play on exit from a layer.
	string Skype__mkcs___soundPath,
; For default-level scripts, the namespace for the CommandKey value in the MKCSOptions section of the effective config file.
	string Skype__mkcs___namespace,
; The file name and jcf section used by the scripts, if explicitly set in a mkcs__setCommandKey call.
	string Skype__mkcs___fname,
	string Skype__mkcs___sect,
; The command key that starts all sequences recognized here.
	string Skype__mkcs___commandKey

string function mkcs___getActiveConfiguration(optional int wantDomain)
; Wrapper for the JAWS getActiveConfiguration() function that returns the application config when True is passed but no domain scripts are active.
; The JAWS documentation says getActiveConfiguration does this, but through at least JAWS 18 it returns "" in this case. [DGL, 2017-08-19
if !wantDomain
	return getActiveConfiguration()
endIf
var string cfg = getActiveConfiguration(wantDomain)
if cfg
	return cfg
endIf
return getActiveConfiguration()
endFunction

script mkcs__editCommandKey()
; Allow the user to alter the command key used to start key sequences for these scripts.
; Traditionally assigned to JAWSKey+C / Insert+C at the first level after the currently active command key.
var string sKey = mkcs__getCommandKey()
var string sKey0 = sKey
var int okPressed = inputBox(mkcs___msgEditPrompt, mkcs___msgEditTitle, sKey)
if !okPressed || stringCompare(sKey, sKey0, True) == 0
	sayMessage(OT_User_Requested_Information, mkcs___msgNoChange)
	return
endIf

; Set the key in the active or appropriate config file first, so mkcs__setCommandKey() will find it correctly.
var string jcfname = Skype__mkcs___fname
; That would be a jkm file name if not blank.
if jcfname
jcfname = stringLeft(jcfname, stringContainsFromRight(jcfname, ".")-1) +".jcf"
else
	jcfname = mkcs___getActiveConfiguration(True) +".jcf"
endIf
if !stringContains(jcfname, "\\")
	jcfname = getJAWSSettingsDirectory() +"\\" +jcfname
endIf
var string jcfSect = "MKCSOptions"
var string jcfKey = Skype__mkcs___namespace +"CommandKey"
var string userCommandKey = iniReadString(jcfSect, jcfKey, "", jcfname)
if stringCompare(userCommandKey, sKey) == 0
	sayMessage(OT_User_Requested_Information, mkcs___msgNoChange)
	return
endIf
iniWriteString(jcfSect, jcfKey, sKey, jcfname)
mkcs__setCommandKey(sKey)
endScript

int function mkcs__setCommandKey(string sKey, /*?*/ string fname, string sect)
; Sets the key used as a prefix for all command sequences in this application.
; Also performs other minor initialization tasks.
; Call from AutoStartEvent or appropriate initialization code.
; Also call if the key is ever changed.
; Updates the jkm file if necessary.
; If fname is not passed, defaults to <activeConfiguration>.jkm.
; If sect is not passed, defaults to "Common Keys."
; The given command key can be overridden by a value for CommandKey in an MKCSOptions section of the config file as of June 13, 2017.
; When using this in default-level scripts, first call mkcs__setNamespace("MyScripts") so the value will come from MyScriptsCommandKey instead.
; The config file defaults to <activeConfiguration>.jcf but gets the basename from fname if passed.
; Returns 0 if things are not left in sync, 1 if they are, and 2 if the jkm changed.
; Things should not be left out of sync unless there is an error.
Skype__mkcs___soundPath = ""
if !sKey
	; Can't set to null key, doing nothing is inconsistent,
	; so we set to a string that is not a valid key.
	sKey = "<NoKey>"
endIf
Skype__mkcs___commandKey = sKey
if fname
	Skype__mkcs___fname = fname
else
	fname = Skype__mkcs___fname
endIf
if !fname
	fname = mkcs___getActiveConfiguration(True) +".jkm"
endIf
if !stringContains(fname, "\\")
	fname = getJAWSSettingsDirectory() +"\\" +fname
endIf
if sect
	Skype__mkcs___sect = sect
else
	sect = Skype__mkcs___sect
endIf
if !sect
	sect = "Common Keys"
endIf
; Allow an MKCSOptions config value to override the command key.
; This is mainly for non-English installations, where author command key choice may block standard things like accented letter typing.
var string jcfname = stringLeft(fname, stringContainsFromRight(fname, ".")-1) +".jcf"
var string jcfSect = "MKCSOptions"
var string jcfKey = Skype__mkcs___namespace +"CommandKey"
var string userCommandKey = iniReadString(jcfSect, jcfKey, "", jcfname)
if userCommandKey
	Skype__mkcs___commandKey = userCommandKey
endIf

var string jkm = util__stringFromFile(fname)
if !jkm
	; Chances are this means we're on the secure desktop and can't read files.
	return 0
endIf
var string sep = "\n"
var int pos = stringContains(stringLower(jkm), stringLower("&escape=mkcs__exit"))
if !pos
	sayMessage(OT_Error, mkcs___msgNotFound1)
	return 0
endIf
var string sKey0 = stringLeft(jkm, pos-1)
pos = stringContainsFromRight(sKey0, sep)
if !pos
	sayMessage(OT_Error, mkcs___msgNotFound2)
	return 0
endIf
sKey0 = stringChopLeft(sKey0, pos)
; In case the first layer entry is not the command key entry.
pos = stringContainsFromRight(sKey0, "&")
if pos
	sKey0 = stringLeft(sKey0, pos-1)
endIf
if stringCompare(sKey0, Skype__mkcs___commandKey, False) == 0
	return 1
endIf

sayUsingVoice(VCTX_Message, mkcs___msgUpdating, OT_Message)
var int n = stringSegmentCount(jkm, sep)
var string jkm1 = ""
var int i = 1
var int changed = False
var int inSect = False
var string line
while i <= n
	line = stringSegment(jkm, sep, i)
	line = stringTrimLeadingBlanks(stringTrimTrailingBlanks(line))
	if stringStartsWith(line, "[" +sect +"]")
		inSect = True
	elif stringLeft(line, 1) == "[" && stringRight(line, 1) == "]"
		inSect = False
	elif inSect
		if stringStartsWith(line, sKey0+"&")
			line = Skype__mkcs___commandKey +stringChopLeft(line, stringLength(sKey0))
			changed = True
		endIf
	endIf
	if jkm1
		jkm1 = jkm1 +"\r" +sep
	endIf
	jkm1 = jkm1 +line
	i = i +1
endWhile
jkm1 = jkm1 +"\r" +sep
if changed
	util__stringToFile(fname, jkm1)
	sayUsingVoice(VCTX_Message, mkcs___msgDone, OT_Message)
	scheduleFunction("reloadAllConfigs", 2)
	return 2
else
	sayMessage(OT_Error, mkcs___msgUpdateError)
	return 0
endIf
endFunction

string function mkcs__getCommandKey()
; Returns the key used as a prefix for all command sequences in this application.
; This will be the value passed to the latest call to mkcs__setCommandKey().
return Skype__mkcs___commandKey
endFunction

string function mkcs___getActiveKeyboardLayout()
; Returns the name of the keyboard layout in effect or null if this can't be determined.
; Tested in JAWS 9.0.2169.
var string fname, string layout
fname = getJAWSSettingsDirectory() +"\\default.jcf"
layout = iniReadString("Options", "KeyboardType", "", fname)
if !layout
	fname = getSharedSettingsDirectory() +"\\default.jcf"
	layout = iniReadString("Options", "KeyboardType", "", fname)
endIf
return layout
endFunction

string function mkcs___getActiveKeymapFile()
; Returns the full path to the active .jkm file or nul if this can't be determined.
var
	string fname
fname = mkcs___getActiveConfiguration(True)
if !fname
	return ""
endIf
fname = fname +".jkm"
fname = findJAWSSettingsFile(fname)
if !fname
	return ""
endIf
return fname
endFunction

int function mkcs___updateKeyList(string byRef list, string prefix, string jkm, string layout, optional int includeScriptNames)
; Adds any key sequences to the given list that begin with the given prefix.
; Key sequences in the given keyboard layout section of the indicated jkm file are scanned for matches.
; Returns the number (0 or more) of entries added to the list.
var
	string keys,
	string sKey, string sKey0,
	int pos,
	int n, int i,
	int prefixLen,
	int additions
keys = iniReadSectionKeys(layout +" Keys", jkm)
n = stringSegmentCount(keys, "|")
if !keys || !n
	return 0
endIf
prefix = stringStripAllBlanks(prefix)
prefixLen = stringLength(prefix)
additions = 0
i = 1
while i < n
	sKey = stringStripAllBlanks(stringSegment(keys, "|", i))
	sKey0 = sKey
	; Case-insensitive prefix match that doesn't let sKey be a substring of the prefix.
	if stringLeft(sKey, prefixLen) == prefix && stringLength(sKey) > prefixLen
		; +1 cuts off the "&" after the prefix.
		sKey = stringChopLeft(sKey, prefixLen+1)
		; Only consider this key level, not subsequent ones.
		; Also remove the trailing "*" that means "stay at this level" if it is there.
		pos = stringContains(sKey, "&")
		if pos
			sKey = stringLeft(sKey, pos-1)
		elif stringRight(sKey, 1) == "*"
			sKey = stringChopRight(sKey, 1)
		endIf
		; Avoid adding duplicates or keys that are part of this help system.
		if !stringContains(stringLower(sKey), "tab")
		&& !stringContains(stringLower(sKey), "enter")
		&& !stringContains(stringLower(sKey), "esc")
		&& !stringSegmentIndex(list, "|", sKey, True)
			list = list +"|" +sKey
			if includeScriptNames
				list = list +"=" +mkcs___getScriptAssignedTo(sKey0)
			endIf
			additions = additions +1
		endIf
	endIf
	i = i +1
endWhile
return additions
endFunction

string function mkcs___getKeyList(string prefix, optional int includeScriptNames)
; Returns a list of the key sequences starting with the given prefix.
; Key sequences in Common Keys and the active keyboard layout section of the active jkm file are scanned for matches.
; Returned sequences omit the prefix.
var
	string layout,
	string jkm,
	string list
layout = mkcs___getActiveKeyboardLayout()
if !layout
	return ""
endIf
jkm = mkcs___getActiveKeymapFile()
if !jkm
	return ""
endIf
list = ""
mkcs___updateKeyList(list, prefix, jkm, "Common", includeScriptNames)
mkcs___updateKeyList(list, prefix, jkm, layout, includeScriptNames)
list = stringChopLeft(list, 1)  ; remove extra leading separator.
return list
endFunction

const
; Values for iKeyStatus in keymapChangedEvent
	mkcs___KeySequenceComplete = 0,
	mkcs___KeySequenceRepeating = 1,
	mkcs___KeySequencePending = 2,
	mkcs___KeySequenceCanceled = 3,
	mkcs___KeySequenceStart = 4,
	mkcs___KeySequenceRestarted = 5
globals
	int Skype__mkcs___active,
	string Skype__mkcs___helpText,
	int Skype__mkcs___lastKeyCode,
	string Skype__mkcs___sequence,
; The next three are used for the Enter/exec functionality.
	string Skype__mkcs___lastSeq,
	string Skype__mkcs___execScript,
	int Skype__mkcs___lastIndex,
	int Skype__mkcs___lastKeyStatus,
	int Skype__mkcs___helpIndex

int function mkcs___sayKeyPrompt(string seq)
; Speak the prompt for the given (or default) key sequence if there is a prompt for it.
; Returns True if the speaking of the prompt was handled and False if not.
var
	int idx,
	string prompt,
	int isScheduledCall
isScheduledCall = False
if !seq
	seq = Skype__mkcs___sequence
	isScheduledCall = True
endIf
; Cut off the command key from the sequence before looking up prompts.
idx = stringContains(seq, "&")
if idx
	seq = stringChopLeft(seq, idx)
else
	seq = ""
endIf
; Find the key's entry in the help text.
idx = stringContains(stringLower(Skype__mkcs___helpText), "\10"+stringLower(seq)+"=")
if idx
	; Chop off everything through seq=
	prompt = stringChopLeft(Skype__mkcs___helpText, idx+stringLength(seq)+1)
	; Chop off everything past the key's entry.
	prompt = stringSegment(prompt, "\10", 1)
	; Chop off the actual help text (left of the prompt).
	prompt = stringSegment(prompt, "|", 2)
	if prompt
		if !isScheduledCall
			stopSpeech()
		endIf
		sayMessage(OT_Help, prompt)
	endIf
	return True
endIf
return False
endFunction

void function mkcs__setHelpText(string helpText)
; Set the help text to be used by this system.
; Call from AutoStartEvent in the application script file.
Skype__mkcs___helpText = "\10" +helpText +"\10"
endFunction

script mkcs__help(string which)
; Implements round-robin keyboard help for the current level of a multikey sequence.
; Which: next to move forward, prev to move backward, cur to stay put.
; These should correspond to Tab, Shift+Tab, and JAWSKey+Tab/Insert+Tab, respectively.
; Exec, which corresponds to Enter, executes the last entry reported.
; For all this to work, appropriate help must have been registered with mkcs__setHelpText().
; Required help format:
;	- One line per help entry.
;	- Line format: <sequence>=<help>[|<prompt>]
;	- <sequence> is the sequence to which the help applies, minus the command prefix and its "&".
;	- <help> is the help text to speak for this sequence. "@" means use the script's jsd Synopsis or, failing that, Description value.
;	- <prompt> (optional) is the prompt to speak when the user actually enters this sequence (not when getting help).
;	- The first line should provide the prompt for the actual command key (e.g., "=|Command").
var
	string prefix,
	int len,
	int pos,
	int i, int n,
	string seg,
	string helps,
	string seq

; Get the prefix, which is the current position in key sequences without the leading command key or its "&".
; Also get len, its length.
; Prefix will end with an "&" if it is not empty.
; It's empty for top-level keys though (keys that immediately follow the command key).
if which == "exec"
	; The sequence should have just now completed,
	; so the sequence is in Skype__mkcs___lastSeq, not Skype__mkcs___sequence.
	seq = Skype__mkcs___lastSeq
else
	seq = Skype__mkcs___sequence
endIf
prefix = seq
pos = stringContains(prefix, "&")
if pos
	; Get rid of the leading command character and its "&"
	; Then put an "&" at the end.
	prefix = stringChopLeft(prefix, pos) +"&"
else
	; We're at the top level (right after entry of the command key).
	prefix = ""
endIf
len = stringLength(prefix)

; Loop through help text lines, collecting those that apply to the current level.
; JAWS seems to use \10, not \13, to separate lines in a Messages text string.
n = stringSegmentCount(Skype__mkcs___helpText, "\10")
helps = ""
i = 2  ; mkcs__setHelpText prepends a delimiter, so the first segment is null.
while i <= n
	seg = stringSegment(Skype__mkcs___helpText, "\10", i)
	seg = stringTrimLeadingBlanks(stringTrimTrailingBlanks(seg))
	; The first check here keeps out the initial Command prompt line.
	; The second part lets through current entries if we're not at the top level.
	; The third lets through top-level entries if that's where we are.
	if stringLeft(seg, 1) != "=" && ((len && stringLeft(seg, stringLength(prefix)) == prefix) || !len)
		; Chop all but the current level's keystroke off this entry if we're not at the top level.
		if len
			seg = stringChopLeft(seg, len)
		endIf
		; Look for entries that are just this level, not deeper ones.
		; This logic allows "&" to occur in help or prompt text.
		if !stringContains(seg, "&") || stringContains(seg, "&") > stringContains(seg, "=")
			; Remove the prompt part if it's there.
			pos = stringContainsFromRight(seg, "|")
			if pos > stringContains(seg, "=")
				seg = stringLeft(seg, pos-1)
			endIf
			helps = helps +"|" +seg
		endIf
	endIf
	i = i +1
endWhile
helps = stringChopLeft(helps, 1)  ; extra leading delimiter

; Figure out which entry to speak, and do it.
if which == "next" || which == "prev"
	if which == "next"
		Skype__mkcs___helpIndex = Skype__mkcs___helpIndex +1
	elif which == "prev"
		Skype__mkcs___helpIndex = Skype__mkcs___helpIndex -1
	endIf
	if Skype__mkcs___helpIndex <= 0
		Skype__mkcs___helpIndex = stringSegmentCount(helps, "|")
	elif Skype__mkcs___helpIndex > stringSegmentCount(helps, "|")
		Skype__mkcs___helpIndex = 1
	endIf
endIf  ; next or prev
var int idx = Skype__mkcs___helpIndex
if which == "exec"
	idx = Skype__mkcs___lastIndex
endIf
seg = stringSegment(helps, "|", idx)
if !seg
	; Happens when there are no non-help-system keys defined.
	if which == "exec"
		beep()
	else
		sayMessage(OT_Error, mkcs___msgNoKeys)
	endIf
	return
endIf
if which == "exec"
	var string sKey = stringSegment(seg, "=", 1)
	;sayMessage(OT_User_Requested_Information, sKey)
	; TODO: Disabled as this does not work in a layer: The key is typed
	; but the script is not run. Simulating seq+"&"+sKey actually fails.
	; [DGL, 2014-08-24, JAWS 15.0]
	;simulateKey(sKey)
	if !mkcs___simulateKey(seq, sKey)
		beep()
	endIf
else
	; This says the help text, pauses, then says the keystroke.
	sKey = stringSegment(seg, "=", 1)
	var string txt = stringSegment(seg, "=", 2)
	if txt == "@"
		txt = mkcs___getScriptHelp(mkcs___getScriptAssignedTo(seq+"&"+sKey))
	endIf
	sayMessage(OT_User_Requested_Information, formatString("%2, %1",
		sKey, txt
	))
endIf
endScript

int function mkcs___simulateKey(string seq, string sKey)
; Simulate the final key in a multikey sequence by calling its script.
sKey = formatString("%1&%2", seq, sKey)
var string mapping = mkcs___getScriptAssignedTo(sKey)
if !mapping
	return False
endIf
; Don't call directly because this may cause recursion problems.
; Don't use scheduleFunction directly in case "()" appears in the jkm.
; Schedule a formatStringWithEmbeddedFunctions to handle all this.
Skype__mkcs___execScript = mapping
scheduleFunction("mkcs___exec", 0)
return True
endFunction

string function mkcs___getScriptAssignedTo(string sKey)
; Get the script name attached to the given key or sequence.
; Works on sequences where getScriptAssignedTo() fails.
var string layouts = mkcs___getActiveKeyboardLayout()
if layouts
	layouts = layouts +" Keys|"
endIf
layouts = layouts +"Common Keys"
var string files = ""
var string cfg = mkcs___getActiveConfiguration(True)
if cfg
	files = files +formatString("|%1|%2",
		getJAWSSettingsDirectory() +"\\" +cfg +".jkm",
		getSharedSettingsDirectory() +"\\" +cfg +".jkm"
	)
endIf
files = files +formatString("|%1|%2",
	getJAWSSettingsDirectory() +"\\default.jkm",
	getSharedSettingsDirectory() +"\\default.jkm"
)
files = stringChopLeft(files, 1)
var int nfiles = stringSegmentCount(files, "|")
var int nlayouts = stringSegmentCount(layouts, "|")
var int i, int j, string fpath, string layout
var string mapping = ""
for i = 1 to nfiles
	fpath = stringSegment(files, "|", i)
	for j = 1 to nlayouts
		layout = stringSegment(layouts, "|", j)
		mapping = iniReadString(layout, sKey, "", fpath)
		if mapping
			return mapping
		endIf
		mapping = iniReadString(layout, sKey+"*", "", fpath)
		if mapping
			return mapping
		endIf
	endFor
endFor
return ""
endFunction

void function mkcs___exec()
; Execute a script called via Enter.
formatStringWithEmbeddedFunctions("<$"+Skype__mkcs___execScript+">")
Skype__mkcs___execScript = ""
endFunction

int function mkcs___playExitSound()
; Play the "exit sequence" sound file.
if !Skype__mkcs___soundPath
	Skype__mkcs___soundPath = mkcs___findSound("mkcsExit.wav")
	if !Skype__mkcs___soundPath
		Skype__mkcs___soundPath = mkcs___findSound("CameraShutter3.wav")
	endIf
endIf
if Skype__mkcs___soundPath
	return playSound(Skype__mkcs___soundPath)
endIf
return False
endFunction

string function mkcs___findSound(string sound)
; Find and return the full path to the given sound file,
; or return null if it can't be found.
var string spath, string base
;spath = findJAWSSoundFile(sound)
base = getJAWSSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
base = getSharedSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
return ""
endFunction

void function keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
if mkcs__keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
	return
endIf
if getJFWVersion() >= 1100000
	; Chaining in older JAWS versions causes "unknown function call to KeyMapChangedEvent" messages for some reason.
	; JAWS 11.0 starts including a KeymapChangedEvent in default.jss.
	keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
endIf
endFunction

void function mkcs__keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
; Helper for KeyMapChangedEvent.
; Returns True if the default KeyMapChangedEvent should be blocked and False if not.

;sayString(formatString("Last %1 %2, current %4 %5",
;	Skype__mkcs___sequence, intToString(Skype__mkcs___lastKeyStatus), intToString(Skype__mkcs___lastKeyCode),
;sKeyName, intToString(iKeyStatus), intToString(iKeyCode)
;))
var
	string seq
; Filter out unrelated sequences and make any wanted sounds indicating changes into andout of our sequence.
if iKeyStatus == mkcs___KeySequenceStart || iKeyStatus == mkcs___KeySequenceRestarted
	if !isSameKey(sKeyName, Skype__mkcs___commandKey)
		; Not our sequence.
		Skype__mkcs___active = False
		return False
	endIf
	; Start or restart of our sequence.
	Skype__mkcs___active = True
elif !Skype__mkcs___active
	; Something other than activation going on, but it's not our sequence.
	return False
elif iKeyStatus == mkcs___KeySequenceCanceled
|| (iKeyStatus == mkcs___KeySequenceComplete && (Skype__mkcs___lastKeyStatus != mkcs___KeySequenceComplete && Skype__mkcs___lastKeyStatus != mkcs___KeySequenceCanceled))
	; End of our sequence.
	mkcs___playExitSound()
	Skype__mkcs___active = False
endIf

; Establish the sequence of keys that got us here.
if iKeyStatus == mkcs___KeySequenceStart || iKeyStatus == mkcs___KeySequenceRestarted
	Skype__mkcs___sequence = ""
	Skype__mkcs___helpIndex = 0
endIf
seq = ""
if Skype__mkcs___sequence
	seq = Skype__mkcs___sequence +"&"
endIf
seq = seq +sKeyName

; Announce the new key if there's an announcement to make for it.
if iKeyStatus == mkcs___KeySequenceCanceled
	if iKeyCode == Skype__mkcs___lastKeyCode
	&& (Skype__mkcs___lastKeyStatus == mkcs___KeySequenceStart || Skype__mkcs___lastKeyStatus == mkcs___KeySequenceRestarted)
		; Pass a repeat of the root key through and speak its name.
		sayMessage(OT_Keyboard, sKeyName)
		if !gbKeyboardHelp
			typeKey(sKeyName)
		endIf
	else
		; Avoid saying the invalid key's name and stop any previous speech.
		; Then announce the sequence cancellation.
		stopSpeech()
		;sayMessage(OT_Help, "Canceled", "")
	endIf
elif iKeyStatus != mkcs___KeySequenceComplete || Skype__mkcs___lastKeyStatus != mkcs___KeySequenceComplete
	; Say the key sequence's prompt if there is one, or the key name if not..
	; The above condition keeps this block from applying to non-layered script keys.
	if !mkcs___sayKeyPrompt(seq)
		stopSpeech()
		; TODO: Disabled because we don't always want default labels.
		;sayMessage(OT_Keyboard, sKeyName)
	endIf
endIf

; Set the sequence to remember for next time.
; Also set other globals to remember parameters to this call.
if iKeyStatus == mkcs___KeySequenceComplete || iKeyStatus == mkcs___KeySequenceCanceled
	if iKeyStatus == mkcs___KeySequenceComplete
		Skype__mkcs___lastSeq = Skype__mkcs___sequence
		Skype__mkcs___lastIndex = Skype__mkcs___helpIndex
	endIf
	Skype__mkcs___sequence = ""
	Skype__mkcs___helpIndex = 0
elif iKeyStatus == mkcs___KeySequenceRepeating
	; Do nothing; the new key stayed at this level.
	;scheduleFunction("mkcs___sayKeyPrompt", 0)
else
	; Pending:  Include the new key.
	; Start/Restarted:  Start with only the new key (already done above).
	Skype__mkcs___sequence = seq
	Skype__mkcs___helpIndex = 0
endIf
Skype__mkcs___lastKeyCode = iKeyCode
Skype__mkcs___lastKeyStatus = iKeyStatus
; This avoids an extra top-level prompt on the first non-Command scripted key after a sequence cancellation.
if iKeyStatus == mkcs___KeySequenceCanceled
	Skype__mkcs___lastKeyStatus = mkcs___KeySequenceComplete
endIf

; Chain this event.
;sayString("map " +sKeyName +" " +stringSegment("complete|repeating|pending|canceled|start|restarted", "|", iKeyStatus+1))
return True
endFunction

void function cancelLayeredKeySequence()
if Skype__mkcs___active
	mkcs___playExitSound()
	Skype__mkcs___active = False
endIf
cancelLayeredKeySequence()
endFunction

script mkcs__exit()
;sayMessage(OT_Help, "Exit", "", False)
endScript

int function mkcs__isSequenceActive()
; Returns True if one of our key sequences is active (in progress).
return Skype__mkcs___active
endFunction

string function mkcs___getScriptHelp(string scriptName)
; Get help for the given script. Tries Synopsis, Description, and default Synopsis and Description, in that order.
; On failure, returns the null string.
var string h = getScriptSynopsis(scriptName)
if !h h = getScriptDescription(scriptName) endIf
if !h h = getScriptSynopsis(scriptName, True) endIf
if !h h = getScriptDescription(scriptName, True) endIf
return h
endFunction

;=============================== End module mkcs ===============================


;================================= Module stats ================================
; A module for keeping and reporting statistics, often for debugging purposes.
;
; Author:  Doug Lee of Level Access

globals
	collection Skype__stats___stats,
	int Skype__stats___enabled

void function stats__enable()
; Enable the keeping of stats.
Skype__stats___enabled = True
endFunction

void function stats__disable()
; Disable the keeping of stats for runtime efficiency.
Skype__stats___enabled = False
endFunction

void function stats__inc(string name, optional collection byRef stats)
; Add 1 to the indicated value, creating it at 1 if necessary.
stats__delta(name, 1, stats)
endFunction

void function stats__dec(string name, optional collection byRef stats)
; Subtract 1 from the indicated value, creating it at -1 if necessary.
stats__delta(name, -1, stats)
endFunction

void function stats__delta(string name, int delta, optional collection byRef stats)
; Apply a change to the indicated value.
if !Skype__stats___enabled return endIf
if !Skype__stats___stats
	Skype__stats___stats = new collection
endIf
if !stats
	stats = Skype__stats___stats
endIf
stats[name] = stats[name] +delta
endFunction

void function stats__set(string name, int val, optional collection byRef stats)
; Set the indicated value.
if !Skype__stats___enabled return endIf
if !Skype__stats___stats
	Skype__stats___stats = new collection
endIf
if !stats
	stats = Skype__stats___stats
endIf
stats[name] = val
endFunction

void function stats__clear(optional collection byRef stats)
if !Skype__stats___enabled return endIf
if stats
	stats = new collection
else
	Skype__stats___stats = new collection
endIf
endFunction

void function stats__say(optional collection byRef stats)
stats__dump(False, stats)
endFunction

void function stats__show(optional collection byRef stats)
stats__dump(True, stats)
endFunction

void function stats__dump(int show, optional collection byRef stats)
var string statname, int val
var int customStats = True
if !stats
	stats = Skype__stats___stats
	customStats = False
endIf
var int ot = OT_Message
if show
	ot = OT_User_Buffer
endIf
var int found = False
forEach statname in stats
	val = stats[statname]
	if val
		sayMessage(ot, statname +" " +intToString(val))
		found = True
	endIf
endForEach
if !found
	beep()
endIf
var int enabled = Skype__stats___enabled
Skype__stats___enabled = True
if customStats
stats__clear(stats)
else
	stats__clear()
endIf
Skype__stats___enabled = enabled
endFunction

;=============================== End module stats ==============================


;=============================== Module winclick ===============================
; Support for sending mouse clicks of various sorts directly to a window.
; This permits clicking on obscured windows
; or clicking without checking for what is in the foreground.
;
; Author:  Doug Lee, with credit to Jim Bauer for the idea and some initial research.
;
; TODO:
;	- No support for mouse wheel events.
;	- Alt+clicks not supported.
;	- Non-client clicks not supported, such as for scrollbar elements.


int function winclick__send(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

; Returns the sendMessage() result.
return winclick___doClick(True, hwnd, which, x, y)
endFunction

void function winclick__post(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

winclick___doClick(False, hwnd, which, x, y)
endFunction

int function winclick___doClick(int useSend, handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() or postMessage call.
; See winclick__sendClick() or winclick__postClick() for details on usage.
; Returns the sendMessage() result if sendMessage() is used, or nothing if postMessage() is used.
; Fine print: This function sends two messages, ; one for buttonDown and one for buttonUp.
; The sendMessage result comes from the first of these calls when sendMessage is used.
; Only one message is sent for a hover or leave.
which = stringLower(which)
var int msgDown, int msgUp
var int wParam = 0
if stringContains(which, "hover")
	msgDown = 0x02A1  ; WM_MOUSEHOVER
	msgUp = 0  ; skip that one.
elif stringContains(which, "leave")
	msgDown = 0x02A3  ; WM_MOUSELEAVE
	msgUp = 0  ; skip that one.
elif stringContains(which, "left")
	if stringContains(which, "dbl")
		msgDown = 0x0203  ; WM_LBUTTONDBLCLK
	else
		msgDown = 0x0201  ; WM_LBUTTONDOWN
	endIf
	msgUp = 0x0202  ; WM_LBUTTONUP
	wParam = wParam | 1  ; MK_LBUTTON
elif stringContains(which, "right")
	if stringContains(which, "dbl")
		msgDown = 0x0206  ; WM_RBUTTONDBLCLK
	else
		msgDown = 0x0204  ; WM_RBUTTONDOWN
	endIf
	msgUp = 0x0205  ; WM_RBUTTONUP
	wParam = wParam | 2  ; MK_RBUTTON
elif stringContains(which, "middle")
	if stringContains(which, "dbl")
		msgDown = 0x0209  ; WM_MBUTTONDBLCLK
	else
		msgDown = 0x0207  ; WM_MBUTTONDOWN
	endIf
	msgUp = 0x0208  ; WM_MBUTTONUP
	wParam = wParam | 0x10  ; MK_MBUTTON
else
	beep()
	return False
endIf
if stringContains(which, "shift")
	wParam = wParam | 4  ; MK_SHIFT
endIf
if stringContains(which, "ctrl") || stringContains(which, "control")
	wParam = wParam | 8  ; MK_CONTROL
endIf
var int lParam = 0
if x || y
	if !msgUp
		saveCursor()  JAWSCursor()  saveCursor()
		moveTo(x, y)
	endIf
	var int left, int right, int top, int bottom
	; -4 is ObjID_Client
	if MSAA__windowRect(hwnd, -4, left, right, top, bottom)
		x = x -left
		y = y -top
	else
		x = x -getWindowLeft(hwnd)
		y = y -getWindowTop(hwnd)
	endIf
	lParam = makeLong(x, y)
endIf
var int result
if useSend
	result = sendMessage(hwnd, msgDown, wParam, lParam)
	if msgUp
		sendMessage(hwnd, msgUp, 0, lParam)
	endIf
else
	postMessage(hwnd, msgDown, wParam, lParam)
	if msgUp
		postMessage(hwnd, msgUp, 0, lParam)
	endIf
endIf
if !msgUp
	pause()
endIf
return result
endFunction

;============================= End module winclick =============================


;================================= Module MSAA =================================
; Miscellaneous functions for MSAA access
;
; Author:  Doug Lee

string function MSAA__windowName(handle hwnd, /*?*/ int childID, int objID)
; Get the MSAA name for the given window.
return MSAA__windowProp(hwnd, "name", childID, objID)
endFunction

collection function MSAA__windowRect(handle hwnd, int objID, optional int byRef left, int byRef right, int byRef top, int byRef bottom)
; Return by reference and as a collection the rectangle for the given object ID in the given window.
; Returns a collection on success and null on failure.
left = 0
right = 0
top = 0
bottom = 0
var object o, int childID
o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o
	return
endIf
return MSAA__rect(o, 0, left, right, top, bottom)
endFunction

variant function MSAA__windowProp(handle hwnd, string whichProp, /*?*/ int childID, int objID)
; Returns any one of several possible MSAA properties for the given window.
; Supported properties:
;	Name, Role, State, Value.
;	Description, KeyboardShortcut, Help, ChildCount.
var
	variant result,
	object o
if whichProp == "childCount" || whichProp == "role" || whichProp == "state"
	result = util__vcast(0)
else
	result = util__vcast("")
endIf
if !hwnd
	return result
endIf
o = MSAA__objectFromWindow(hwnd, objID)
if !o
	return result
endIf
childID = childID +0
if whichProp == "name"
	return o.accName(childID)
elif whichProp == "state"
	return o.accState(childID)
elif whichProp == "role"
	return o.accRole(childID)
elif whichProp == "value"
	return o.accValue(childID)
elif whichProp == "description"
	return o.accDescription(childID)
elif whichProp == "help"
	return o.accHelp(childID)
elif whichProp == "keyboardShortcut"
	return o.accKeyboardShortcut(childID)
elif whichProp == "childCount"
	if childID
		return o.accChild(childID).accChildCount
	else
		return o.accChildCount
	endIf
endIf
return result
endFunction

int function MSAA__click(object o, int childID)
; Click the center of the indicated MSAA element.
; This function does not check directly for invisible, offscreen, obscured, etc.,
; but it avoids clicking if the element at the click location seems to be the wrong one.
; Returns True if clicked and False if not.
var
	int left, int right, int top, int bottom
var string name = o.accName(childID)
var string role = o.accRole(childID)
if !MSAA__rect(o, childID, left, right, top, bottom)
	return False
endIf
saveCursor()  JAWSCursor()  saveCursor()
var int x = (left+right) / 2
var int y = (top+bottom) / 2
moveTo(x, y)
if util__abs(x-getCursorCol()) > 2 || util__abs(y-getCursorRow()) > 2
	; That's to protect against off-screen coordinates mostly.
	return False
endIf
var object o1, int childID1
o1 = getObjectAtPoint(childID1, x, y)
if childID1 != childID
|| o.accRole(childID1) != role
|| stringCompare(o1.accName(childID1), name, True) != 0
	return False
endIf
	leftMouseButton()
	return True
endFunction

int function MSAA__childByName(object o, string name, /*?*/ string flags, int startID, int endID)
; Returns the child ID whose object's name matches the given name, or 0 if no match is found.
; Flags:
;	c = case sensitive, otherwise not.
;	p = prefix match (returned name may be longer), otherwise full match.
;	s = substring (name contains given text).
;	x = exact, do not strip leading and trailing blanks.
; p and s are mutually exclusive.
; startID: The starting childID to check.
; endID:  The last childID to check.
var
	int childID,
	string curname
if o.accRole(0) == 9  ; 9 == Role_System_Window
	o = o.accChild(-4)  ; client object
endIf
if !startID
	startID = 1
endIf
childID = o.accChildCount
if !endID || endID > childID
	endID = childID
endIf
childID = startID
var int stripEnds = (stringContains(flags, "x") == 0)
if stripEnds
	name = stringTrimLeadingBlanks(stringTrimTrailingBlanks(name))
endIf
var int ignoreCase = (stringContains(flags, "c") == 0)
if ignoreCase
	name = stringLower(name)
endIf
var int len = stringLength(name)
var int substringMatch = (stringContains(flags, "s") > 0)
var int prefixMatch = (stringContains(flags, "p") >= 0)
var int curlen
while childID <= endID
	curname = o.accName(childID)
	if stripEnds
		curname = stringTrimLeadingBlanks(stringTrimTrailingBlanks(curname))
	endIf
	curlen = stringLength(curname)
	if curlen == 0
	&& !(o.accState(childID) & (State_System_OffScreen | State_System_Invisible))
		; Use on-screen text instead.
		curName = MSAA__rectText(o, childID)
		if stripEnds
			curname = stringTrimLeadingBlanks(stringTrimTrailingBlanks(curname))
		endIf
		curlen = stringLength(curname)
	endIf
	if ignoreCase
		curname = stringLower(curname)
	endIf
	if substringMatch
		if stringContains(curname, name)
			return childID
		endIf
	elif prefixMatch
		if curlen >= len
			curname = stringLeft(curname, len)
			if name == curname
				return childID
			endIf
		endIf
	elif stringCompare(name, curname, True)
		return childID
	endIf
	childID = childID +1
endWhile
return 0
endFunction

string function MSAA__rectText(object o, int childID, /*?*/ int attributes, int textColor, int backgroundColor, int bIncludeLineBreaks, int bGetBrailleText)
var int left, int right, int top, int bottom
if !MSAA__rect(o, childID, left, right, top, bottom)
	return ""
endIf
; This handles optional parameters as best we can do portably.
if bGetBrailleText
	return getTextInRect(left, top, right, bottom, attributes, textColor, backgroundColor, bIncludeLineBreaks, bGetBrailleText)
elif bIncludeLineBreaks
	return getTextInRect(left, top, right, bottom, attributes, textColor, backgroundColor, bIncludeLineBreaks)
elif stringLength(util__vcast(backgroundColor))
	return getTextInRect(left, top, right, bottom, attributes, textColor, backgroundColor)
elif stringLength(util__vcast(textColor))
	return getTextInRect(left, top, right, bottom, attributes, textColor)
elif stringLength(util__vcast(attributes))
	return getTextInRect(left, top, right, bottom, attributes)
else
	return getTextInRect(left, top, right, bottom)
endIf
endFunction

collection function MSAA__rect(object o, int childID, optional int byRef left, int byRef right, int byRef top, int byRef bottom)
; Get the bounding rectangle for the given MSAA object and childID.
; The rectangle is returned 1-based, as JAWS likes its rectangles.
; MSAA provides 0-based left/top/width/height.
; The rectangle is returned by reference and also as a collection. Returns null on failure.
if !o
	left = 0
	right = 0
	top = 0
	bottom = 0
	return
endIf
if !childID
	childID = 0
endIf
o.accLocation(intRef(left), intRef(top), intRef(right), intRef(bottom), childID)
if !left && !right && !top && !bottom
	return
endIf
; but we want 1-based left/right/top/bottom.
left = left +1
top = top +1
right = left +right
bottom = top +bottom
var collection c c = new collection
c.left = left
c.right = right
c.top = top
c.bottom = bottom
return c
endFunction

globals
	handle Skype__MSAA___lastHwnd,
	int Skype__MSAA___lastObjID,
	object Skype__MSAA___lastObj,
	int Skype__MSAA___lastTC

object function MSAA__objectFromWindow(handle hwnd, /*?*/ int objID)
; Return the requested (default ObjID_Client) MSAA object from the given window.
; Caches results a bit for better performance.
if !stringLength(util__vcast(objID))
	objID = -4  ; ObjID_Client
endIf
var int tc = getTickCount()
if hwnd == Skype__MSAA___lastHwnd && objID == Skype__MSAA___lastObjID && Skype__MSAA___lastObj
&& Skype__MSAA___lastTC && tc - Skype__MSAA___lastTC < 500
	return Skype__MSAA___lastObj
endIf
var object o, int childID
o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o
	return o
endIf
Skype__MSAA___lastHwnd = hwnd
Skype__MSAA___lastObjID = objID
Skype__MSAA___lastObj = o
Skype__MSAA___lastTC = tc
return o
endFunction

;=============================== End module MSAA ===============================


;================================== Module UIA =================================
; Code for accessing and manipulating the UIAScriptAPI in JAWS 14 and later.
;
; Globals not under namespace control: gc_cachedAPIObjects.
;
; Author: Doug Lee of Level Access

globals
; The main UIA interface, cached in case this speeds things.
	object Skype__UIA___gUIA,
; An optional cache of API objects.
; This is a workaround for a JAWS (at least 14-18) issue whereby retrieval of an API object can start failing until the next JAWS restart.
; This global is deliberately JAWS-wide, not namespace-controlled, so that one script set building a cache will benefit all others.
	collection gc_cachedAPIObjects,
; An indicator of the version of the UIA support (see UIA__version()).
; Set by the first UIA__getScriptAPI() call.
	int Skype__UIA___version,
; Lists of IDs from UIA.jsh / UI Automation.
	collection Skype__UIA___props,
	collection Skype__UIA___vals,
	collection Skype__UIA___events,
	collection Skype__UIA___attributes,
	collection Skype__UIA___textUnits,
	collection Skype__UIA___fullnames,
; A flag indicating when all those lists are completely built.
	int Skype__UIA___built

; Messages for this code that should not be localized,
; because they are for scripters and should not be heard by users.
messages
@UIA___msgInvalidScopeChar
Invalid search scope character: %1
@@
@UIA___msgNullConditionString
Null condition string
@@
@UIA___msgStringNotTerminated
String literal in condition spec was not terminated
@@
; %1 is a reference like %11 (only up to %10 allowed).
@UIA___msgInvalidRef
Invalid reference in condition: %1
@@
; %1 is a list type (property, value, event, etc.).
; %2 is an invalid property/value/event/etc. name.
@UIA___msgInvalidName
Module UIA: Invalid %1 name %2
@@
; %1 is the number of unused stack values.
@UIA___msgSpecIncomplete
%1 value(s) not used - missing trailing 'and'?
@@
@UIA___msgUnknownError
Unknown error in condition specification
@@
endMessages

object function UIA__getScriptAPI(optional int forceNew)
; Get the main UIAScriptAPI object.
; Also sets UIA__version on first call.
; If forceNew is True, the object is a new reference.
; Otherwise it may (but may not) be a shared object.
; Shared objects may be shared by other scripts.
; You may want to pass True when hooking events.
if !Skype__UIA___gUIA
	Skype__UIA___gUIA = UIA___rawCreate()
	if !Skype__UIA___gUIA
		; Use the API object cache if somebody built one.
		Skype__UIA___gUIA = UIA___getCachedAPIObject(forceNew)
	endIf
	if !Skype__UIA___gUIA
		return
	endIf
	if !Skype__UIA___version
		; UIA__version() sets Skype__UIA___version to avoid recursion.
		Skype__UIA___version = UIA__version(Skype__UIA___gUIA)
	endIf
endIf
scheduleFunction("UIA___UIAClear", 50)
if forceNew
	var object newObj = UIA___rawCreate()
	if !newObj
		; Use the API object cache if somebody built one.
		newObj = UIA___getCachedAPIObject(forceNew)
	endIf
	return newObj
endIf
return Skype__UIA___gUIA
endFunction

object function UIA___rawCreate()
; Raw creator of the UIAScriptAPI object.
var object uia = createObjectEx("FreedomSci.UIA", FALSE, "UIAScriptAPI.x.manifest")
return uia
endFunction

int function UIA__version(optional object oUIA)
; Indicate the version of the available or given UIA support:
;	-1 - Unknown.
;	0 - Not set yet.
;	1 - JAWS 14.
;	2 - JAWS 15 before update 1.
;	3 - JAWS 15 update 1 (15.0.5056).
; The version of the available UIA support is cached by UIA__getScriptAPI().
var int ver = -1
if !oUIA
	if Skype__UIA___version
		; Already figured out for this JAWS instance.
		return Skype__UIA___version
	endIf
	; UIA__getScriptAPI() passes a valid object,
	; so this will not cause recursion.
	; Set Skype__UIA___version here to keep UIA__getScriptAPI() from calling us though.
	Skype__UIA___version = -1
	oUIA = UIA__getScriptAPI()
	if !oUIA
		return -1
	endIf
endIf
var variant names = COMGetAvailableNames(oUIA)
names = stringLower(names)
if stringContains(names, "addtextedittextchangedeventhandler()")
	; Added in JAWS 15.0.5056.
	; Anything newer than that JAWS version must be detected here.
	ver = 3
elif stringContains(names, "events()")
	; Included in JAWS 14 but removed in JAWS 15.
	ver = 1
else
	; All we have left is the initial JAWS 15 release.
	ver = 2
endIf
return ver
endFunction

object function UIA___getCachedAPIObject(int forceNew)
; Return (and remove from cache if not shared) a cached UIAScriptAPI object.
var collection cache = gc_cachedAPIObjects
if !cache
	return
endIf
var int i = 1
if forceNew
	i = cache.count
	if i <= 1
		; Out of new ones... (cache item 1 is reserved as the shared object).
		UIA___reportAPICacheSize()
		return
	endIf
endIf
var object api = cache[intToString(i)]
if i > 1
	var variant null
	cache[intToString(i)] = null
	cache.count = cache.count -1
endIf
UIA___reportAPICacheSize()
return api
endFunction

void function UIA___reportAPICacheSize()
; Indicate to the user the API object cache size. This is called when the size changes.
var string fname = getJAWSSettingsDirectory() +"\\Sounds\\gauge" +intToString(gc_cachedAPIObjects.count) +".wav"
if !fileExists(fname)
	return
endIf
playSound(fname)
endFunction

object function UIA__elementFromMSAA(object o, int childID)
; Return the UIA element corresponding to the given MSAA object and childID, or null on error.
var object oUIA = UIA__getScriptAPI()
if !oUIA
	return
endIf
return oUIA.getElementFromIAccessible(o, childID)
endFunction

variant function UIA__boundingRectangle(object el, optional int adjustForJAWS)
; Return an object or collection with left/right/top/bottom properties for the given element.
; In JAWS 15.0.5056 and later, an object is returned.
; Otherwise a collection simulating that object is returned.
; If adjustForJAWS is passed and True, 1 is added to each coordinate,
; and a collection is always returned.
; Warning: When an object is returned, its properties are read-only.
var variant rect, object rect0
if UIA__version() >= 3
	if !adjustForJAWS
		return el.boundingRectangle()
	endIf
	rect0 = el.boundingRectangle()
	rect = new collection
	rect.left = rect0.left +1
	rect.right = rect0.right +1
	rect.top = rect0.top +1
	rect.bottom = rect0.bottom +1
	return rect
endIf

rect = new collection
var int left, int right, int top, int bottom
if !el.boundingRectangle(intRef(left), intRef(top), intRef(right), intRef(bottom))
	return
endIf
if adjustForJAWS
rect.left = left +1
rect.right = right +1
rect.top = top +1
rect.bottom = bottom +1
else
rect.left = left
rect.right = right
rect.top = top
rect.bottom = bottom
endIf
return rect
endFunction

object function UIA__findAll(object el, int treeScope, variant cond, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Return an array of matches to the given condition.
; The condition may be an object as expected by oUIA.FindAll(),
; or a string as expected by UIA__condition().
if getVariantType(cond) == 8  ; VT_String
	cond = UIA__condition(cond, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
	if !cond
		return
	endIf
endIf
return el.findAll(treeScope, cond)
endFunction

collection function UIA__elementIterator(object els)
; Return an iterator for UIA__nextElement.
; Includes .count for the element count, for convenience.
; At any instant, .index is the 1-based index of iteration.
; .elements is the actual set of elements.
; .currentElement is the current element during iteration.
; ._version is a (private) indicator of UIAScriptAPI version.
; Useful when supporting JAWS 14.
var collection iter iter = new collection
var variant null
iter.elements = els
iter.count = els.count +els.length
UIA__resetIterator(iter)
iter._version = UIA__version()
return iter
endFunction

void function UIA__resetIterator(collection iter)
; Reset the iterator to start at the beginning.
iter.index = 0
var variant null
iter.currentElement = null
endFunction

int function UIA__nextElement(collection iter)
; Moves to the next element in iter, as returned by UIA__elementIterator().
; Returns True if there is another element and False otherwise.
; .currentElement becomes the next element.
var variant null
iter.index = iter.index +1
if iter.index > iter.count
	iter.currentElement = null
	return False
endIf
var object els = iter.elements
var int idx = iter.index -1
if iter._version >= 3
	iter.currentElement = els(idx)
else
	iter.currentElement = els.getElement(idx)
endIf
return True
endFunction

; Cache of conditions by spec, excluding those with %<n> parameters.
globals collection Skype__UIA___conds

object function UIA__condition(string spec, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Build and return a UIAScriptAPI Condition object from a string specification.
; Values and other Condition objects may be passed and referred to in the spec.
; The spec is in post-fix notation, case-insensitive.
; Items in the spec are separated by a single space or line ending.  Allowed items:
;	contentView, controlView, rawView: Require the given UIA tree view type.
;	focusedProcess: Require the process ID to match the focused application's process ID.
;	FPRaw/Control/ContentView: One of the above with focusedProcess.
;	not: Reverse the previous condition: "focusedProcess not."
;	and, or: Combine previous two conditions: "contentView focusedProcess and."
;	property name: The name of a property with a PropertyID.
;		These are from UIA.jsh without UIA_ prefix and PropertyId suffix.
;		Examples: Name, ControlType.
;	value name: The name of a value with an ID. Categories:
;		Control types like CheckBox, Edit, and Hyperlink.
;		Other constants without *_ prefix, like Horizontal and LargeIncrement.
;		Other raw constants like Next that contain no underscore.
;	False, True: Create a boolPropertyCondition with the previous item.
;		Example: "IsEnabled True."
;		If not after a property, creates a False or True condition
;		(directly in JAWS 15.0.5056+, simulated before that).
;	int like 1, -50, etc.: An integer value.
;		Makes an IntPropertyCondition with the previous item.
;		Example: "RangeValueMinimum -5"
;	String like 'hello': A string value.
;		Makes a StringPropertyCondition with the previous item.
;		Example: "Name 'My UIA Application'"
;		Apostrophes in the string are written as ''. Backslash escapes like \t also work.
;	%n (1-10): Use argument n, 1 being the first after the spec, etc.
;	May be a string or a Condition object. Ints are not supported here.
; Examples:
;	c1 = UIA__condition("ControlType GroupControl Name %1 and", getObjectName(True, 1))
;		(Must be a group control named whatever getObjectName(True, 1) returns.)
;	c2 = UIA__condition("ContentView %1 and", c1)
;		Same as c1 but must also be in the ContentView tree.
; Returns an IDispatch Condition object for further use with the UIAScriptAPI.
; Returns null on error and may speak a message explaining the problem.
if !Skype__UIA___conds
	Skype__UIA___conds = new collection
endIf
var object cnd, variant null
spec = stringReplaceChars(spec, "\10\13\160", " ")
spec = stringReplaceSubstrings(spec, "  ", " ")
if !stringContains(spec, "%")
	cnd = Skype__UIA___conds[spec]
	if cnd
		return cnd
	endIf
endIf
var string spec0 = spec
var object oUIA = UIA__getScriptAPI()
if !oUIA
	UIA___sayError(UIA___msgNoUIAObject)
	return
endIf
spec = stringTrimLeadingBlanks(stringTrimTrailingBlanks(spec))
if !spec
	UIA___sayError(UIA___msgNullConditionString)
	return
endIf
; Shortcuts allowed starting in JAWS 15.0.5056.
if UIA__version() >= 3
	cnd = null
	if stringCompare(spec, "rawview", False) == 0
		cnd = oUIA.rawViewCondition
	elif stringCompare(spec, "controlview", False) == 0
		cnd = oUIA.controlViewCondition
	elif stringCompare(spec, "contentview", False) == 0
		cnd = oUIA.contentViewCondition
	endIf
	if cnd
		Skype__UIA___conds[spec0] = cnd
		return cnd
	endIf
endIf
var variant stack = schat__stl__stack(stringLength(spec))
var string sItem, int pos
var object c1, object c2
UIA___buildLists()
; This is used to disambiguate %1 etc. as condition versus value.
var int expectValue = False
while spec
	if stringLeft(spec, 1) == "'"
		; A string literal that may contain '' apostrophe escapes.
		; JAWS will translate the other escapes like \t, \n, etc.
		expectValue = False
		spec = stringChopLeft(spec, 1)
		sItem = ""
		pos = stringContains(spec, "'")
		if !pos
			UIA___sayError(UIA___msgStringNotTerminated, spec0)
			return
		endIf
		while substring(spec, pos+1, 1) == "'"
			; Include the first of the two apostrophes.
			sItem = formatString("%1%2", sItem, stringLeft(spec, pos))
			; Then chop both off the spec.
			spec = stringChopLeft(spec, pos+1)
			; And finally find the next one.
			pos = stringContains(spec, "'")
			if !pos
				UIA___sayError(UIA___msgStringNotTerminated, spec0)
				return
			endIf
		endWhile
		sItem = formatString("%1%2", sItem, stringLeft(spec, pos-1))
		spec = stringChopLeft(spec, pos)
		schat__stl__push(stack, oUIA.CreateStringPropertyCondition(
			schat__stl__pop(stack).value, sItem
		))
	else
		pos = stringContains(spec, " ")
		if pos
			sItem = stringLeft(spec, pos-1)
			spec = stringChopLeft(spec, pos)
		else
			sItem = spec
			spec = ""
		endIf
		; Some shortcuts for developer (and slight runtime) efficiency.
		if stringStartsWith(sItem, "fp") && stringContains("fprawview fpcontrolview fpcontentview", stringLower(sItem))
			var object c0
			if sItem == "FPRawView"
				c0 = oUIA.CreateRawViewCondition()
			elif sItem == "FPControlView"
				c0 = oUIA.CreateControlViewCondition()
			else  ; FPContentView
				c0 = oUIA.CreateContentViewCondition()
			endIf
			c0 = oUIA.CreateAndCondition(c0,
				oUIA.CreateIntPropertyCondition(UIA__propertyID("processID"),
				oUIA.GetFocusedElement().ProcessID)
			)
			schat__stl__push(stack, c0)
			expectValue = False
		elif stringCompare(sItem, "RawView", False) == 0
			schat__stl__push(stack, oUIA.CreateRawViewCondition())
			expectValue = False
		elif stringCompare(sItem, "ControlView", False) == 0
			schat__stl__push(stack, oUIA.CreateControlViewCondition())
			expectValue = False
		elif stringCompare(sItem, "ContentView", False) == 0
			schat__stl__push(stack, oUIA.CreateContentViewCondition())
			expectValue = False
		elif stringCompare(sItem, "FocusedProcess", False) == 0
			schat__stl__push(stack,
				oUIA.CreateIntPropertyCondition(UIA__propertyID("processID"),
				oUIA.GetFocusedElement().ProcessID)
			)
			expectValue = False
		elif stringCompare(sItem, "True", False) == 0
			UIA___handleTrueAndFalse(stack, oUIA, True)
			expectValue = False
		elif stringCompare(sItem, "False", False) == 0
			UIA___handleTrueAndFalse(stack, oUIA, False)
			expectValue = False
		elif stringCompare(sItem, "not", False) == 0
			schat__stl__push(stack,
				oUIA.CreateNotCondition(schat__stl__pop(stack).value)
			)
			expectValue = False
		elif stringCompare(sItem, "and", False) == 0
			c2 = schat__stl__pop(stack).value
			c1 = schat__stl__pop(stack).value
			schat__stl__push(stack, oUIA.CreateAndCondition(c1, c2))
			expectValue = False
		elif stringCompare(sItem, "or", False) == 0
			c2 = schat__stl__pop(stack).value
			c1 = schat__stl__pop(stack).value
			schat__stl__push(stack, oUIA.CreateOrCondition(c1, c2))
			expectValue = False
		elif stringToInt(sItem) || sItem == "0" || sItem == "-0"
				schat__stl__push(stack,
					oUIA.CreateIntPropertyCondition(schat__stl__pop(stack).value, stringToInt(sItem))
				)
			expectValue = False
		elif stringLeft(sItem, 1) == "%"
			var int idx = stringToInt(stringChopLeft(sItem, 1))
			var variant v
			if idx == 1 v = v1
			elif idx == 2 v = v2
			elif idx == 3 v = v3
			elif idx == 4 v = v4
			elif idx == 5 v = v5
			elif idx == 6 v = v6
			elif idx == 7 v = v7
			elif idx == 8 v = v8
			elif idx == 9 v = v9
			elif idx == 10 v = v10
			else
				UIA___sayError(formatString(UIA___msgInvalidRef, sItem), spec0)
				return
			endIf
			if expectValue
				; ToDo: Assuming references like this are strings.
				schat__stl__push(stack, oUIA.CreateStringPropertyCondition(
					schat__stl__pop(stack).value, v
				))
			else
				schat__stl__push(stack, v)
			endIf
			expectValue = False
		else
			; All that's left are property and value names.
			var int val
			if collectionItemExists(Skype__UIA___props, sItem)
				; Save for the value to compare it to.
				val = UIA__propertyID(sItem)
				schat__stl__push(stack, val)
				expectValue = True
			else
				val = UIA__valueID(sItem)
				if val == UIA__Error
					return
				endIf
				; Now we have one of those values.
				; All of these should be ints.
				schat__stl__push(stack, oUIA.CreateIntPropertyCondition(
					schat__stl__pop(stack).value, val
				))
				expectValue = False
			endIf  ; Property name or value name.
		endIf  ; list of possible keywords.
	endIf  ; leading ' or no leading '.
	spec = stringTrimLeadingBlanks(spec)
endWhile
if schat__stl__size(stack) != 1
	UIA___sayError(formatString(UIA___msgSpecIncomplete, intToString(schat__stl__size(stack)-1)), spec0)
	return
endIf
var object condition = schat__stl__pop(stack).value
if !condition
	UIA___sayError(UIA___msgUnknownError, spec0)
endIf
Skype__UIA___conds[spec0] = condition
return condition
endFunction

void function UIA___handleTrueAndFalse(variant stack, object oUIA, int val)
; Handle "false" and "true" in specs.
; Helper for UIA__condition().
var int standAlone = False
var variant prevVal
var int size = schat__stl__size(stack)
if size < 1
	standAlone = True
else
	prevVal = schat__stl__pop(stack).value
	if getVariantType(prevVal) == 9  ; VT_Dispatch
		; This should be a Condition object.
		standAlone = True
	endIf
endIf
if standAlone
	var object newVal
	if val
		if UIA__version() >= 3
			newVal = oUIA.CreateTrueCondition()
		else
			; Something that should always be True.
			newVal = oUIA.CreateRawViewCondition()
		endIf
	else
		if UIA__version() >= 3
			newVal = oUIA.CreateFalseCondition()
		else
			; Something that should always be False.
			; This is controlType = <invalidValue>.
			newVal = oUIA.CreateIntPropertyCondition(30003, 0xFFF00000L)
		endIf
	endIf
	schat__stl__push(stack, newVal)
	return
endIf
schat__stl__push(stack,
	oUIA.CreateBoolPropertyCondition(prevVal, val)
)
endFunction

variant function UIA__propertyValue(object el, variant id)
; Get a UIA property value by its ID or name.
; Returns null and speaks an error message if an invalid property name is passed.
if !stringToInt(id)
	id = UIA__propertyId(id)
	if id == UIA__Error
		return
	endIf
endIf
return el.getPropertyValue(id)
endFunction

int function UIA__propertyID(string name, optional int quiet)
; Get a UIA property ID.
; Returns the constant UIA__Error if an invalid property name is passed.
; Also speaks an error message if that happens unless quiet is True.
return UIA___getID("property", name, quiet)
endFunction

int function UIA__valueID(string name)
; Get a UIA value ID.
; Returns the constant UIA__Error if an invalid value name is passed.
; Also speaks an error message if that happens.
return UIA___getID("value", name)
endFunction

int function UIA__eventID(string name)
; Get a UIA event ID.
; Returns the constant UIA__Error if an invalid event name is passed.
; Also speaks an error message if that happens.
return UIA___getID("event", name)
endFunction

int function UIA__attributeID(string name)
; Get a UIA attribute ID.
; Returns the constant UIA__Error if an invalid attribute name is passed.
; Also speaks an error message if that happens.
return UIA___getID("attribute", name)
endFunction

int function UIA__textUnitID(string name)
; Get a UIA text unit id.
; Returns the constant UIA__Error if an invalid text unit name is passed.
; Also speaks an error message if that happens.
return UIA___getID("textUnit", name)
endFunction

collection function UIA__getList(string listType)
; Get a list by a string identifier.
; The list is returned as a collection, name to value.
UIA___buildLists()
var collection lst
if listType == "property"
	lst = Skype__UIA___props
elif listType == "value"
	lst = Skype__UIA___vals
elif listType == "event"
	lst = Skype__UIA___events
elif listType == "attribute"
	lst = Skype__UIA___attributes
elif listType == "textUnit"
	lst = Skype__UIA___textUnits
endIf
return lst
endFunction

int function UIA___getID(string listType, string name, optional int quiet)
; Get a property, value, event, attribute, or text unit id.
; Returns the constant UIA__Error if an invalid name is passed.
; Also speaks an error message if that happens unless quiet is True.
var int result = UIA___getID0(listType, name, quiet)
if result != UIA__Error
	return result
endIf
; Rebuild lists in case they were updated since the last JAWS restart.
; (This implies a script update and reload.)
UIA___buildLists(True)
result = UIA___getID0(listType, name, quiet)
if result != UIA__Error
	return result
endIf
if !quiet
	UIA___sayError(formatString(UIA___msgInvalidName, listType, name))
endIf
return UIA__Error
endFunction

int function UIA___getID0(string listType, string name, optional int quiet)
; Helper for UIA___getID().
var collection lst = UIA__getList(listType)
if stringLeft(name, 4) == "UIA_"
	name = stringChopLeft(name, 4)
endIf
if collectionItemExists(lst, name)
	return lst[name]
elif collectionItemExists(Skype__UIA___fullnames, name)
	return Skype__UIA___fullnames[name]
endIf
return UIA__Error
endFunction

void function UIA___UIAClear()
; Release the UIAScriptAPI cached object.
var variant null
Skype__UIA___gUIA = null
endFunction

void function UIA___buildLists(optional int force)
; Build internal lists of property, value, event, attribute, and textUnit Ids.
if Skype__UIA___built && !force
	return
endIf
Skype__UIA___built = False
Skype__UIA___props = new collection
Skype__UIA___vals = new collection
Skype__UIA___events = new collection
Skype__UIA___attributes = new collection
Skype__UIA___textUnits = new collection
Skype__UIA___fullnames = new collection
Skype__UIA___fullnames["UIATrue"] = 0xffffffff
Skype__UIA___vals["UIATrue"] = 0xffffffff
Skype__UIA___fullnames["UIAError"] = 0xfffe
Skype__UIA___vals["UIAError"] = 0xfffe
Skype__UIA___fullnames["ButtonControlTypeId"] = 50000
Skype__UIA___vals["Button"] = 50000
Skype__UIA___fullnames["CalendarControlTypeId"] = 50001
Skype__UIA___vals["Calendar"] = 50001
Skype__UIA___fullnames["CheckBoxControlTypeId"] = 50002
Skype__UIA___vals["CheckBox"] = 50002
Skype__UIA___fullnames["ComboBoxControlTypeId"] = 50003
Skype__UIA___vals["ComboBox"] = 50003
Skype__UIA___fullnames["EditControlTypeId"] = 50004
Skype__UIA___vals["Edit"] = 50004
Skype__UIA___fullnames["HyperlinkControlTypeId"] = 50005
Skype__UIA___vals["Hyperlink"] = 50005
Skype__UIA___fullnames["ImageControlTypeId"] = 50006
Skype__UIA___vals["Image"] = 50006
Skype__UIA___fullnames["ListItemControlTypeId"] = 50007
Skype__UIA___vals["ListItem"] = 50007
Skype__UIA___fullnames["ListControlTypeId"] = 50008
Skype__UIA___vals["List"] = 50008
Skype__UIA___fullnames["MenuControlTypeId"] = 50009
Skype__UIA___vals["Menu"] = 50009
Skype__UIA___fullnames["MenuBarControlTypeId"] = 50010
Skype__UIA___vals["MenuBar"] = 50010
Skype__UIA___fullnames["MenuItemControlTypeId"] = 50011
Skype__UIA___vals["MenuItem"] = 50011
Skype__UIA___fullnames["ProgressBarControlTypeId"] = 50012
Skype__UIA___vals["ProgressBar"] = 50012
Skype__UIA___fullnames["RadioButtonControlTypeId"] = 50013
Skype__UIA___vals["RadioButton"] = 50013
Skype__UIA___fullnames["ScrollBarControlTypeId"] = 50014
Skype__UIA___vals["ScrollBar"] = 50014
Skype__UIA___fullnames["SliderControlTypeId"] = 50015
Skype__UIA___vals["Slider"] = 50015
Skype__UIA___fullnames["SpinnerControlTypeId"] = 50016
Skype__UIA___vals["Spinner"] = 50016
Skype__UIA___fullnames["StatusBarControlTypeId"] = 50017
Skype__UIA___vals["StatusBar"] = 50017
Skype__UIA___fullnames["TabControlTypeId"] = 50018
Skype__UIA___vals["Tab"] = 50018
Skype__UIA___fullnames["TabItemControlTypeId"] = 50019
Skype__UIA___vals["TabItem"] = 50019
Skype__UIA___fullnames["TextControlTypeId"] = 50020
Skype__UIA___vals["Text"] = 50020
Skype__UIA___fullnames["ToolBarControlTypeId"] = 50021
Skype__UIA___vals["ToolBar"] = 50021
Skype__UIA___fullnames["ToolTipControlTypeId"] = 50022
Skype__UIA___vals["ToolTip"] = 50022
Skype__UIA___fullnames["TreeControlTypeId"] = 50023
Skype__UIA___vals["Tree"] = 50023
Skype__UIA___fullnames["TreeItemControlTypeId"] = 50024
Skype__UIA___vals["TreeItem"] = 50024
Skype__UIA___fullnames["CustomControlTypeId"] = 50025
Skype__UIA___vals["Custom"] = 50025
Skype__UIA___fullnames["GroupControlTypeId"] = 50026
Skype__UIA___vals["Group"] = 50026
Skype__UIA___fullnames["ThumbControlTypeId"] = 50027
Skype__UIA___vals["Thumb"] = 50027
Skype__UIA___fullnames["DataGridControlTypeId"] = 50028
Skype__UIA___vals["DataGrid"] = 50028
Skype__UIA___fullnames["DataItemControlTypeId"] = 50029
Skype__UIA___vals["DataItem"] = 50029
Skype__UIA___fullnames["DocumentControlTypeId"] = 50030
Skype__UIA___vals["Document"] = 50030
Skype__UIA___fullnames["SplitButtonControlTypeId"] = 50031
Skype__UIA___vals["SplitButton"] = 50031
Skype__UIA___fullnames["WindowControlTypeId"] = 50032
Skype__UIA___vals["Window"] = 50032
Skype__UIA___fullnames["PaneControlTypeId"] = 50033
Skype__UIA___vals["Pane"] = 50033
Skype__UIA___fullnames["HeaderControlTypeId"] = 50034
Skype__UIA___vals["Header"] = 50034
Skype__UIA___fullnames["HeaderItemControlTypeId"] = 50035
Skype__UIA___vals["HeaderItem"] = 50035
Skype__UIA___fullnames["TableControlTypeId"] = 50036
Skype__UIA___vals["Table"] = 50036
Skype__UIA___fullnames["TitleBarControlTypeId"] = 50037
Skype__UIA___vals["TitleBar"] = 50037
Skype__UIA___fullnames["SeparatorControlTypeId"] = 50038
Skype__UIA___vals["Separator"] = 50038
Skype__UIA___fullnames["SemanticZoomControlTypeId"] = 50039
Skype__UIA___vals["SemanticZoom"] = 50039
Skype__UIA___fullnames["AppBarControlTypeId"] = 50040
Skype__UIA___vals["AppBar"] = 50040
Skype__UIA___fullnames["RuntimeIdPropertyId"] = 30000
Skype__UIA___props["RuntimeId"] = 30000
Skype__UIA___fullnames["BoundingRectanglePropertyId"] = 30001
Skype__UIA___props["BoundingRectangle"] = 30001
Skype__UIA___fullnames["ProcessIdPropertyId"] = 30002
Skype__UIA___props["ProcessId"] = 30002
Skype__UIA___fullnames["ControlTypePropertyId"] = 30003
Skype__UIA___props["ControlType"] = 30003
Skype__UIA___fullnames["LocalizedControlTypePropertyId"] = 30004
Skype__UIA___props["LocalizedControlType"] = 30004
Skype__UIA___fullnames["NamePropertyId"] = 30005
Skype__UIA___props["Name"] = 30005
Skype__UIA___fullnames["AcceleratorKeyPropertyId"] = 30006
Skype__UIA___props["AcceleratorKey"] = 30006
Skype__UIA___fullnames["AccessKeyPropertyId"] = 30007
Skype__UIA___props["AccessKey"] = 30007
Skype__UIA___fullnames["HasKeyboardFocusPropertyId"] = 30008
Skype__UIA___props["HasKeyboardFocus"] = 30008
Skype__UIA___fullnames["IsKeyboardFocusablePropertyId"] = 30009
Skype__UIA___props["IsKeyboardFocusable"] = 30009
Skype__UIA___fullnames["IsEnabledPropertyId"] = 30010
Skype__UIA___props["IsEnabled"] = 30010
Skype__UIA___fullnames["AutomationIdPropertyId"] = 30011
Skype__UIA___props["AutomationId"] = 30011
Skype__UIA___fullnames["ClassNamePropertyId"] = 30012
Skype__UIA___props["ClassName"] = 30012
Skype__UIA___fullnames["HelpTextPropertyId"] = 30013
Skype__UIA___props["HelpText"] = 30013
Skype__UIA___fullnames["ClickablePointPropertyId"] = 30014
Skype__UIA___props["ClickablePoint"] = 30014
Skype__UIA___fullnames["CulturePropertyId"] = 30015
Skype__UIA___props["Culture"] = 30015
Skype__UIA___fullnames["IsControlElementPropertyId"] = 30016
Skype__UIA___props["IsControlElement"] = 30016
Skype__UIA___fullnames["IsContentElementPropertyId"] = 30017
Skype__UIA___props["IsContentElement"] = 30017
Skype__UIA___fullnames["LabeledByPropertyId"] = 30018
Skype__UIA___props["LabeledBy"] = 30018
Skype__UIA___fullnames["IsPasswordPropertyId"] = 30019
Skype__UIA___props["IsPassword"] = 30019
Skype__UIA___fullnames["NativeWindowHandlePropertyId"] = 30020
Skype__UIA___props["NativeWindowHandle"] = 30020
Skype__UIA___fullnames["ItemTypePropertyId"] = 30021
Skype__UIA___props["ItemType"] = 30021
Skype__UIA___fullnames["IsOffscreenPropertyId"] = 30022
Skype__UIA___props["IsOffscreen"] = 30022
Skype__UIA___fullnames["OrientationPropertyId"] = 30023
Skype__UIA___props["Orientation"] = 30023
Skype__UIA___fullnames["FrameworkIdPropertyId"] = 30024
Skype__UIA___props["FrameworkId"] = 30024
Skype__UIA___fullnames["IsRequiredForFormPropertyId"] = 30025
Skype__UIA___props["IsRequiredForForm"] = 30025
Skype__UIA___fullnames["ItemStatusPropertyId"] = 30026
Skype__UIA___props["ItemStatus"] = 30026
Skype__UIA___fullnames["IsDockPatternAvailablePropertyId"] = 30027
Skype__UIA___props["IsDockPatternAvailable"] = 30027
Skype__UIA___fullnames["IsExpandCollapsePatternAvailablePropertyId"] = 30028
Skype__UIA___props["IsExpandCollapsePatternAvailable"] = 30028
Skype__UIA___fullnames["IsGridItemPatternAvailablePropertyId"] = 30029
Skype__UIA___props["IsGridItemPatternAvailable"] = 30029
Skype__UIA___fullnames["IsGridPatternAvailablePropertyId"] = 30030
Skype__UIA___props["IsGridPatternAvailable"] = 30030
Skype__UIA___fullnames["IsInvokePatternAvailablePropertyId"] = 30031
Skype__UIA___props["IsInvokePatternAvailable"] = 30031
Skype__UIA___fullnames["IsMultipleViewPatternAvailablePropertyId"] = 30032
Skype__UIA___props["IsMultipleViewPatternAvailable"] = 30032
Skype__UIA___fullnames["IsRangeValuePatternAvailablePropertyId"] = 30033
Skype__UIA___props["IsRangeValuePatternAvailable"] = 30033
Skype__UIA___fullnames["IsScrollPatternAvailablePropertyId"] = 30034
Skype__UIA___props["IsScrollPatternAvailable"] = 30034
Skype__UIA___fullnames["IsScrollItemPatternAvailablePropertyId"] = 30035
Skype__UIA___props["IsScrollItemPatternAvailable"] = 30035
Skype__UIA___fullnames["IsSelectionItemPatternAvailablePropertyId"] = 30036
Skype__UIA___props["IsSelectionItemPatternAvailable"] = 30036
Skype__UIA___fullnames["IsSelectionPatternAvailablePropertyId"] = 30037
Skype__UIA___props["IsSelectionPatternAvailable"] = 30037
Skype__UIA___fullnames["IsTablePatternAvailablePropertyId"] = 30038
Skype__UIA___props["IsTablePatternAvailable"] = 30038
Skype__UIA___fullnames["IsTableItemPatternAvailablePropertyId"] = 30039
Skype__UIA___props["IsTableItemPatternAvailable"] = 30039
Skype__UIA___fullnames["IsTextPatternAvailablePropertyId"] = 30040
Skype__UIA___props["IsTextPatternAvailable"] = 30040
Skype__UIA___fullnames["IsTogglePatternAvailablePropertyId"] = 30041
Skype__UIA___props["IsTogglePatternAvailable"] = 30041
Skype__UIA___fullnames["IsTransformPatternAvailablePropertyId"] = 30042
Skype__UIA___props["IsTransformPatternAvailable"] = 30042
Skype__UIA___fullnames["IsValuePatternAvailablePropertyId"] = 30043
Skype__UIA___props["IsValuePatternAvailable"] = 30043
Skype__UIA___fullnames["IsWindowPatternAvailablePropertyId"] = 30044
Skype__UIA___props["IsWindowPatternAvailable"] = 30044
Skype__UIA___fullnames["ValueValuePropertyId"] = 30045
Skype__UIA___props["ValueValue"] = 30045
Skype__UIA___fullnames["ValueIsReadOnlyPropertyId"] = 30046
Skype__UIA___props["ValueIsReadOnly"] = 30046
Skype__UIA___fullnames["RangeValueValuePropertyId"] = 30047
Skype__UIA___props["RangeValueValue"] = 30047
Skype__UIA___fullnames["RangeValueIsReadOnlyPropertyId"] = 30048
Skype__UIA___props["RangeValueIsReadOnly"] = 30048
Skype__UIA___fullnames["RangeValueMinimumPropertyId"] = 30049
Skype__UIA___props["RangeValueMinimum"] = 30049
Skype__UIA___fullnames["RangeValueMaximumPropertyId"] = 30050
Skype__UIA___props["RangeValueMaximum"] = 30050
Skype__UIA___fullnames["RangeValueLargeChangePropertyId"] = 30051
Skype__UIA___props["RangeValueLargeChange"] = 30051
Skype__UIA___fullnames["RangeValueSmallChangePropertyId"] = 30052
Skype__UIA___props["RangeValueSmallChange"] = 30052
Skype__UIA___fullnames["ScrollHorizontalScrollPercentPropertyId"] = 30053
Skype__UIA___props["ScrollHorizontalScrollPercent"] = 30053
Skype__UIA___fullnames["ScrollHorizontalViewSizePropertyId"] = 30054
Skype__UIA___props["ScrollHorizontalViewSize"] = 30054
Skype__UIA___fullnames["ScrollVerticalScrollPercentPropertyId"] = 30055
Skype__UIA___props["ScrollVerticalScrollPercent"] = 30055
Skype__UIA___fullnames["ScrollVerticalViewSizePropertyId"] = 30056
Skype__UIA___props["ScrollVerticalViewSize"] = 30056
Skype__UIA___fullnames["ScrollHorizontallyScrollablePropertyId"] = 30057
Skype__UIA___props["ScrollHorizontallyScrollable"] = 30057
Skype__UIA___fullnames["ScrollVerticallyScrollablePropertyId"] = 30058
Skype__UIA___props["ScrollVerticallyScrollable"] = 30058
Skype__UIA___fullnames["SelectionSelectionPropertyId"] = 30059
Skype__UIA___props["SelectionSelection"] = 30059
Skype__UIA___fullnames["SelectionCanSelectMultiplePropertyId"] = 30060
Skype__UIA___props["SelectionCanSelectMultiple"] = 30060
Skype__UIA___fullnames["SelectionIsSelectionRequiredPropertyId"] = 30061
Skype__UIA___props["SelectionIsSelectionRequired"] = 30061
Skype__UIA___fullnames["GridRowCountPropertyId"] = 30062
Skype__UIA___props["GridRowCount"] = 30062
Skype__UIA___fullnames["GridColumnCountPropertyId"] = 30063
Skype__UIA___props["GridColumnCount"] = 30063
Skype__UIA___fullnames["GridItemRowPropertyId"] = 30064
Skype__UIA___props["GridItemRow"] = 30064
Skype__UIA___fullnames["GridItemColumnPropertyId"] = 30065
Skype__UIA___props["GridItemColumn"] = 30065
Skype__UIA___fullnames["GridItemRowSpanPropertyId"] = 30066
Skype__UIA___props["GridItemRowSpan"] = 30066
Skype__UIA___fullnames["GridItemColumnSpanPropertyId"] = 30067
Skype__UIA___props["GridItemColumnSpan"] = 30067
Skype__UIA___fullnames["GridItemContainingGridPropertyId"] = 30068
Skype__UIA___props["GridItemContainingGrid"] = 30068
Skype__UIA___fullnames["DockDockPositionPropertyId"] = 30069
Skype__UIA___props["DockDockPosition"] = 30069
Skype__UIA___fullnames["ExpandCollapseExpandCollapseStatePropertyId"] = 30070
Skype__UIA___props["ExpandCollapseExpandCollapseState"] = 30070
Skype__UIA___fullnames["MultipleViewCurrentViewPropertyId"] = 30071
Skype__UIA___props["MultipleViewCurrentView"] = 30071
Skype__UIA___fullnames["MultipleViewSupportedViewsPropertyId"] = 30072
Skype__UIA___props["MultipleViewSupportedViews"] = 30072
Skype__UIA___fullnames["WindowCanMaximizePropertyId"] = 30073
Skype__UIA___props["WindowCanMaximize"] = 30073
Skype__UIA___fullnames["WindowCanMinimizePropertyId"] = 30074
Skype__UIA___props["WindowCanMinimize"] = 30074
Skype__UIA___fullnames["WindowWindowVisualStatePropertyId"] = 30075
Skype__UIA___props["WindowWindowVisualState"] = 30075
Skype__UIA___fullnames["WindowWindowInteractionStatePropertyId"] = 30076
Skype__UIA___props["WindowWindowInteractionState"] = 30076
Skype__UIA___fullnames["WindowIsModalPropertyId"] = 30077
Skype__UIA___props["WindowIsModal"] = 30077
Skype__UIA___fullnames["WindowIsTopmostPropertyId"] = 30078
Skype__UIA___props["WindowIsTopmost"] = 30078
Skype__UIA___fullnames["SelectionItemIsSelectedPropertyId"] = 30079
Skype__UIA___props["SelectionItemIsSelected"] = 30079
Skype__UIA___fullnames["SelectionItemSelectionContainerPropertyId"] = 30080
Skype__UIA___props["SelectionItemSelectionContainer"] = 30080
Skype__UIA___fullnames["TableRowHeadersPropertyId"] = 30081
Skype__UIA___props["TableRowHeaders"] = 30081
Skype__UIA___fullnames["TableColumnHeadersPropertyId"] = 30082
Skype__UIA___props["TableColumnHeaders"] = 30082
Skype__UIA___fullnames["TableRowOrColumnMajorPropertyId"] = 30083
Skype__UIA___props["TableRowOrColumnMajor"] = 30083
Skype__UIA___fullnames["TableItemRowHeaderItemsPropertyId"] = 30084
Skype__UIA___props["TableItemRowHeaderItems"] = 30084
Skype__UIA___fullnames["TableItemColumnHeaderItemsPropertyId"] = 30085
Skype__UIA___props["TableItemColumnHeaderItems"] = 30085
Skype__UIA___fullnames["ToggleToggleStatePropertyId"] = 30086
Skype__UIA___props["ToggleToggleState"] = 30086
Skype__UIA___fullnames["TransformCanMovePropertyId"] = 30087
Skype__UIA___props["TransformCanMove"] = 30087
Skype__UIA___fullnames["TransformCanResizePropertyId"] = 30088
Skype__UIA___props["TransformCanResize"] = 30088
Skype__UIA___fullnames["TransformCanRotatePropertyId"] = 30089
Skype__UIA___props["TransformCanRotate"] = 30089
Skype__UIA___fullnames["IsLegacyIAccessiblePatternAvailablePropertyId"] = 30090
Skype__UIA___props["IsLegacyIAccessiblePatternAvailable"] = 30090
Skype__UIA___fullnames["LegacyIAccessibleChildIdPropertyId"] = 30091
Skype__UIA___props["LegacyIAccessibleChildId"] = 30091
Skype__UIA___fullnames["LegacyIAccessibleNamePropertyId"] = 30092
Skype__UIA___props["LegacyIAccessibleName"] = 30092
Skype__UIA___fullnames["LegacyIAccessibleValuePropertyId"] = 30093
Skype__UIA___props["LegacyIAccessibleValue"] = 30093
Skype__UIA___fullnames["LegacyIAccessibleDescriptionPropertyId"] = 30094
Skype__UIA___props["LegacyIAccessibleDescription"] = 30094
Skype__UIA___fullnames["LegacyIAccessibleRolePropertyId"] = 30095
Skype__UIA___props["LegacyIAccessibleRole"] = 30095
Skype__UIA___fullnames["LegacyIAccessibleStatePropertyId"] = 30096
Skype__UIA___props["LegacyIAccessibleState"] = 30096
Skype__UIA___fullnames["LegacyIAccessibleHelpPropertyId"] = 30097
Skype__UIA___props["LegacyIAccessibleHelp"] = 30097
Skype__UIA___fullnames["LegacyIAccessibleKeyboardShortcutPropertyId"] = 30098
Skype__UIA___props["LegacyIAccessibleKeyboardShortcut"] = 30098
Skype__UIA___fullnames["LegacyIAccessibleSelectionPropertyId"] = 30099
Skype__UIA___props["LegacyIAccessibleSelection"] = 30099
Skype__UIA___fullnames["LegacyIAccessibleDefaultActionPropertyId"] = 30100
Skype__UIA___props["LegacyIAccessibleDefaultAction"] = 30100
Skype__UIA___fullnames["AriaRolePropertyId"] = 30101
Skype__UIA___props["AriaRole"] = 30101
Skype__UIA___fullnames["AriaPropertiesPropertyId"] = 30102
Skype__UIA___props["AriaProperties"] = 30102
Skype__UIA___fullnames["IsDataValidForFormPropertyId"] = 30103
Skype__UIA___props["IsDataValidForForm"] = 30103
Skype__UIA___fullnames["ControllerForPropertyId"] = 30104
Skype__UIA___props["ControllerFor"] = 30104
Skype__UIA___fullnames["DescribedByPropertyId"] = 30105
Skype__UIA___props["DescribedBy"] = 30105
Skype__UIA___fullnames["FlowsToPropertyId"] = 30106
Skype__UIA___props["FlowsTo"] = 30106
Skype__UIA___fullnames["ProviderDescriptionPropertyId"] = 30107
Skype__UIA___props["ProviderDescription"] = 30107
Skype__UIA___fullnames["IsItemContainerPatternAvailablePropertyId"] = 30108
Skype__UIA___props["IsItemContainerPatternAvailable"] = 30108
Skype__UIA___fullnames["IsVirtualizedItemPatternAvailablePropertyId"] = 30109
Skype__UIA___props["IsVirtualizedItemPatternAvailable"] = 30109
Skype__UIA___fullnames["IsSynchronizedInputPatternAvailablePropertyId"] = 30110
Skype__UIA___props["IsSynchronizedInputPatternAvailable"] = 30110
Skype__UIA___fullnames["OptimizeForVisualContentPropertyId"] = 30111
Skype__UIA___props["OptimizeForVisualContent"] = 30111
Skype__UIA___fullnames["IsObjectModelPatternAvailablePropertyId"] = 30112
Skype__UIA___props["IsObjectModelPatternAvailable"] = 30112
Skype__UIA___fullnames["AnnotationAnnotationTypeIdPropertyId"] = 30113
Skype__UIA___props["AnnotationAnnotationTypeId"] = 30113
Skype__UIA___fullnames["AnnotationAnnotationTypeNamePropertyId"] = 30114
Skype__UIA___props["AnnotationAnnotationTypeName"] = 30114
Skype__UIA___fullnames["AnnotationAuthorPropertyId"] = 30115
Skype__UIA___props["AnnotationAuthor"] = 30115
Skype__UIA___fullnames["AnnotationDateTimePropertyId"] = 30116
Skype__UIA___props["AnnotationDateTime"] = 30116
Skype__UIA___fullnames["AnnotationTargetPropertyId"] = 30117
Skype__UIA___props["AnnotationTarget"] = 30117
Skype__UIA___fullnames["IsAnnotationPatternAvailablePropertyId"] = 30118
Skype__UIA___props["IsAnnotationPatternAvailable"] = 30118
Skype__UIA___fullnames["IsTextPattern2AvailablePropertyId"] = 30119
Skype__UIA___props["IsTextPattern2Available"] = 30119
Skype__UIA___fullnames["StylesStyleIdPropertyId"] = 30120
Skype__UIA___props["StylesStyleId"] = 30120
Skype__UIA___fullnames["StylesStyleNamePropertyId"] = 30121
Skype__UIA___props["StylesStyleName"] = 30121
Skype__UIA___fullnames["StylesFillColorPropertyId"] = 30122
Skype__UIA___props["StylesFillColor"] = 30122
Skype__UIA___fullnames["StylesFillPatternStylePropertyId"] = 30123
Skype__UIA___props["StylesFillPatternStyle"] = 30123
Skype__UIA___fullnames["StylesShapePropertyId"] = 30124
Skype__UIA___props["StylesShape"] = 30124
Skype__UIA___fullnames["StylesFillPatternColorPropertyId"] = 30125
Skype__UIA___props["StylesFillPatternColor"] = 30125
Skype__UIA___fullnames["StylesExtendedPropertiesPropertyId"] = 30126
Skype__UIA___props["StylesExtendedProperties"] = 30126
Skype__UIA___fullnames["IsStylesPatternAvailablePropertyId"] = 30127
Skype__UIA___props["IsStylesPatternAvailable"] = 30127
Skype__UIA___fullnames["IsSpreadsheetPatternAvailablePropertyId"] = 30128
Skype__UIA___props["IsSpreadsheetPatternAvailable"] = 30128
Skype__UIA___fullnames["SpreadsheetItemFormulaPropertyId"] = 30129
Skype__UIA___props["SpreadsheetItemFormula"] = 30129
Skype__UIA___fullnames["SpreadsheetItemAnnotationObjectsPropertyId"] = 30130
Skype__UIA___props["SpreadsheetItemAnnotationObjects"] = 30130
Skype__UIA___fullnames["SpreadsheetItemAnnotationTypesPropertyId"] = 30131
Skype__UIA___props["SpreadsheetItemAnnotationTypes"] = 30131
Skype__UIA___fullnames["IsSpreadsheetItemPatternAvailablePropertyId"] = 30132
Skype__UIA___props["IsSpreadsheetItemPatternAvailable"] = 30132
Skype__UIA___fullnames["Transform2CanZoomPropertyId"] = 30133
Skype__UIA___props["Transform2CanZoom"] = 30133
Skype__UIA___fullnames["IsTransformPattern2AvailablePropertyId"] = 30134
Skype__UIA___props["IsTransformPattern2Available"] = 30134
Skype__UIA___fullnames["LiveSettingPropertyId"] = 30135
Skype__UIA___props["LiveSetting"] = 30135
Skype__UIA___fullnames["IsTextChildPatternAvailablePropertyId"] = 30136
Skype__UIA___props["IsTextChildPatternAvailable"] = 30136
Skype__UIA___fullnames["IsDragPatternAvailablePropertyId"] = 30137
Skype__UIA___props["IsDragPatternAvailable"] = 30137
Skype__UIA___fullnames["DragIsGrabbedPropertyId"] = 30138
Skype__UIA___props["DragIsGrabbed"] = 30138
Skype__UIA___fullnames["DragDropEffectPropertyId"] = 30139
Skype__UIA___props["DragDropEffect"] = 30139
Skype__UIA___fullnames["DragDropEffectsPropertyId"] = 30140
Skype__UIA___props["DragDropEffects"] = 30140
Skype__UIA___fullnames["IsDropTargetPatternAvailablePropertyId"] = 30141
Skype__UIA___props["IsDropTargetPatternAvailable"] = 30141
Skype__UIA___fullnames["DropTargetDropTargetEffectPropertyId"] = 30142
Skype__UIA___props["DropTargetDropTargetEffect"] = 30142
Skype__UIA___fullnames["DropTargetDropTargetEffectsPropertyId"] = 30143
Skype__UIA___props["DropTargetDropTargetEffects"] = 30143
Skype__UIA___fullnames["DragGrabbedItemsPropertyId"] = 30144
Skype__UIA___props["DragGrabbedItems"] = 30144
Skype__UIA___fullnames["Transform2ZoomLevelPropertyId"] = 30145
Skype__UIA___props["Transform2ZoomLevel"] = 30145
Skype__UIA___fullnames["Transform2ZoomMinimumPropertyId"] = 30146
Skype__UIA___props["Transform2ZoomMinimum"] = 30146
Skype__UIA___fullnames["Transform2ZoomMaximumPropertyId"] = 30147
Skype__UIA___props["Transform2ZoomMaximum"] = 30147
Skype__UIA___fullnames["FlowsFromPropertyId"] = 30148
Skype__UIA___props["FlowsFrom"] = 30148
Skype__UIA___fullnames["IsTextEditPatternAvailablePropertyId"] = 30149
Skype__UIA___props["IsTextEditPatternAvailable"] = 30149
Skype__UIA___fullnames["IsPeripheralPropertyId"] = 30150
Skype__UIA___props["IsPeripheral"] = 30150
Skype__UIA___fullnames["PositionInSetPropertyId"] = 30152
Skype__UIA___props["PositionInSet"] = 30152
Skype__UIA___fullnames["SizeOfSetPropertyId"] = 30153
Skype__UIA___props["SizeOfSet"] = 30153
Skype__UIA___fullnames["LevelPropertyId"] = 30154
Skype__UIA___props["Level"] = 30154
Skype__UIA___fullnames["AnnotationTypesPropertyId"] = 30155
Skype__UIA___props["AnnotationTypes"] = 30155
Skype__UIA___fullnames["AnnotationObjectsPropertyId"] = 30156
Skype__UIA___props["AnnotationObjects"] = 30156
Skype__UIA___fullnames["LandmarkTypePropertyId"] = 30157
Skype__UIA___props["LandmarkType"] = 30157
Skype__UIA___fullnames["LocalizedLandmarkTypePropertyId"] = 30158
Skype__UIA___props["LocalizedLandmarkType"] = 30158
Skype__UIA___fullnames["FullDescriptionPropertyId"] = 30159
Skype__UIA___props["FullDescription"] = 30159
Skype__UIA___fullnames["IsCustomNavigationPatternAvailablePropertyId"] = 30151
Skype__UIA___props["IsCustomNavigationPatternAvailable"] = 30151
Skype__UIA___fullnames["FillColorPropertyId"] = 30160
Skype__UIA___props["FillColor"] = 30160
Skype__UIA___fullnames["OutlineColorPropertyId"] = 30161
Skype__UIA___props["OutlineColor"] = 30161
Skype__UIA___fullnames["FillTypePropertyId"] = 30162
Skype__UIA___props["FillType"] = 30162
Skype__UIA___fullnames["VisualEffectsPropertyId"] = 30163
Skype__UIA___props["VisualEffects"] = 30163
Skype__UIA___fullnames["OutlineThicknessPropertyId"] = 30164
Skype__UIA___props["OutlineThickness"] = 30164
Skype__UIA___fullnames["CenterPointPropertyId"] = 30165
Skype__UIA___props["CenterPoint"] = 30165
Skype__UIA___fullnames["RotationPropertyId"] = 30166
Skype__UIA___props["Rotation"] = 30166
Skype__UIA___fullnames["SizePropertyId"] = 30167
Skype__UIA___props["Size"] = 30167
Skype__UIA___fullnames["UIA_AnnotationType_Unknown"] = 60000
Skype__UIA___vals["UIA_AnnotationType_Unknown"] = 60000
Skype__UIA___fullnames["UIA_AnnotationType_SpellingError"] = 60001
Skype__UIA___vals["UIA_AnnotationType_SpellingError"] = 60001
Skype__UIA___fullnames["UIA_AnnotationType_GrammarError"] = 60002
Skype__UIA___vals["UIA_AnnotationType_GrammarError"] = 60002
Skype__UIA___fullnames["UIA_AnnotationType_Comment"] = 60003
Skype__UIA___vals["UIA_AnnotationType_Comment"] = 60003
Skype__UIA___fullnames["UIA_AnnotationType_FormulaError"] = 60004
Skype__UIA___vals["UIA_AnnotationType_FormulaError"] = 60004
Skype__UIA___fullnames["UIA_AnnotationType_TrackChanges"] = 60005
Skype__UIA___vals["UIA_AnnotationType_TrackChanges"] = 60005
Skype__UIA___fullnames["UIA_AnnotationType_Header"] = 60006
Skype__UIA___vals["UIA_AnnotationType_Header"] = 60006
Skype__UIA___fullnames["UIA_AnnotationType_Footer"] = 60007
Skype__UIA___vals["UIA_AnnotationType_Footer"] = 60007
Skype__UIA___fullnames["UIA_AnnotationType_Highlighted"] = 60008
Skype__UIA___vals["UIA_AnnotationType_Highlighted"] = 60008
Skype__UIA___fullnames["UIA_AnnotationType_Endnote"] = 60009
Skype__UIA___vals["UIA_AnnotationType_Endnote"] = 60009
Skype__UIA___fullnames["UIA_AnnotationType_Footnote"] = 60010
Skype__UIA___vals["UIA_AnnotationType_Footnote"] = 60010
Skype__UIA___fullnames["UIA_AnnotationType_InsertionChange"] = 60011
Skype__UIA___vals["UIA_AnnotationType_InsertionChange"] = 60011
Skype__UIA___fullnames["UIA_AnnotationType_DeletionChange"] = 60012
Skype__UIA___vals["UIA_AnnotationType_DeletionChange"] = 60012
Skype__UIA___fullnames["UIA_AnnotationType_MoveChange"] = 60013
Skype__UIA___vals["UIA_AnnotationType_MoveChange"] = 60013
Skype__UIA___fullnames["UIA_AnnotationType_FormatChange"] = 60014
Skype__UIA___vals["UIA_AnnotationType_FormatChange"] = 60014
Skype__UIA___fullnames["UIA_AnnotationType_UnsyncedChange"] = 60015
Skype__UIA___vals["UIA_AnnotationType_UnsyncedChange"] = 60015
Skype__UIA___fullnames["UIA_AnnotationType_EditingLockedChange"] = 60016
Skype__UIA___vals["UIA_AnnotationType_EditingLockedChange"] = 60016
Skype__UIA___fullnames["UIA_AnnotationType_ExternalChange"] = 60017
Skype__UIA___vals["UIA_AnnotationType_ExternalChange"] = 60017
Skype__UIA___fullnames["UIA_AnnotationType_ConflictingChange"] = 60018
Skype__UIA___vals["UIA_AnnotationType_ConflictingChange"] = 60018
Skype__UIA___fullnames["UIA_AnnotationType_Author"] = 60019
Skype__UIA___vals["UIA_AnnotationType_Author"] = 60019
Skype__UIA___fullnames["UIA_AnnotationType_AdvancedProofingIssue"] = 60020
Skype__UIA___vals["UIA_AnnotationType_AdvancedProofingIssue"] = 60020
Skype__UIA___fullnames["UIA_AnnotationType_DataValidationError"] = 60021
Skype__UIA___vals["UIA_AnnotationType_DataValidationError"] = 60021
Skype__UIA___fullnames["UIA_AnnotationType_CircularReferenceError"] = 60022
Skype__UIA___vals["UIA_AnnotationType_CircularReferenceError"] = 60022
Skype__UIA___fullnames["ToolTipOpenedEventId"] = 20000
Skype__UIA___events["ToolTipOpened"] = 20000
Skype__UIA___fullnames["ToolTipClosedEventId"] = 20001
Skype__UIA___events["ToolTipClosed"] = 20001
Skype__UIA___fullnames["StructureChangedEventId"] = 20002
Skype__UIA___events["StructureChanged"] = 20002
Skype__UIA___fullnames["MenuOpenedEventId"] = 20003
Skype__UIA___events["MenuOpened"] = 20003
Skype__UIA___fullnames["AutomationPropertyChangedEventId"] = 20004
Skype__UIA___events["AutomationPropertyChanged"] = 20004
Skype__UIA___fullnames["AutomationFocusChangedEventId"] = 20005
Skype__UIA___events["AutomationFocusChanged"] = 20005
Skype__UIA___fullnames["AsyncContentLoadedEventId"] = 20006
Skype__UIA___events["AsyncContentLoaded"] = 20006
Skype__UIA___fullnames["MenuClosedEventId"] = 20007
Skype__UIA___events["MenuClosed"] = 20007
Skype__UIA___fullnames["LayoutInvalidatedEventId"] = 20008
Skype__UIA___events["LayoutInvalidated"] = 20008
Skype__UIA___fullnames["Invoke_InvokedEventId"] = 20009
Skype__UIA___events["Invoke_Invoked"] = 20009
Skype__UIA___fullnames["SelectionItem_ElementAddedToSelectionEventId"] = 20010
Skype__UIA___events["SelectionItem_ElementAddedToSelection"] = 20010
Skype__UIA___fullnames["SelectionItem_ElementRemovedFromSelectionEventId"] = 20011
Skype__UIA___events["SelectionItem_ElementRemovedFromSelection"] = 20011
Skype__UIA___fullnames["SelectionItem_ElementSelectedEventId"] = 20012
Skype__UIA___events["SelectionItem_ElementSelected"] = 20012
Skype__UIA___fullnames["Selection_InvalidatedEventId"] = 20013
Skype__UIA___events["Selection_Invalidated"] = 20013
Skype__UIA___fullnames["Text_TextSelectionChangedEventId"] = 20014
Skype__UIA___events["Text_TextSelectionChanged"] = 20014
Skype__UIA___fullnames["Text_TextChangedEventId"] = 20015
Skype__UIA___events["Text_TextChanged"] = 20015
Skype__UIA___fullnames["Window_WindowOpenedEventId"] = 20016
Skype__UIA___events["Window_WindowOpened"] = 20016
Skype__UIA___fullnames["Window_WindowClosedEventId"] = 20017
Skype__UIA___events["Window_WindowClosed"] = 20017
Skype__UIA___fullnames["MenuModeStartEventId"] = 20018
Skype__UIA___events["MenuModeStart"] = 20018
Skype__UIA___fullnames["MenuModeEndEventId"] = 20019
Skype__UIA___events["MenuModeEnd"] = 20019
Skype__UIA___fullnames["InputReachedTargetEventId"] = 20020
Skype__UIA___events["InputReachedTarget"] = 20020
Skype__UIA___fullnames["InputReachedOtherElementEventId"] = 20021
Skype__UIA___events["InputReachedOtherElement"] = 20021
Skype__UIA___fullnames["InputDiscardedEventId"] = 20022
Skype__UIA___events["InputDiscarded"] = 20022
Skype__UIA___fullnames["SystemAlertEventId"] = 20023
Skype__UIA___events["SystemAlert"] = 20023
Skype__UIA___fullnames["LiveRegionChangedEventId"] = 20024
Skype__UIA___events["LiveRegionChanged"] = 20024
Skype__UIA___fullnames["HostedFragmentRootsInvalidatedEventId"] = 20025
Skype__UIA___events["HostedFragmentRootsInvalidated"] = 20025
Skype__UIA___fullnames["DragStartEventId"] = 20026
Skype__UIA___events["DragStart"] = 20026
Skype__UIA___fullnames["DragCancelEventId"] = 20027
Skype__UIA___events["DragCancel"] = 20027
Skype__UIA___fullnames["Drag_CompleteEventId"] = 20028
Skype__UIA___events["Drag_Complete"] = 20028
Skype__UIA___fullnames["DragEnterEventId"] = 20029
Skype__UIA___events["DragEnter"] = 20029
Skype__UIA___fullnames["DragLeaveEventId"] = 20030
Skype__UIA___events["DragLeave"] = 20030
Skype__UIA___fullnames["DroppedEventId"] = 20031
Skype__UIA___events["Dropped"] = 20031
Skype__UIA___fullnames["TextEdit_TextChangedEventId"] = 20032
Skype__UIA___events["TextEdit_TextChanged"] = 20032
Skype__UIA___fullnames["TextEdit_ConversionTargetChangedEventId"] = 20033
Skype__UIA___events["TextEdit_ConversionTargetChanged"] = 20033
Skype__UIA___fullnames["OrientationType_None"] = 0
Skype__UIA___vals["OrientationType_None"] = 0
Skype__UIA___fullnames["OrientationType_Horizontal"] = 1
Skype__UIA___vals["OrientationType_Horizontal"] = 1
Skype__UIA___fullnames["OrientationType_Vertical"] = 2
Skype__UIA___vals["OrientationType_Vertical"] = 2
Skype__UIA___fullnames["TreeScope_Element"] = 1
Skype__UIA___vals["TreeScope_Element"] = 1
Skype__UIA___fullnames["TreeScope_Children"] = 2
Skype__UIA___vals["TreeScope_Children"] = 2
Skype__UIA___fullnames["TreeScope_Descendants"] = 4
Skype__UIA___vals["TreeScope_Descendants"] = 4
Skype__UIA___fullnames["TreeScope_Parent"] = 8
Skype__UIA___vals["TreeScope_Parent"] = 8
Skype__UIA___fullnames["TreeScope_Ancestors"] = 16
Skype__UIA___vals["TreeScope_Ancestors"] = 16
Skype__UIA___fullnames["TreeScope_Subtree"] = 7
Skype__UIA___vals["TreeScope_Subtree"] = 7
Skype__UIA___fullnames["next"] = 1
Skype__UIA___vals["next"] = 1
Skype__UIA___fullnames["prior"] = 0
Skype__UIA___vals["prior"] = 0
Skype__UIA___fullnames["ScrollAmount_LargeDecrement"] = 0
Skype__UIA___vals["ScrollAmount_LargeDecrement"] = 0
Skype__UIA___fullnames["ScrollAmount_SmallDecrement"] = 1
Skype__UIA___vals["ScrollAmount_SmallDecrement"] = 1
Skype__UIA___fullnames["ScrollAmount_NoAmount"] = 2
Skype__UIA___vals["ScrollAmount_NoAmount"] = 2
Skype__UIA___fullnames["ScrollAmount_LargeIncrement"] = 3
Skype__UIA___vals["ScrollAmount_LargeIncrement"] = 3
Skype__UIA___fullnames["ScrollAmount_SmallIncrement"] = 4
Skype__UIA___vals["ScrollAmount_SmallIncrement"] = 4
Skype__UIA___fullnames["StructureChangeType_ChildAdded"] = 0
Skype__UIA___vals["StructureChangeType_ChildAdded"] = 0
Skype__UIA___fullnames["StructureChangeType_ChildRemoved"] = 1
Skype__UIA___vals["StructureChangeType_ChildRemoved"] = 1
Skype__UIA___fullnames["StructureChangeType_ChildrenInvalidated"] = 2
Skype__UIA___vals["StructureChangeType_ChildrenInvalidated"] = 2
Skype__UIA___fullnames["StructureChangeType_ChildrenBulkAdded"] = 3
Skype__UIA___vals["StructureChangeType_ChildrenBulkAdded"] = 3
Skype__UIA___fullnames["StructureChangeType_ChildrenBulkRemoved"] = 4
Skype__UIA___vals["StructureChangeType_ChildrenBulkRemoved"] = 4
Skype__UIA___fullnames["StructureChangeType_ChildrenReordered"] = 5
Skype__UIA___vals["StructureChangeType_ChildrenReordered"] = 5
Skype__UIA___fullnames["AnimationStyleAttributeId"] = 40000
Skype__UIA___attributes["AnimationStyle"] = 40000
Skype__UIA___fullnames["AnnotationObjectsAttributeId"] = 40032
Skype__UIA___attributes["AnnotationObjects"] = 40032
Skype__UIA___fullnames["AnnotationTypesAttributeId"] = 40031
Skype__UIA___attributes["AnnotationTypes"] = 40031
Skype__UIA___fullnames["BackgroundColorAttributeId"] = 40001
Skype__UIA___attributes["BackgroundColor"] = 40001
Skype__UIA___fullnames["BulletStyleAttributeId"] = 40002
Skype__UIA___attributes["BulletStyle"] = 40002
Skype__UIA___fullnames["CapStyleAttributeId"] = 40003
Skype__UIA___attributes["CapStyle"] = 40003
Skype__UIA___fullnames["CaretBidiModeAttributeId"] = 40039
Skype__UIA___attributes["CaretBidiMode"] = 40039
Skype__UIA___fullnames["CaretPositionAttributeId"] = 40038
Skype__UIA___attributes["CaretPosition"] = 40038
Skype__UIA___fullnames["CultureAttributeId"] = 40004
Skype__UIA___attributes["Culture"] = 40004
Skype__UIA___fullnames["FontNameAttributeId"] = 40005
Skype__UIA___attributes["FontName"] = 40005
Skype__UIA___fullnames["FontSizeAttributeId"] = 40006
Skype__UIA___attributes["FontSize"] = 40006
Skype__UIA___fullnames["FontWeightAttributeId"] = 400007
Skype__UIA___attributes["FontWeight"] = 400007
Skype__UIA___fullnames["ForegroundColorAttributeId"] = 40008
Skype__UIA___attributes["ForegroundColor"] = 40008
Skype__UIA___fullnames["HorizontalTextAlignmentAttributeId"] = 40009
Skype__UIA___attributes["HorizontalTextAlignment"] = 40009
Skype__UIA___fullnames["IndentationFirstLineAttributeId"] = 40010
Skype__UIA___attributes["IndentationFirstLine"] = 40010
Skype__UIA___fullnames["IndentationLeadingAttributeId"] = 40011
Skype__UIA___attributes["IndentationLeading"] = 40011
Skype__UIA___fullnames["IndentationTrailingAttributeId"] = 40012
Skype__UIA___attributes["IndentationTrailing"] = 40012
Skype__UIA___fullnames["IsActiveAttributeId"] = 40036
Skype__UIA___attributes["IsActive"] = 40036
Skype__UIA___fullnames["IsHiddenAttributeId"] = 40013
Skype__UIA___attributes["IsHidden"] = 40013
Skype__UIA___fullnames["IsItalicAttributeId"] = 40014
Skype__UIA___attributes["IsItalic"] = 40014
Skype__UIA___fullnames["IsReadOnlyAttributeId"] = 40015
Skype__UIA___attributes["IsReadOnly"] = 40015
Skype__UIA___fullnames["IsSubscriptAttributeId"] = 40016
Skype__UIA___attributes["IsSubscript"] = 40016
Skype__UIA___fullnames["IsSuperscriptAttributeId"] = 40017
Skype__UIA___attributes["IsSuperscript"] = 40017
Skype__UIA___fullnames["LinkAttributeId"] = 40035
Skype__UIA___attributes["Link"] = 40035
Skype__UIA___fullnames["MarginBottomAttributeId"] = 40018
Skype__UIA___attributes["MarginBottom"] = 40018
Skype__UIA___fullnames["MarginLeadingAttributeId"] = 40019
Skype__UIA___attributes["MarginLeading"] = 40019
Skype__UIA___fullnames["MarginTopAttributeId"] = 40020
Skype__UIA___attributes["MarginTop"] = 40020
Skype__UIA___fullnames["MarginTrailingAttributeId"] = 40021
Skype__UIA___attributes["MarginTrailing"] = 40021
Skype__UIA___fullnames["OutlineStylesAttributeId"] = 40022
Skype__UIA___attributes["OutlineStyles"] = 40022
Skype__UIA___fullnames["OverlineColorAttributeId"] = 40023
Skype__UIA___attributes["OverlineColor"] = 40023
Skype__UIA___fullnames["OverlineStyleAttributeId"] = 40024
Skype__UIA___attributes["OverlineStyle"] = 40024
Skype__UIA___fullnames["SelectionActiveEndAttributeId"] = 40037
Skype__UIA___attributes["SelectionActiveEnd"] = 40037
Skype__UIA___fullnames["StrikethroughColorAttributeId"] = 40025
Skype__UIA___attributes["StrikethroughColor"] = 40025
Skype__UIA___fullnames["StrikethroughStyleAttributeId"] = 40026
Skype__UIA___attributes["StrikethroughStyle"] = 40026
Skype__UIA___fullnames["StyleIdAttributeId"] = 40034
Skype__UIA___attributes["StyleId"] = 40034
Skype__UIA___fullnames["StyleNameAttributeId"] = 40033
Skype__UIA___attributes["StyleName"] = 40033
Skype__UIA___fullnames["TabsAttributeId"] = 40027
Skype__UIA___attributes["Tabs"] = 40027
Skype__UIA___fullnames["TextFlowDirectionsAttributeId"] = 40028
Skype__UIA___attributes["TextFlowDirections"] = 40028
Skype__UIA___fullnames["UnderlineColorAttributeId"] = 40029
Skype__UIA___attributes["UnderlineColor"] = 40029
Skype__UIA___fullnames["UnderlineStyleAttributeId"] = 40030
Skype__UIA___attributes["UnderlineStyle"] = 40030
Skype__UIA___fullnames["TextUnit_Character"] = 0
Skype__UIA___textUnits["Character"] = 0
Skype__UIA___fullnames["TextUnit_Format"] = 1
Skype__UIA___textUnits["Format"] = 1
Skype__UIA___fullnames["TextUnit_Word"] = 2
Skype__UIA___textUnits["Word"] = 2
Skype__UIA___fullnames["TextUnit_Line"] = 3
Skype__UIA___textUnits["Line"] = 3
Skype__UIA___fullnames["TextUnit_Paragraph"] = 4
Skype__UIA___textUnits["Paragraph"] = 4
Skype__UIA___fullnames["TextUnit_Page"] = 5
Skype__UIA___textUnits["Page"] = 5
Skype__UIA___fullnames["TextUnit_Document"] = 6
Skype__UIA___textUnits["Document"] = 6
Skype__UIA___fullnames["TextPatternRangeEndpoint_Start"] = 0
Skype__UIA___vals["TextPatternRangeEndpoint_Start"] = 0
Skype__UIA___fullnames["TextPatternRangeEndpoint_End"] = 1
Skype__UIA___vals["TextPatternRangeEndpoint_End"] = 1
Skype__UIA___fullnames["SupportedTextSelection_None"] = 0
Skype__UIA___vals["SupportedTextSelection_None"] = 0
Skype__UIA___fullnames["SupportedTextSelection_Single"] = 1
Skype__UIA___vals["SupportedTextSelection_Single"] = 1
Skype__UIA___fullnames["SupportedTextSelection_Multiple"] = 2
Skype__UIA___vals["SupportedTextSelection_Multiple"] = 2
Skype__UIA___fullnames["DockPosition_Top"] = 0
Skype__UIA___vals["DockPosition_Top"] = 0
Skype__UIA___fullnames["DockPosition_Left"] = 1
Skype__UIA___vals["DockPosition_Left"] = 1
Skype__UIA___fullnames["DockPosition_Bottom"] = 2
Skype__UIA___vals["DockPosition_Bottom"] = 2
Skype__UIA___fullnames["DockPosition_Right"] = 3
Skype__UIA___vals["DockPosition_Right"] = 3
Skype__UIA___fullnames["DockPosition_Fill"] = 4
Skype__UIA___vals["DockPosition_Fill"] = 4
Skype__UIA___fullnames["DockPosition_None"] = 5
Skype__UIA___vals["DockPosition_None"] = 5
Skype__UIA___fullnames["SynchronizedInputType_KeyUp"] = 0
Skype__UIA___vals["SynchronizedInputType_KeyUp"] = 0
Skype__UIA___fullnames["SynchronizedInputType_KeyDown"] = 1
Skype__UIA___vals["SynchronizedInputType_KeyDown"] = 1
Skype__UIA___fullnames["SynchronizedInputType_LeftMouseUp"] = 2
Skype__UIA___vals["SynchronizedInputType_LeftMouseUp"] = 2
Skype__UIA___fullnames["SynchronizedInputType_LeftMouseDown"] = 3
Skype__UIA___vals["SynchronizedInputType_LeftMouseDown"] = 3
Skype__UIA___fullnames["SynchronizedInputType_RightMouseUp"] = 4
Skype__UIA___vals["SynchronizedInputType_RightMouseUp"] = 4
Skype__UIA___fullnames["SynchronizedInputType_RightMouseDown"] = 5
Skype__UIA___vals["SynchronizedInputType_RightMouseDown"] = 5
Skype__UIA___fullnames["RowOrColumnMajor_RowMajor"] = 0
Skype__UIA___vals["RowOrColumnMajor_RowMajor"] = 0
Skype__UIA___fullnames["RowOrColumnMajor_ColumnMajor"] = 1
Skype__UIA___vals["RowOrColumnMajor_ColumnMajor"] = 1
Skype__UIA___fullnames["RowOrColumnMajor_Indeterminate"] = 2
Skype__UIA___vals["RowOrColumnMajor_Indeterminate"] = 2
Skype__UIA___fullnames["ZoomUnit_NoAmount"] = 0
Skype__UIA___vals["ZoomUnit_NoAmount"] = 0
Skype__UIA___fullnames["ZoomUnit_LargeDecrement"] = 1
Skype__UIA___vals["ZoomUnit_LargeDecrement"] = 1
Skype__UIA___fullnames["ZoomUnit_SmallDecrement"] = 2
Skype__UIA___vals["ZoomUnit_SmallDecrement"] = 2
Skype__UIA___fullnames["ZoomUnit_LargeIncrement"] = 3
Skype__UIA___vals["ZoomUnit_LargeIncrement"] = 3
Skype__UIA___fullnames["ZoomUnit_SmallIncrement"] = 4
Skype__UIA___vals["ZoomUnit_SmallIncrement"] = 4
Skype__UIA___fullnames["WindowVisualState_Normal"] = 0
Skype__UIA___vals["WindowVisualState_Normal"] = 0
Skype__UIA___fullnames["WindowVisualState_Maximized"] = 1
Skype__UIA___vals["WindowVisualState_Maximized"] = 1
Skype__UIA___fullnames["WindowVisualState_Minimized"] = 2
Skype__UIA___vals["WindowVisualState_Minimized"] = 2
Skype__UIA___fullnames["WindowInteractionState_Running"] = 0
Skype__UIA___vals["WindowInteractionState_Running"] = 0
Skype__UIA___fullnames["WindowInteractionState_Closing"] = 1
Skype__UIA___vals["WindowInteractionState_Closing"] = 1
Skype__UIA___fullnames["WindowInteractionState_ReadyForUserInteraction"] = 2
Skype__UIA___vals["WindowInteractionState_ReadyForUserInteraction"] = 2
Skype__UIA___fullnames["WindowInteractionState_BlockedByModalWindow"] = 3
Skype__UIA___vals["WindowInteractionState_BlockedByModalWindow"] = 3
Skype__UIA___fullnames["WindowInteractionState_NotResponding"] = 4
Skype__UIA___vals["WindowInteractionState_NotResponding"] = 4
Skype__UIA___fullnames["TextEditChangeType_None"] = 0
Skype__UIA___vals["TextEditChangeType_None"] = 0
Skype__UIA___fullnames["TextEditChangeType_AutoCorrect"] = 1
Skype__UIA___vals["TextEditChangeType_AutoCorrect"] = 1
Skype__UIA___fullnames["TextEditChangeType_Composition"] = 2
Skype__UIA___vals["TextEditChangeType_Composition"] = 2
Skype__UIA___fullnames["TextEditChangeType_CompositionFinalized"] = 3
Skype__UIA___vals["TextEditChangeType_CompositionFinalized"] = 3
Skype__UIA___fullnames["ExpandCollapseState_Collapsed"] = 0
Skype__UIA___vals["ExpandCollapseState_Collapsed"] = 0
Skype__UIA___fullnames["ExpandCollapseState_Expanded"] = 1
Skype__UIA___vals["ExpandCollapseState_Expanded"] = 1
Skype__UIA___fullnames["ExpandCollapseState_PartiallyExpanded"] = 2
Skype__UIA___vals["ExpandCollapseState_PartiallyExpanded"] = 2
Skype__UIA___fullnames["ExpandCollapseState_LeafNode"] = 3
Skype__UIA___vals["ExpandCollapseState_LeafNode"] = 3
Skype__UIA___fullnames["StyleId_BulletedList"] = 70015
Skype__UIA___vals["StyleId_BulletedList"] = 70015
Skype__UIA___fullnames["StyleId_Custom"] = 70000
Skype__UIA___vals["StyleId_Custom"] = 70000
Skype__UIA___fullnames["StyleId_Emphasis"] = 70013
Skype__UIA___vals["StyleId_Emphasis"] = 70013
Skype__UIA___fullnames["StyleId_Heading1"] = 70001
Skype__UIA___vals["StyleId_Heading1"] = 70001
Skype__UIA___fullnames["StyleId_Heading2"] = 70002
Skype__UIA___vals["StyleId_Heading2"] = 70002
Skype__UIA___fullnames["StyleId_Heading3"] = 70003
Skype__UIA___vals["StyleId_Heading3"] = 70003
Skype__UIA___fullnames["StyleId_Heading4"] = 70004
Skype__UIA___vals["StyleId_Heading4"] = 70004
Skype__UIA___fullnames["StyleId_Heading5"] = 70005
Skype__UIA___vals["StyleId_Heading5"] = 70005
Skype__UIA___fullnames["StyleId_Heading6"] = 70006
Skype__UIA___vals["StyleId_Heading6"] = 70006
Skype__UIA___fullnames["StyleId_Heading7"] = 70007
Skype__UIA___vals["StyleId_Heading7"] = 70007
Skype__UIA___fullnames["StyleId_Heading8"] = 70008
Skype__UIA___vals["StyleId_Heading8"] = 70008
Skype__UIA___fullnames["StyleId_Heading9"] = 70009
Skype__UIA___vals["StyleId_Heading9"] = 70009
Skype__UIA___fullnames["StyleId_Normal"] = 70012
Skype__UIA___vals["StyleId_Normal"] = 70012
Skype__UIA___fullnames["StyleId_NumberedList"] = 70016
Skype__UIA___vals["StyleId_NumberedList"] = 70016
Skype__UIA___fullnames["StyleId_Quote"] = 70014
Skype__UIA___vals["StyleId_Quote"] = 70014
Skype__UIA___fullnames["StyleId_Subtitle"] = 70011
Skype__UIA___vals["StyleId_Subtitle"] = 70011
Skype__UIA___fullnames["StyleId_Title"] = 70010
Skype__UIA___vals["StyleId_Title"] = 70010
Skype__UIA___fullnames["UIA_CustomLandmarkTypeId"] = 80000
Skype__UIA___vals["UIA_CustomLandmarkTypeId"] = 80000
Skype__UIA___fullnames["UIA_FormLandmarkTypeId"] = 80001
Skype__UIA___vals["UIA_FormLandmarkTypeId"] = 80001
Skype__UIA___fullnames["UIA_MainLandmarkTypeId"] = 80002
Skype__UIA___vals["UIA_MainLandmarkTypeId"] = 80002
Skype__UIA___fullnames["UIA_NavigationLandmarkTypeId"] = 80003
Skype__UIA___vals["UIA_NavigationLandmarkTypeId"] = 80003
Skype__UIA___fullnames["UIA_SearchLandmarkTypeId"] = 80004
Skype__UIA___vals["UIA_SearchLandmarkTypeId"] = 80004
Skype__UIA___fullnames["ToggleState_Off"] = 0
Skype__UIA___vals["ToggleState_Off"] = 0
Skype__UIA___fullnames["ToggleState_On"] = 1
Skype__UIA___vals["ToggleState_On"] = 1
Skype__UIA___fullnames["ToggleState_Indeterminate"] = 2
Skype__UIA___vals["ToggleState_Indeterminate"] = 2
Skype__UIA___fullnames["Off"] = 0
Skype__UIA___vals["Off"] = 0
Skype__UIA___fullnames["Polite"] = 1
Skype__UIA___vals["Polite"] = 1
Skype__UIA___fullnames["Assertive"] = 2
Skype__UIA___vals["Assertive"] = 2
Skype__UIA___fullnames["SELFLAG_NONE"] = 0
Skype__UIA___vals["SELFLAG_NONE"] = 0
Skype__UIA___fullnames["SELFLAG_TAKEFOCUS"] = 1
Skype__UIA___vals["SELFLAG_TAKEFOCUS"] = 1
Skype__UIA___fullnames["SELFLAG_TAKESELECTION"] = 2
Skype__UIA___vals["SELFLAG_TAKESELECTION"] = 2
Skype__UIA___fullnames["SELFLAG_EXTENDSELECTION"] = 4
Skype__UIA___vals["SELFLAG_EXTENDSELECTION"] = 4
Skype__UIA___fullnames["SELFLAG_ADDSELECTION"] = 8
Skype__UIA___vals["SELFLAG_ADDSELECTION"] = 8
Skype__UIA___fullnames["SELFLAG_REMOVESELECTION"] = 16
Skype__UIA___vals["SELFLAG_REMOVESELECTION"] = 16
Skype__UIA___fullnames["TextRange_NoMaxLength"] = 0xffffffff
Skype__UIA___vals["TextRange_NoMaxLength"] = 0xffffffff
Skype__UIA___fullnames["CaretPosition_Unknown"] = 0
Skype__UIA___vals["CaretPosition_Unknown"] = 0
Skype__UIA___fullnames["CaretPosition_EndOfLine"] = 1
Skype__UIA___vals["CaretPosition_EndOfLine"] = 1
Skype__UIA___fullnames["CaretPosition_BeginningOfLine"] = 2
Skype__UIA___vals["CaretPosition_BeginningOfLine"] = 2
Skype__UIA___fullnames["ActiveEnd_None"] = 0
Skype__UIA___vals["ActiveEnd_None"] = 0
Skype__UIA___fullnames["ActiveEnd_Start"] = 1
Skype__UIA___vals["ActiveEnd_Start"] = 1
Skype__UIA___fullnames["ActiveEnd_End"] = 2
Skype__UIA___vals["ActiveEnd_End"] = 2

Skype__UIA___built = True
endFunction

collection function UIA__propList()
; Return the list of supported property names.
; This is a collection with names as keys and integer property Ids as values.
UIA___buildLists()
return Skype__UIA___props
endFunction

void function UIA___sayError(string s, optional string spec)
; Speak an error message.
if spec
	s = s +": " +spec
endIf
sayMessage(OT_Error, s)
endFunction

;================================ End module UIA ===============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

int function util__waitForKeysToClear(/*?*/ int nTenths)
; Wait up to ntenths of a second for shift-type keys to be released.
; This was initially written for NT-style Windows versions, in which TypeKey (and {keyName} constructs)
; have been seen to fail when such keys are considered down already.
; If nTenths is 0 or not passed, this function will wait indefinitely for the keys to be released.
; The keys waited for are alt, control, shift, and insert.
; Returns True on success and False on failure.
; Fails if GetKeyState fails to get a key state, or if nTenths expires.
; Written 01/02/03 by DGL.
var
	string keyList,
	int i,
	int isDown, int isOn,
	string keyName,
	int count
keyList = "alt|control|shift|insert"
count = 0
i = 1
keyName = stringSegment(keyList, "|", i)
if !nTenths
	nTenths = 0
endIf
while keyName && count <= nTenths
	if !getKeyState(keyName, isDown, isOn)
		; Fail if GetKeyState fails
		return False
	endIf
	if isDown
		; try again in a tenth of a second unless we're out of time
		if nTenths
			count = count +1
		endIf
		if count <= nTenths
			delay(1)
		endIf
	else
		; This key is not down; go to the next one if there is another.
		i = i +1
		keyName = stringSegment(keyList, "|", i)
		; No extra delay or count increment when jumping to another key name.
	endIf
endWhile
; If all keys are up, isDown will be False.
return !isDown
EndFunction

int function util__abs(int n)
; Return the absolute value of n.
if n < 0  return 0-n
else return n
endIf
endFunction

int function util__isSpecialFocus()
; Returns True if there is a special condition for which normal focus handling should not apply.
; Currently, the conditions that are considered special are menusActive, userBufferIsActive, inHJDialog, and switchingTasks (Alt+Tab is active).
return menusActive() ||userBufferIsActive() ||inHJDialog() || util__switchingTasks()
endFunction

int function util__switchingTasks()
; Returns True if the user is currently Alt+Tabbing between tasks.

; This approach is adapted from JAWS 9.0.2169's default.jss FocusChangedEventEx function.
if getWindowClass(getFocus()) == "#32771"
&& getObjectTypeCode()== WT_LISTBOXITEM
	return True
endIf
return False
endFunction

int function util__setRestrictionToWindow(handle hwnd, optional int objID)
; Restrict the active cursor to the given window's borders.
; If objID is passed and non-zero, use that MSAA object ID's window region.
; Example: ObjID_Client restricts to the window's client area.
var int left, int right, int top, int bottom
if !objID
	if !getWindowRect(hwnd, left, right, top, bottom)
		return False
	endIf
	if !left && !right && !top && !bottom
		return False
	endIf
else  ; objID given.
	var object o, int childID
	o = getObjectFromEvent(hwnd, objID, 0, childID)
	if !o
		return False
	endIf
	o.accLocation(intRef(left), intRef(top), intRef(right), intRef(bottom), childID)
	if !left && !right && !top && !bottom
		return False
	endIf
	; but we want 1-based left/right/top/bottom.
	left = left +1
	top = top +1
	right = left +right
	bottom = top +bottom
endIf
return setRestrictionToRect(left, top, right, bottom)
endFunction

string function util__stringFromFile(string fname, /*?*/ string fmt)
; Returns the file contents as a string.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt,
	string s
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 1, 0, iFmt)  ; 1 is ForReading, 0 means don't create
if !f
	;sayString("No code file")
	return ""
endIf
s = f.readAll()
f.close()
if !stringLength(s)
	;sayString("No code in the file")
endIf
return s
endFunction

int function util__stringToFile(string fname, string s, /*?*/ string fmt)
; Writes the string to the indicated file, creating or replacing the file.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 2, 1, iFmt)  ; 2 is ForWriting, 1 means create
if !f
	;sayString("Can't create file")
	return ""
endIf
f.write(s)
f.close()
return True
endFunction

handle function util__getTaskBarWindow()
return findWindow(util__getDesktopWindow(), "MSTaskSwWClass", "")
endFunction

handle function util__getFirstTopLevelWindow()
; Returns the handle of the first window at the top level of the window tree.
var
	handle hwnd

hwnd = util__getDesktopWindow()
if hwnd
	return getFirstChild(hwnd)
endIf

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
hwnd = getFocus()
if !hwnd
	hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf

; Find the top level of the window tree without relying on getAppMainWindow.
; Note that this can actually reach the parent of the top level if we are starting on a ComboLBox window!
while getParent(hwnd)
	hwnd = getParent(hwnd)
endWhile

; make sure this is the level we want, not the desktop itself.
; At least in Windows 95/98, the desktop has window type 14 (desktop), class "#32769", and handle 128.
; The handle is 65566 rather than 128 in Windows NT.  [DGL, 10/07/01]
; We use the first of the above properties because it seems the most likely to remain unique for a window with no parents.
if !getNextWindow(hwnd) && !getPriorWindow(hwnd)
	hwnd = getFirstChild(hwnd)
else
	hwnd = getFirstWindow(hwnd)
endIf
return hwnd
endFunction

handle function util__getDesktopWindow()
; Analogous to the Windows call of the same name.
var
	handle hwnd,
	string hlist, int hcount, int i

; Switching to the invisible cursor keeps GetWindowTypeCode from returning 3 (WT_Edit) for all windows when the virtual cursor is active.
saveCursor() invisibleCursor()

; First try handles known to be the Desktop window in various versions of Windows.
; 128=95/98, 65566=NT, 65556=XP, 65546=2k.
; Thanks to Victor Tsaran for helping collect these handles. [DGL, 04/28/03]
; 65552 for Windows 8.1.
hlist = "128|65566|65556|65552|65546|65536|65526|65516|65506"
hcount = 8
i = 1
while i <= hcount
	hwnd = stringToHandle(stringSegment(hlist, "|", i))
	if getWindowTypeCode(hwnd) == WT_Desktop
		return hwnd
	endIf
	i = i +1
endWhile

; The hard way...

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
hwnd = getFocus()
if !hwnd
	hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf
hwnd = getFirstWindow(getTopLevelWindow(hwnd))
while hwnd
	if getWindowTypeCode(getParent(hwnd)) == WT_Desktop
		return getParent(hwnd)
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

string function util__callStackInfo(int nlevels, string flags, /*?*/ int skip)
; Produce a representation of the current call stack formatted according to the given parameters.
; The following shortenings are applied to each script/function call in the return value:
;	- The shared folder path is replaced with "shared."
;	- The user folder path is removed.
;	- The word "function" is removed.
;	- The word "script" is replaced with "$."
; Requires JAWS 10 or later because the JAWS 10+ getScriptCallStack() is required.
; nlevels: Number of levels from starting position up the stack to print.  1 means just current function.  0 or less means everything.
; flags: f for forward order (default is reverse), s for short (one line and no parameters).
; skip: How many calls to skip over besides this function itself, default 0.
; Returns the requested call stack information or null if the JAWS version is too old.
var
	string userFolder, string sharedFolder,
	int short, int forward,
	int stacksize
if getJFWVersion() < 1000000
	return ""
endIf
flags = stringLower(flags)
short = stringContains(flags, "s")
forward = stringContains(flags, "f")
userFolder = getJAWSSettingsDirectory()
sharedFolder = getSharedSettingsDirectory()
var collection stack = util__callStack()
var int i = 2  ; skip this function's entry.
i = i +skip
stacksize = stack.count
if nlevels > 0
	nlevels = min(nlevels, stacksize-i+1)
else
	nlevels = stacksize-i+1
endIf
if nlevels <= 0
	return ""
endIf
var int delta = 1
if forward
	delta = -1
	i = i +nlevels -1
endIf
var string buf = ""
while nlevels
	if buf
		if forward
			buf = buf +" -> "
		else
			buf = buf +" <- "
		endIf
	endIf
	var collection  entry = stack[intToString(i)]
	var string fname = entry.file
	fname = stringReplaceSubstrings(fname, sharedFolder, "")
	fname = stringReplaceSubstrings(fname, userFolder, "user")
	if stringRight(fname, 4) == ".jsb"
		fname = stringChopRight(fname, 4)
	endIf
	buf = buf +fname +" "
	if entry.type == "script"
		buf = buf +"$"
	endIf
	if short
		buf = buf +stringSegment(entry.call, "(", 1)
	else
		buf = buf +string__ref(entry.call)
	endIf
	i = i +delta
	nlevels = nlevels -1
endWhile
return buf
endFunction

globals object Skype__util___csRE
collection function util__callStack()
; Return a collection of collections representing the script call stack.
; Requires JAWS 11 update 2 or later and the VBScript.RegExp object.
; Index the result by stringified 1-based integers.
; The most recent call is item 1.
; Structure of an item:
;	file: The file from which the script/function came, either "builtin" or a full path.
;	type: The type ("script" or "function", lower case).
;	call: The full call with parameters.
if !Skype__util___csRE
	Skype__util___csRE = util__createObject("VBScript.RegExp")
	if !Skype__util___csRE
		return
	endIf
	Skype__util___csRE.Global = -1
	Skype__util___csRE.IgnoreCase = 0
	Skype__util___csRE.Pattern = "\10([^\10\t]*)\t(script |function )"
endIf
var string sStack = "\10" +getScriptCallStack()
var object matches = Skype__util___csRE.execute(sStack)
var object match
var int n = matches.count
var int i = 1  ; Skip this function itself.
var string tmp, int pos1, int pos2
var collection stack stack = new collection
var collection entry
stack.count = 0
while i < n
	entry = new collection
	stack.count = stack.count +1
	stack[intToString(stack.count)] = entry
	match = matches(i)
	; Skip the \10 here.
	; Also switch from 0-base to 1-base.
	pos1 = match.firstIndex +2
	if i == n-1
		pos2 = stringLength(sStack) +1
	else
		pos2 = matches(i+1).firstIndex +1
	endIf
	tmp = substring(sStack, pos1, pos2-pos1)
	entry.file = string__segmentNext(tmp, "\t")
	entry.type = string__segmentNext(tmp, " ")
	entry.call = tmp
	i = i +1
endWhile
return stack
endFunction

globals object Skype__util___fso
object function util__getFSObject()
; Cached access to a filesystem object.
if !Skype__util___fso
	Skype__util___fso = util__createObject("Scripting.FilesystemObject")
endIf
return Skype__util___fso
endFunction

string function util__getJAWSInfo()
; Returns a list of fields, separated by "|", as follows:
; 1.  The JAWS version number from the JFWUI2 window, in the form [M]M.mm.bbb; example, 5.00.584.
; 2.  The JAWS serial number from the JFWUI2 window.
; 3.  The JAWS serial number from GetJFWSerialNumber().
; The build number returned by this function is accurate, whereas it is "000" in GetJFWVersion in older JAWS versions.
var
	handle hwnd0, handle hwnd,
	int pos,
	string sn, string ver

; Find the JAWS dialog, which exists even when invisible or when JAWS is on the system tray.
hwnd0 = getFirstChild(findTopLevelWindow("JFWUI2", ""))
if !hwnd0
	return ""
endIf

; Get the serial number from the status bar's window name.
hwnd = getNextWindow(hwnd0)
sn = stringLower(util__getFullWindowName(hwnd))
pos = stringContains(sn, "ber")
if pos
	sn = intToString(stringToInt(substring(sn, pos+4, 10)))
else
	sn = ""
endIf

; Find the static containing the version number and get it.
hwnd = getFirstChild(hwnd0)
ver = ""
while hwnd && !ver
	ver = stringTrimTrailingBlanks(util__getFullWindowName(hwnd))
	pos = stringContainsFromRight(ver, " ")
	if pos
		ver = stringChopLeft(ver, pos)
	else
		ver = ""
	endIf
	hwnd = getNextWindow(hwnd)
endWhile

; The above version retrieval sometimes fails, so try the normal way.
if !ver
	ver = intToString(getJFWVersion())
	var int verlen = stringLength(ver)
	ver = formatString("%1.%2.%3",
		stringChopRight(ver, 5),
		substring(ver, verlen-4, 1),
		stringRight(ver, 4)
	)
endIf

if !sn && !ver
	return ""
else
	return ver +"|" +sn +"|" +getJFWSerialNumber()
endIf
endFunction

string function util__getFullWindowName(handle hWnd)
; Gets the full name of the given window if possible, using MSAA when it seems advisable.
; The JAWS GetWindowName() function seems to cut the name at around 255 characters.
; JAWS can get up to 4 K (maybe more sometimes) from MSAA though.
; Some JAWS versions (15?) can also return null for getWindowName when MSAA can get it.
var
	string wName,
	string mName,
	int cutoffLen

; Window names can reach 255 characters, but we choose an earlier cutoff for safety.
cutoffLen = 240

; Zero-length names can mean getWindowName missed it entirely. [DGL, 2014-09-29, JAWS 15.0]
wName = GetWindowName(HWnd)
var int len = stringLength(wName)
if !len || len > cutoffLen
	var int childID, object o
	o = GetObjectFromEvent(HWnd, -4, 0, childID)  ; -4 = ObjID_Client
	mName = o.accName(0)
	if stringLength(mName) > len
		return mName
	endIf
endIf
return wName
EndFunction

int function util__setFocus(handle hwnd, /*?*/ string method, handle hwndDest)
; Enhanced SetFocus function allowing any of a few methods to set focus.
; Example usages:
;	util__setFocus(hwnd):  Same as setFocus(hwnd).
;	util__setFocus(hwnd, "w"):  Use accSelect on the MSAA Window object.
;	util__setFocus(hwnd, "d"):  Use WM_NEXTDLGCTL against hwnd's top-level window.
;	util__setFocus(hwnd, "d", hwndOther):  Use WM_NEXTDLGCTL against hwndOther.
;	util__setFocus(0, "n"):  Use WM_NEXTDLGCTL to move one tab stop forward in the current screen.
; Method:  One of the following:
;	null or not passed:  Default, use the JAWS SetFocus function.
;	w:  Use MSAA accSelect(SELFLAG_TAKEFOCUS, ObjID_Window).
;	c:  Use MSAA accSelect(SELFLAG_TAKEFOCUS, ObjID_Client).
;	d:  Use the WM_NEXTDLGCTL Windows API message.
;	n:  Use WM_NEXTDLGCTL to go to the next tab stop.
;	p:  Use WM_NEXTDLGCTL to go to the previous tab stop.
; hwndDest: Optional and only used with methods d, n, and p above.
; Specifies which window to send the WM_NEXTDLGCTL message to.
; If not passed, the top-level window above hwnd is used.
; For n and p, hwnd may be 0, in which case hwndDest defaults to the current top-level window.
; By default, focus is not changed if focus is already on hwnd.
; Instead, SayObjectTypeAndText() is called to announce as if focus was changed.
; Method can also contain an exclamation (!) to force the focus change even if focus is already on hwnd.
; Returns True if a focus change was attempted and False if not.
var int force = False
if stringContains(method, "!")
	method = stringReplaceSubstrings(method, "!", "")
	force = True
endIf
if !force && hwnd && getFocus() == hwnd
	sayObjectTypeAndText()
	return False
endIf
if !method
	return setFocus(hwnd)
elif method == "w" || method == "c"
	var int objID = 0  ; ObjID_Window
	if method == "c"
		objID = -4  ; ObjID_Client
	endIf
	var int childID
	var object o = getObjectFromEvent(hwnd, objID, 0, childID)
	if !o
		return False
	endIf
	; 1 is SelFlag_TakeFocus
	o.accSelect(1, 0)
	return True
elif method == "d" || method == "n" || method == "p"
	if !hwndDest
		if hwnd
			hwndDest = getTopLevelWindow(hwnd)
		else
			hwndDest = getTopLevelWindow(getFocus())
		endIf
	endIf
	; 40 is WM_NEXTDLGCTL.
	var int msgno = 40
	var int wParam, int lParam
	if method == "n"
		wParam = 0  ; next tab stop
		lParam = 0
	elif method == "p"
		wParam = 1  ; non-zero: previous tab stop.
		lParam = 0
	else  ; "d"
		wParam = util__vcast(hwnd)
		; 1 (True) means set focus to the hwnd given in wParam,
		lParam = 1
	endIf
	sendMessage(hwndDest, msgno, wParam, lParam)
	return True
endIf
return False
endFunction

variant function util__vcast(variant v)
; Cast any type to variant.
return v
endFunction

object function util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o = createObjectEx(objname, False)
if !o
	o = createObject(objname)
endIf
return o
endFunction

int function util__collectionIsEmpty(collection c)
; Return True if the given collection is empty and False if not.
var string k
forEach k in c
	; If this line is reached, there is at least one collection member.
	return False
endForEach
; And if this line is reached, there are none.
return True
endFunction

;=============================== End module util ===============================


;================================ Module string ================================
; String manipulation functions.
;
; Author:  Doug Lee

string function string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContains(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringLeft(s, pos-1)
s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

int function string__segmentCount(string s, string sep)
; Returns the number of segments in the given string.
; If sep is null, the result is 0.
; As of 03/15/03, the null string is defined to have zero segments (previously one).
; Any non-null string containing no separators has one segment.
; WARNING: Unlike the built-in (as of JAWS 7.0) stringSegmentCount function,
; this function counts the null string after a trailing delimiter;
; thus, string__segmentCount("a|b|", "|") is 3, where JAWS would say 2.
; Also, this function considers a multi-character sep as a multi-character delimiter,
; whereas stringSegmentCount allows each character of sep to be a delimiter.
; This function requires JAWS 7.0 or later.
var
	int len,
	int seplen,
	int l,
	int n
seplen = stringLength(sep)
if seplen == 1
	n = stringSegmentCount(s, sep)
	if stringRight(s, 1) == sep
		; JAWS does not count the null string after a trailing delimiter, but I do.
		n = n +1
	endIf
	return n
endIf
len = stringLength(s)
if !len
	return 0
endIf
if !seplen
	return 0
endIf

; A little optimization made possible by JAWS 6.0's StringReplaceSubstrings function...
; Count segments by removing all separators and comparing string lengths.
l = stringLength(string__ref(stringReplaceSubstrings(s, sep, "")))
l = len -l
return (l / seplen) +1
endFunction

int function string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s)
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0")
	whatToDo = "l"
	elif stringToInt(whatToDo)
	whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

whatToDo = stringLower(whatToDo)
trimEnds = stringContains(whatToDo, "t")
considerCase = stringContains(whatToDo, "c")
fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e"
	if trimEnds
		sa = stringStripAllBlanks(s)
	else
		sa = string__ref(s)
	endIf
	s = ""
	l1 = stringLength(sa)
	pos = 1
	while pos < l1
		s = string__cat(s, string__cat(substring(sa, pos, 1), r2))
		pos = pos +1
	endWhile
	s = string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1)
	if fromWhere == "l" || fromWhere == "a"
		if trimEnds
			sb = stringTrimLeadingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringLeft(sb, 1)
		sb = stringChopLeft(sb, 1)
		if trimEnds
			sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds
			sb = stringTrimTrailingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringChopRight(sb, 1)
		if trimEnds
			sa = stringTrimTrailingBlanks(sa)
		endIf
		sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb)
		return 0
	endIf
	s = string__cat(sa, string__cat(r2, sb))
	return 1
endIf

changeCount = 0
while !done
	l1 = stringLength(r1)
	changed = false
	if fromWhere == "l"
		if !stringCompare(stringLeft(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				s = r2 + s
			else
				s = r2 + stringChopLeft(s, l1)
			endIf
			changed = true
		endIf
	elif fromWhere == "r"
		if !stringCompare(stringRight(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				s = s + r2
			else
				s = stringChopRight(s, l1) + r2
			endIf
			changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase
			pos = StringContains(s, r1)
		else
			pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos
			if trimEnds
				sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				s = " " +sa +r2 +sb
				s = stringChopLeft(s, 1)
			else
				s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				s = stringChopLeft(s, 1)
			endIf
			changed = True
		endIf
	endIf
	if fromWhere == "e"
		done = !changed
	else
		done = True
	endIf
	if changed
		changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function string__join(string s1, string sep, string s2)
; Returns the two given strings joined with the given separator.
; If s1 or s2 is empty, the separator is not included.
if stringIsBlank(s1)
	return s2
elif stringIsBlank(s2)
	return s1
else
	return s1 +sep +s2
endIf
endFunction

string function string__quantify(string s)
; A bit like FormatString for singular/plural issues:
; In s, any sequence of the form {<n>:<s1>/<s2>} becomes <s1> if <n> is 1 and <s2> otherwise.
; Also, any sequence of the form {<n>:<s0>/<s1>/<s2>} becomes <s0> if n is 0, <s1> if <n> is 1, and <s2> otherwise.
; Works well in conjunction with FormatString().
; Example: string__quantify(formatString("{%1:No calls/%1 call/%1 calls} {%1:has/have} come in today.", intToString(nchats))):
; nchats=0: "No calls have come in today."
; nchats=1: "1 call has come in today."
; nchats=2: "2 calls have come in today."
; Substrings can be null; example: string__quantify(formatString("%1 call{%1:/s}", intToString(nchats)))
; (0 calls, 1 call, 2 calls, etc.).
var
	int pos1, int pos2,
	string tbuf,
	int n, string lhs, string rhs,
	int zeroSeg,
	string result,
	int safetyCount
safetyCount = 500
result = ""
while safetyCount
	pos1 = stringContains(s, "{")
	if !pos1
		return string__cat(result, s)
	endIf
	pos2 = stringContains(stringChopLeft(s, pos1), "}")
	if !pos2
		return string__cat(result, s)
	endIf
	pos2 = pos2 +pos1
	tbuf = substring(s, pos1+1, pos2-pos1-1) ; string between but not including braces
	lhs = stringLeft(s, pos1-1)  ; up to but not including {
	rhs = stringChopLeft(s, pos2)  ; after but not including }
	if stringContains(tbuf, ":") && stringContains(tbuf, "/")
		n = stringToInt(stringSegment(tbuf, ":", 1))
		tbuf = stringChopLeft(tbuf, stringContains(tbuf, ":"))
		; The next line returns 0 if there's only one slash and 1 if more than one slash.
		zeroSeg = (stringContains(tbuf, "/") != stringContainsFromRight(tbuf, "/"))
		if n == 0 && zeroSeg
			n = 1
		elif n == 1
			n = 1 +zeroSeg
		else
			n = 2 +zeroSeg
		endIf
		tbuf = stringSegment(tbuf, "/", n)
	else
		tbuf = "{" +tbuf +"}"  ; original material not already in lhs/rhs
	endIf
	result = string__cat(result, string__cat(lhs, tbuf))
	s = rhs +""
	safetyCount = safetyCount -1
endWhile
; TODO: Report.
return result
endFunction

string function string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;============================== End module string ==============================

